#include <stdio.h>
#include <stdlib.h>

#if defined(WIN32)
#  include <windows.h>
#  include "GLee.h"
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#  include <GL/glxext.h>
#elif defined(MACOS)
#  include <AGL/agl.h>
#endif

#include <string>

#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "RTTBuffer.h"

#include "Compositor.h"

using namespace glrc;

/**
 * Constructor. You can create an instance without valid OpenGL contexts.
 */
Compositor::Compositor()
{
    m_bEnabled = true;

    m_x = 0.0f;
    m_y = 0.0f;
    m_width = 1.0f;
    m_height = 1.0f;

    m_pRTTBuffer = new RTTBuffer;

    m_bBlendEnabled = true;
    m_rotationAngle = 0.0f;
    m_scaleX = 1.0f;
    m_scaleY = 1.0f;
    m_flip = GLRC_FLIP_NONE;

    m_effectMode = GLRC_EFFECT_ALL;
    m_gamma[0] = 1.0f;
    m_gamma[1] = 1.0f;
    m_gamma[2] = 1.0f;
    m_gamma[3] = 1.0f;
    m_exposure[0] = 1.0f;
    m_exposure[1] = 1.0f;
    m_exposure[2] = 1.0f;
    m_exposure[3] = 1.0f;

    m_bModelViewMatrixEnabled = false;
    m_bProjectionMatrixEnabled = false;
    for (int i=0; i<6; i++)
        m_frustum[i] = 0.0f;

    m_shaderProgIndex = 0;

    m_distTexIndex = 0;
    m_distMeshIndex = 0;
    m_bMeshDistortion = true;
    m_distMeshHorResolution = 64;
    m_distMeshVertResolution = 64;
    m_blendTexIndex = 0;
}

/**
 * Destructor.
 */
Compositor::~Compositor()
{
    if (m_shaderProgIndex) {
#if defined(OPENGL2)
        glDeleteProgram(m_shaderProgIndex);
#endif // OPENGL2;
#if defined(OPENGL15)
        glDeleteObjectARB(m_shaderProgIndex);
#endif // OPENGL15;
    }

    delete m_pRTTBuffer;

    if (glIsTexture(m_distTexIndex))
        glDeleteTextures(1, &m_distTexIndex);
    if (glIsTexture(m_blendTexIndex))
        glDeleteTextures(1, &m_blendTexIndex);
    if (glIsList(m_distMeshIndex))
        glDeleteLists(m_distMeshIndex, 1);
}

/**
 * Initialize the off-screen buffer. You must initialize the buffer in a valid
 * OpenGL context before rendering.
 * Floating point buffer is not available when rttType is READPIXEL nor COPYSUBTEX.
 * At this time, width and height must be power of two!!
 * You can call setup() more than once to re-configure the off-screen buffer.
 *
 * @param rttType RenderToTexture type. GLRC_RTT_AUTO to select the best type automatically.
 * @param width Width of the off-screen buffer.
 * @param height Height of the off-screen buffer.
 * @param bHasAlpha True if you need alpha channel.
 * @param bFPBuffer True if you need a floating point buffer.
 */
void Compositor::setup(GLRC_RTT type, int width, int height, bool bHasAlpha, bool bFPBuffer)
{
    m_pRTTBuffer->init(type, width, height, bHasAlpha, bFPBuffer);
    buildShader();
}

/**
 * Retrieve RenderToTexture type of the off-screen buffer.
 *
 * @return RenderToTexture type of the off-screen buffer.
 */
GLRC_RTT Compositor::getBufferType() const
{
    return m_pRTTBuffer->getType();
}

/**
 * Retrieve width of the off-screen buffer.
 *
 * @return Width of the off-screen buffer.
 */
int Compositor::getBufferWidth() const
{
    return m_pRTTBuffer->getWidth();
}

/**
 * Retrieve height of the off-screen buffer.
 *
 * @return Height of the off-screen buffer.
 */
int Compositor::getBufferHeight() const
{
    return m_pRTTBuffer->getHeight();
}

/**
 * Check whether the off-screen has alpha channel or not.
 *
 * @return True if the off-screen buffer has alpha channel.
 */
bool Compositor::hasAlpha() const
{
    return m_pRTTBuffer->hasAlpha();
}

/**
 * Check whether the off-screen is floating buffer type or not.
 *
 * @return True if the off-screen buffer is a floating point buffer.
 */
bool Compositor::isFPBuffer() const
{
    return m_pRTTBuffer->isFPBuffer();
}

/**
 * Enable mesh distortion or not (use shader for distiortion).
 *
 * @param bEnable True to use mesh distortion, false to use shader for distortion.
 */
void Compositor::setMeshDistortion(bool bEnable)
{
    if (m_bMeshDistortion != bEnable)
    {
        m_bMeshDistortion = bEnable;

        buildShader();
        setDistortionMap(m_distortionFileName.c_str());
    }
}

/**
 * Check whether mesh distortion is enable or not.
 *
 * @return True to use mesh distortion, false to use shader for distortion.
 */
bool Compositor::isMeshDistortion() const
{
    return m_bMeshDistortion;
}

/**
 * Set resolution of the distortion mesh.
 *
 * @param horResolution Horizontal resolution of te distortion mesh.
 * @param vertResolution Vertical resolution of te distortion mesh.
 */
void Compositor::setDistortionMeshResolution(int horResolution, int vertResolution)
{
    if (horResolution == 0)
        horResolution = 64;
    if (vertResolution == 0)
        vertResolution = 64;

    if (m_bMeshDistortion &&
        (m_distMeshHorResolution != horResolution ||
        m_distMeshVertResolution != vertResolution))
    {
        m_distMeshHorResolution = horResolution;
        m_distMeshVertResolution = vertResolution;
        if (m_distTexIndex != 0)
            setDistortionMap(m_distortionFileName.c_str());
    }
}

/**
 * Retrieve horizontal resolution of the distortion mesh.
 *
 * @return Horizontal resolution of te distortion mesh.
 */
int Compositor::getHorizontalResolutionOfDisotrtionMesh() const
{
    return m_distMeshHorResolution;
}

/**
 * Retrieve vertical resolution of the distortion mesh.
 *
 * @return Vertical resolution of te distortion mesh.
 */
int Compositor::getVerticalResolutionOfDisotrtionMesh() const
{
    return m_distMeshVertResolution;
}

/**
 * Set rendering area of the compositor in current OpenGL viewport.
 * Position and size is defined in normalized (0.0-1.0) coordinate
 * in the current OpenGL viewport.
 *
 * @param x X (left) position of the compositor in normalized (0.0-1.0) coordinate.
 * @param y Y (top) position of the compositor in normalized (0.0-1.0) coordinate.
 * @param width Width of the compositor in normalized (0.0-1.0) coordinate.
 * @param height Height of the compositor in normalized (0.0-1.0) coordinate.
 */
void Compositor::setViewport(float x, float y, float width, float height)
{
    m_x = x;
    m_y = y;
    m_width = width;
    m_height = height;
}

/**
 * Retrieve x (left) position of the compositor.
 *
 * @param X (left) position of the compositor in normalized (0.0-1.0) coordinate.
 */
float Compositor::getX() const
{
    return m_x;
}

/**
 * Retrieve y (left) position of the compositor.
 *
 * @param Y (left) position of the compositor in normalized (0.0-1.0) coordinate.
 */
float Compositor::getY() const
{
    return m_y;
}

/**
 * Retrieve width of the compositor.
 *
 * @param Width of the compositor in normalized (0.0-1.0) coordinate.
 */
float Compositor::getWidth() const
{
    return m_width;
}

/**
 * Retrieve height of the compositor.
 *
 * @param Height of the compositor in normalized (0.0-1.0) coordinate.
 */
float Compositor::getHeight() const
{
    return m_height;
}

/**
 * Enable or disable off-screen rendering (and composite processing).
 * When disabled, both beginRender() / endRender() do nothing, so you can get usual rendering results.
 *
 * @param bEnabled True to enable off-screen rendering (and composite processing).
 */
void Compositor::setEnabled(bool bEnabled)
{
    m_bEnabled = bEnabled;
}

/**
 * Check whether off-screen rendering (and composite processing) is enabled or not.
 *
 * @return True if off-screen rendering (and composite processing) is enabled.
 */
bool Compositor::getEnabled() const
{
    return m_bEnabled;
}

/**
 * Declare to begin to render your scene. Activate off-screen rendering to capture the rendering result,
 * and modify current ModelView / Projection matrices.
 */
void Compositor::beginRender()
{
    if (!m_bEnabled)
        return;

    m_pRTTBuffer->beginRenderToTexture();

    if (m_bProjectionMatrixEnabled)
    {
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glMultMatrixf(m_projectionMatrix.ptr());
        glMatrixMode(GL_MODELVIEW);
    }
    if (m_bModelViewMatrixEnabled)
    {
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();
        glMultMatrixf(m_modelViewMatrix.ptr());
    }
}

/**
 * Declare to end the rendering. Deactivate off-screen rendering,
 * then render the compositor object to on-screen (or other Compositor)
 * with composite processings.
 */
void Compositor::endRender()
{
    if (!m_bEnabled)
        return;

    if (m_bModelViewMatrixEnabled)
    {
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
    }
    if (m_bProjectionMatrixEnabled)
    {
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
    }

    m_pRTTBuffer->endRenderToTexture();

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    m_pRTTBuffer->beginUseTexture();

    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    glViewport((GLsizei)(m_x * viewport[2]), (GLsizei)(m_y * viewport[3]),
               (GLsizei)(m_width * viewport[2]), (GLsizei)(m_height * viewport[3]));

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(-0.5, 0.5, 0.5, -0.5, 0.01, 100);

    glMatrixMode(GL_MODELVIEW);

	if (m_rotationAngle != 0.0f)
	    glRotatef(m_rotationAngle, 0.0f, 0.0f, 1.0f);
	if (m_scaleX != 1.0f || m_scaleY != 1.0f)
	    glScalef(m_scaleX, m_scaleY, 1.0f);
    if (m_flip & GLRC_FLIP_HORIZONTAL)
        glScaled(-1.0f, 1.0f, 1.0f);
    if (m_flip & GLRC_FLIP_VERTICAL)
        glScaled(1.0f, -1.0f, 1.0f);

    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);

    if (m_effectMode != GLRC_EFFECT_NONE &&
        !((m_effectMode & GLRC_EFFECT_DISTORTION) && m_bMeshDistortion))
    {
#if defined(OPENGL2)
        glUseProgram(m_shaderProgIndex);
        glUniform1i(glGetUniformLocation(m_shaderProgIndex, "textureImage"), 0);
        if (m_effectMode & GLRC_EFFECT_DISTORTION && !m_bMeshDistortion)
            glUniform1i(glGetUniformLocation(m_effectProgIndex, "textureDistortion"), 1);
        if (m_effectMode & GLRC_EFFECT_HDR) {
            glUniform4fv(glGetUniformLocation(m_shaderProgIndex, "exposure"), 1, m_exposure);
            glUniform4fv(glGetUniformLocation(m_shaderProgIndex, "gamma"), 1, m_gamma);
        }
        if (m_effectMode & GLRC_EFFECT_COLORMATRIX) {
            glUniformMatrix4fv(glGetUniformLocation(m_shaderProgIndex, "colorMatrix"), 1, false, m_colorMatrix.ptr());
        }
#endif // OPENGL2
#if defined(OPENGL15)
        glUseProgramObjectARB(m_shaderProgIndex);
        glUniform1iARB(glGetUniformLocationARB(m_shaderProgIndex, "textureImage"), 0);
        if (m_effectMode & GLRC_EFFECT_DISTORTION && !m_bMeshDistortion)
            glUniform1iARB(glGetUniformLocationARB(m_shaderProgIndex, "textureDistortion"), 1);
        if (m_effectMode & GLRC_EFFECT_HDR) {
            glUniform4fvARB(glGetUniformLocationARB(m_shaderProgIndex, "exposure"), 1, m_exposure);
            glUniform4fvARB(glGetUniformLocationARB(m_shaderProgIndex, "gamma"), 1, m_gamma);
        }
        if (m_effectMode & GLRC_EFFECT_COLORMATRIX) {
            glUniformMatrix4fvARB(glGetUniformLocationARB(m_shaderProgIndex, "colorMatrix"), 1, false, m_colorMatrix.ptr());
        }
#endif // OPENGL15
    }

    if (m_bBlendEnabled)
    {
        glEnable(GL_BLEND);
//      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    }
	else
	{
        glDisable(GL_BLEND);
	}

    if ((m_effectMode & GLRC_EFFECT_DISTORTION) && m_bMeshDistortion && m_distMeshIndex != 0)
    {
        glPushMatrix();
        glTranslatef(-0.5f, -0.5f, -1.0f);
        glCallList(m_distMeshIndex);
        glPopMatrix();
    }
    else
    {
        glBegin(GL_QUADS);
        // offscreen rendered image is upside down!
        glTexCoord2d(0, 1); glVertex3d(-0.5, -0.5, -1);
        glTexCoord2d(1, 1); glVertex3d( 0.5, -0.5, -1);
        glTexCoord2d(1, 0); glVertex3d( 0.5,  0.5, -1);
        glTexCoord2d(0, 0); glVertex3d(-0.5,  0.5, -1);
        glEnd();
    }

    if (m_effectMode != GLRC_EFFECT_NONE &&
        !(m_effectMode & GLRC_EFFECT_DISTORTION && m_bMeshDistortion)) {
#if defined(OPENGL2)
        glUseProgram(0);
#endif // OPENGL2
#if defined(OPENGL15)
        glUseProgramObjectARB(0);
#endif // OPENGL15
    }

    if ((m_effectMode & GLRC_EFFECT_DISTORTION) && m_blendTexIndex != 0)
    {
        glBindTexture(GL_TEXTURE_2D, m_blendTexIndex);
        glEnable(GL_BLEND);
        glBlendFunc(GL_DST_COLOR, GL_ZERO);

        glBegin(GL_QUADS);
        glTexCoord2d(0, 0); glVertex3d(-0.5, -0.5, -1);
        glTexCoord2d(1, 0); glVertex3d( 0.5, -0.5, -1);
        glTexCoord2d(1, 1); glVertex3d( 0.5,  0.5, -1);
        glTexCoord2d(0, 1); glVertex3d(-0.5,  0.5, -1);
        glEnd();
    }

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    m_pRTTBuffer->endUseTexture();

    glPopAttrib();
}

/**
 * Enable or disable compositor rendering with OpenGL blending.
 * When blending is enabled and cleared with proper alpha value ( < 1.0),
 * you can overlay the rendering results with transparent background.
 *
 * @param bEnabled True to composite with blending.
 */
void Compositor::setBlendEnabled(bool bEnabled)
{
    m_bBlendEnabled = bEnabled;
}

/**
 * Check whether compositiong with blending is enabled or not.
 *
 * @return True if composition with blending is enabled.
 */
bool Compositor::getBlendEnabled() const
{
    return m_bBlendEnabled;
}

/**
 * Set rotation angle of the compositor.
 *
 * @param angle Rotation angle in degree (0.0 is default).
 */
void Compositor::setRotation(float angle)
{
    m_rotationAngle = angle;
}

/**
 * Retrieve rotation angle of the compositor.
 *
 * @return Rotation angle in degree (0.0 is default).
 */
float Compositor::getRotation() const
{
    return m_rotationAngle;
}

/**
 * Set scale of the compositor.
 *
 * @param x Horizontal scale of the compositor (1.0 is default).
 * @param y Vertical scale of the compositor (1.0 is default).
 */
void Compositor::setScale(float x, float y)
{
    m_scaleX = x;
    m_scaleY = y;
}

/**
 * Retrieve horizontal scale of the compositor.
 *
 * @return Horizontal scale of the compositor (1.0 is default).
 */
float Compositor::getScaleX() const
{
    return m_scaleX;
}

/**
 * Retrieve vertical scale of the compositor.
 *
 * @return Vertical scale of the compositor (1.0 is default).
 */
float Compositor::getScaleY() const
{
    return m_scaleY;
}

/**
 * Set flipping mode of the compositor.
 *
 * @param flip Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
 */
void Compositor::setFlip(unsigned int flip)
{
    m_flip = flip;
}

/**
 * Retrieve flipping mode of the compositor.
 *
 * @return Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
 */
unsigned int Compositor::getFlip() const
{
    return m_flip;
}

/**
 * Enable or disable model-view matrix modification.
 *
 * @param bEnabled True to enable model-view matrix modification.
 */
void Compositor::setModelViewMatrixEnabled(bool bEnabled)
{
    m_bModelViewMatrixEnabled = bEnabled;
}

/**
 * Check whether model-view matrix modification is enabled or not.
 *
 * @return True if model-view matrix modification is enabled.
 */
bool Compositor::getModelViewMatrixEnabled() const
{
    return m_bModelViewMatrixEnabled;
}

/**
 * Set model-view modification matrix.
 *
 * @param matrix Model-view modification matrix.
 */
void Compositor::setModelViewMatrix(const Matrix& matrix)
{
    m_modelViewMatrix = matrix;
}

/**
 * Retrieve model-view modification matrix.
 *
 * @return Model-view modification matrix.
 */
Matrix Compositor::getModelViewMatrix() const
{
    return m_modelViewMatrix;
}

/**
 * Retrieve translation of the model-view modification matrix.
 *
 * @param x Distance in x-axis direction.
 * @param y Distance in y-axis direction.
 * @param z Distance in z-axis direction.
 */
void Compositor::getModelViewTranslation(float& x, float& y, float& z) const
{
    m_modelViewMatrix.getTranslation(x, y, z);
}

/**
 * Retrieve rotation of the model-view modification matrix.
 *
 * @param angle Angle of rotation in degree.
 * @param x X element value of the axis vector.
 * @param y Y element value of the axis vector.
 * @param z Z element value of the axis vector.
 */
void Compositor::getModelViewRotation(float& angle, float& x, float& y, float& z) const
{
    m_modelViewMatrix.getRotation(angle, x, y, z);
}

/**
 * Retrieve quatenion rotation of the model-view modification matrix.
 *
 * @param x X element of quatenion.
 * @param y Y element of quatenion.
 * @param z Z element of quatenion.
 * @param w W element of quatenion.
 */
void Compositor::getModelViewQuatenion(float& x, float&y, float& z, float& w) const
{
    m_modelViewMatrix.getQuatenion(x, y, z, w);
}

/**
 * Retrieve scaling of the model-view modification matrix.
 *
 * @param x Scaling factor in x-axis.
 * @param y Scaling factor in y-axis.
 * @param z Scaling factor in z-axis.
 */
void Compositor::getModelViewScaling(float& x, float& y, float& z) const
{
    m_modelViewMatrix.getScaling(x, y, z);
}

/**
 * Enable or disable projection matrix modification.
 *
 * @param bEnabled True to enable projection matrix modification.
 */
void Compositor::setProjectionMatrixEnabled(bool bEnabled)
{
    m_bProjectionMatrixEnabled = bEnabled;
}

/**
 * Check whether projection matrix modification is enabled or not.
 *
 * @return True if projection matrix modification is enabled.
 */
bool Compositor::getProjectionMatrixEnabled() const
{
    return m_bProjectionMatrixEnabled;
}

/**
 * Set projection modification matrix.
 *
 * @param matrix Projection modification matrix.
 */
void Compositor::setProjectionMatrix(const Matrix& matrix)
{
    m_projectionMatrix = matrix;
    // decode to frustum...
}

/**
 * Retrieve projection modification matrix.
 *
 * @param matrix Projection modification matrix.
 */
Matrix Compositor::getProjectionMatrix() const
{
    return m_projectionMatrix;
}

/**
 * Set projection modification matrix in glFrustum format.
 *
 * @param frustum Projection modification matrix in glFrustum format.
 */
void Compositor::setProjectionFrustum(float frustum[6])
{
    m_frustum[0] = frustum[0];
    m_frustum[1] = frustum[1];
    m_frustum[2] = frustum[2];
    m_frustum[3] = frustum[3];
    m_frustum[4] = frustum[4];
    m_frustum[5] = frustum[5];

    m_projectionMatrix.makeFrustum(
        m_frustum[0], m_frustum[1], m_frustum[2],
        m_frustum[3], m_frustum[4], m_frustum[5]);
}

/**
 * Retrieve projection modification matrix in glFrustum format.
 *
 * @param frustum Projection modification matrix in glFrustum format.
 */
void Compositor::getProjectionFrustum(float frustum[6]) const
{
    frustum[0] = m_frustum[0];
    frustum[1] = m_frustum[1];
    frustum[2] = m_frustum[2];
    frustum[3] = m_frustum[3];
    frustum[4] = m_frustum[4];
    frustum[5] = m_frustum[5];
}

/**
 * Load matrix configuration file.
 *
 * @param fileName Matrix configuration file name.
 * @return True if loaded successfully.
 */
bool Compositor::loadMatrixFile(const char* fileName)
{
    FILE* fp = fopen(fileName, "r");
    if (!fp) return false;

    char token[1024];
    while (fscanf(fp, "%s", token)==1)
    {
        if (strcmp(token, "Offset")==0) {
            m_bModelViewMatrixEnabled = true;
            m_modelViewMatrix.makeIdentity();
        }
        else if (strcmp(token, "Translate")==0) {
            float x, y, z;
            fscanf(fp, "%f %f %f", &x, &y, &z);
            m_modelViewMatrix.translate(x, y, z);
        }
        else if (strcmp(token, "Rotate")==0) {
            float angle, x, y, z;
            fscanf(fp, "%f %f %f %f", &angle, &x, &y, &z);
            m_modelViewMatrix.rotate(-angle, x, y, z);
        }
        else if (strcmp(token, "Scale")==0) {
            float x, y, z;
            fscanf(fp, "%f %f %f", &x, &y, &z);
            m_modelViewMatrix.scale(x, y, z);
        }
        else if (strcmp(token, "Lens")==0) {
            m_bProjectionMatrixEnabled = true;
            m_projectionMatrix.makeIdentity();
        }
        else if (strcmp(token, "Frustum")==0) {
            fscanf(fp, "%f %f %f %f %f %f",
                &m_frustum[0], &m_frustum[1], &m_frustum[2],
                &m_frustum[3], &m_frustum[4], &m_frustum[5]);
            m_projectionMatrix.makeFrustum(
                m_frustum[0], m_frustum[1], m_frustum[2],
                m_frustum[3], m_frustum[4], m_frustum[5]);
        }
    }
    fclose(fp);

    return true;
}

/**
 * Set effect processing mode to composite the rendering results.
 *
 * @param effectMode Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
 */
void Compositor::setEffectMode(unsigned int effectMode)
{
    if (m_effectMode != effectMode)
    {
        m_effectMode = effectMode;
        buildShader();
    }
}

/**
 * Retrieve effect processing mode to composite the rendering results.
 *
 * @return Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
 */
unsigned int Compositor::getEffectMode() const
{
    return m_effectMode;
}

/**
 * Load a distortion map file to distort the rendering results.
 *
 * @param fileName Distortion map file name.
 * @return True if the distortion map is loaded successfully.
 */
bool Compositor::setDistortionMap(const char* fileName)
{
    m_distortionFileName = fileName;

    ILuint id;
    ilGenImages(1, &id);
    ilBindImage(id);
    if (!ilLoadImage((ILstring)fileName))
        return false;

    if (!m_bMeshDistortion)
    {
        glActiveTexture(GL_TEXTURE1);
        glEnable(GL_TEXTURE_2D);

        if (glIsTexture(m_distTexIndex))
            glDeleteTextures(1, &m_distTexIndex);
        m_distTexIndex = ilutGLBindTexImage();
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glDisable(GL_TEXTURE_2D);
        glActiveTexture(GL_TEXTURE0);

        ilDeleteImages(1, &id);

        return (glIsTexture(m_distTexIndex)==GL_TRUE);
    }
    else
    {
        ILubyte* pData = ilGetData();
        int bytesPerPixel = ilGetInteger(IL_IMAGE_BYTES_PER_PIXEL);

        if (glIsList(m_distMeshIndex))
            glDeleteLists(m_distMeshIndex, 1);
        m_distMeshIndex = glGenLists(1);
        glNewList(m_distMeshIndex, GL_COMPILE);
        int imageWidth = ilGetInteger(IL_IMAGE_WIDTH);
        int imageHeight = ilGetInteger(IL_IMAGE_HEIGHT);
        float dx = 1.0f / m_distMeshHorResolution;
        float dy = 1.0f / m_distMeshHorResolution;
        float dtx = (float)(imageWidth-1) / m_distMeshHorResolution;
        float dty = (float)(imageHeight-1) / m_distMeshVertResolution;
        for (int y=0; y<m_distMeshVertResolution; ++y)
        {
            glBegin(GL_QUAD_STRIP);
            for (int x=0; x<=m_distMeshHorResolution; ++x)
            {
                for (int yy=0; yy<2; ++yy)
                {
                    int pos = imageWidth*bytesPerPixel*(int)(dty*(y+yy))+bytesPerPixel*(int)(dtx*x);
                    int r = pData[pos+0];
                    int g = pData[pos+1];
                    int b = pData[pos+2];
                    float tx = ((float)b / 255.0f + (r % 16)) / 16.0f;
                    float ty = 1.0f - ((float)g / 255.0f + (r / 16)) / 16.0f;
                    glTexCoord2f(tx, ty);
                    glVertex3f(dx*x, dy*(y+yy), 0.0f);
                }
            }
            glEnd();
        }
        glEndList();
    }

    ilDeleteImages(1, &id);

    return true;
}

/**
 * Retrieve the distortion map file name currently loaded.
 *
 * @return Distortion map file name currently loaded.
 */
const char* Compositor::getDistortionMapFileName() const
{
    return m_distortionFileName.c_str();
}

/**
 * Load a blend map file to blend the rendering results.
 *
 * @param fileName Blend map file name.
 * @return True if the blend map is loaded successfully.
 */
bool Compositor::setBlendMap(const char* fileName)
{
    m_blendFileName = fileName;

    ILuint id;
    ilGenImages(1, &id);
    ilBindImage(id);
    if (!ilLoadImage((ILstring)fileName))
        return false;

    if (glIsTexture(m_blendTexIndex))
        glDeleteTextures(1, &m_blendTexIndex);
    m_blendTexIndex = ilutGLBindTexImage();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    ilDeleteImages(1, &id);

    return true;
}

/**
 * Retrieve the blend map file name currently loaded.
 *
 * @return Blend map file name currently loaded.
 */
const char* Compositor::getBlendMapFileName() const
{
    return m_blendFileName.c_str();
}

/**
 * Set gamma value for tone mapping of high-dynamic range rendering.
 * LDR = (HDR * exposure) ^ gamma.
 *
 * @param Gamma value (1.0 is default).
 */
void Compositor::setGamma(float gamma)
{
    m_gamma[0] = gamma;
    m_gamma[1] = gamma;
    m_gamma[2] = gamma;
    m_gamma[3] = 1.0;
}

/**
 * Retrieve gamma value for high-dynamic range rendering.
 *
 * @return Gamma value (1.0 is default).
 */
float Compositor::getGamma() const
{
    return m_gamma[0];
}

/**
 * Set exposure value for tone mapping of high-dynamic range rendering.
 * LDR = (HDR * exposure) ^ gamma.
 *
 * @param Exposure value (1.0 is default).
 */
void Compositor::setExposure(float exposure)
{
    m_exposure[0] = exposure;
    m_exposure[1] = exposure;
    m_exposure[2] = exposure;
    m_exposure[3] = 1.0f;
}

/**
 * Retrieve exposure value for high-dynamic range rendering.
 *
 * @return Exposure value (1.0 is default).
 */
float Compositor::getExposure() const
{
    return m_exposure[0];
}

/**
 * Set color conversion matrix.
 *
 * @param matrix Color conversion matrix.
 */
void Compositor::setColorMatrix(const Matrix& matrix)
{
    m_colorMatrix = matrix;
}

/**
 * Retrieve color conversion matrix.
 *
 * @result Color conversion matrix.
 */
Matrix Compositor::getColorMatrix() const
{
    return m_colorMatrix;
}

/**
 * Save the current off-screen buffer to an image file.
 * Image file format is determined by its file extention.
 * Supported format is depending on DevIL library (and .hdr format).
 *
 * @param Image file name to save the off-screen buffer.
 * @return True if image is saved successfully.
 */
bool Compositor::saveBuffer(const char* fileName)
{
    return m_pRTTBuffer->saveBuffer(fileName);
}

/**
 * (Re-)build OpenGL shader depends on current effect mode. Called by setEffectMode().
 */
void Compositor::buildShader()
{
    if (m_shaderProgIndex) {
#if defined(OPENGL2)
        glDeleteProgram(m_shaderProgIndex);
#endif // OPENGL2;
#if defined(OPENGL15)
        glDeleteObjectARB(m_shaderProgIndex);
#endif // OPENGL15;
    }

    std::string vertProgramText =
        "void main()\n"
        "{\n"
        "gl_Position = ftransform();\n"
        "gl_TexCoord[0] = gl_MultiTexCoord0;\n"
        "}\n";

    std::string fragProgramText =
        "uniform sampler2D textureImage;\n";
    if (m_effectMode & GLRC_EFFECT_HDR) {
        fragProgramText +=
            "uniform vec4 exposure;\n"
            "uniform vec4 gamma;\n";
    }
    if (m_effectMode & GLRC_EFFECT_DISTORTION && !m_bMeshDistortion) {
        fragProgramText +=
            "uniform sampler2D textureDistortion;\n";
    }
    if (m_effectMode & GLRC_EFFECT_COLORMATRIX) {
        fragProgramText +=
            "uniform mat4 colorMatrix;\n";
    }
    fragProgramText +=
        "void main()\n"
        "{\n";
    if (m_effectMode & GLRC_EFFECT_DISTORTION && !m_bMeshDistortion) {
        fragProgramText +=
            "    vec4 texCoord;\n"
            "    vec4 distColor;\n"
            "    vec4 tmp = gl_TexCoord[0];\n"
            "    tmp.t = 1.0 - tmp.t\n;"
            "    distColor = texture2D(textureDistortion, tmp.st);\n"
            "\n"
            "    texCoord.s = (distColor.b + mod(floor(distColor.r * 255.5), 16.0)) / 16.0;\n"
            "    texCoord.t = 1.0 - (distColor.g + floor(floor(distColor.r * 255.5) / 16.0)) / 16.0;\n"
            "\n"
            "    gl_FragColor = texture2D(textureImage,  texCoord.st);\n";
    }
    else
    {
        fragProgramText +=
            "    gl_FragColor = texture2D(textureImage, gl_TexCoord[0].st);\n";
    }
    if (m_effectMode & GLRC_EFFECT_HDR) {
        fragProgramText +=
        "    gl_FragColor = gl_FragColor * exposure;\n"
        "    gl_FragColor = pow(gl_FragColor, gamma);\n";
    }
    if (m_effectMode & GLRC_EFFECT_COLORMATRIX) {
        fragProgramText +=
        "    gl_FragColor = colorMatrix * gl_FragColor;\n";
    }
    fragProgramText +=
        "}\n";

    m_shaderProgIndex = createShader(vertProgramText.c_str(), fragProgramText.c_str());
}

void printShaderInfoLog(GLuint shader);
void printProgramInfoLog(GLuint program);

/**
 * Create a shader program with specified code.
 *
 * @param vertShaderCode Vertex shader source code.
 * @param fragShaderCode Fragment shader source code.
 * @return Compiled/linked OpenGL shader program object index.
 */
unsigned int Compositor::createShader(const char* vertShaderCode, const char* fragShaderCode)
{
    // if GLSL is available...



    // create shader objects
#if defined(OPENGL2)
    GLuint vertShaderIndex = glCreateShader(GL_VERTEX_SHADER);
    GLuint fragShaderIndex = glCreateShader(GL_FRAGMENT_SHADER);
#endif // OPENGL2
#if defined(OPENGL15)
    GLuint vertShaderIndex = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    GLuint fragShaderIndex = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
#endif // OPENGL15

    // load shader programs
    int vertLength, fragLength;
    vertLength = strlen(vertShaderCode);
    fragLength = strlen(fragShaderCode);
#if defined(OPENGL2)
    glShaderSource(vertShaderIndex, 1, &vertShaderCode, &vertLength);
    glShaderSource(fragShaderIndex, 1, &fragShaderCode, &fragLength);
#endif // OPENGL2
#if defined(OPENGL15)
    glShaderSourceARB(vertShaderIndex, 1, &vertShaderCode, &vertLength);
    glShaderSourceARB(fragShaderIndex, 1, &fragShaderCode, &fragLength);
#endif // OPENGL15

    int compiled;

    // compile vertex shader
#if defined(OPENGL2)
    glCompileShader(vertShaderIndex);
    glGetShaderiv(vertShaderIndex, GL_COMPILE_STATUS, &compiled);
#endif // OPENGL2
#if defined(OPENGL15)
    glCompileShaderARB(vertShaderIndex);
    glGetObjectParameterivARB(vertShaderIndex, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
#endif // OPENGL15
    printShaderInfoLog(vertShaderIndex);
    if (compiled == GL_FALSE) {
        fprintf(stderr, "error : Compile error in vertex shader.\n");
        return 0;
    }

    // compile fragment shader
#if defined(OPENGL2)
    glCompileShader(fragShaderIndex);
    glGetShaderiv(fragShaderIndex, GL_COMPILE_STATUS, &compiled);
#endif // OPENGL2
#if defined(OPENGL15)
    glCompileShaderARB(fragShaderIndex);
    glGetObjectParameterivARB(fragShaderIndex, GL_OBJECT_COMPILE_STATUS_ARB, &compiled);
#endif // OPENGL15
    printShaderInfoLog(fragShaderIndex);
    if (compiled == GL_FALSE) {
        fprintf(stderr, "error : Compile error in fragment shader.\n");
        return 0;
    }

#if defined(OPENGL2)
    // create shader program object
    GLuint glslProgIndex = glCreateProgram();

    // atatch shaders to the shader program object
    glAttachShader(glslProgIndex, vertShaderIndex);
    glAttachShader(glslProgIndex, fragShaderIndex);

    // delete shaders
    glDeleteShader(vertShaderIndex);
    glDeleteShader(fragShaderIndex);
#endif // OPENGL2
#if defined(OPENGL15)
    // create shader program object
    GLuint glslProgIndex = glCreateProgramObjectARB();

    // atatch shaders to the shader program object
    glAttachObjectARB(glslProgIndex, vertShaderIndex);
    glAttachObjectARB(glslProgIndex, fragShaderIndex);

    // delete shaders
    glDeleteObjectARB(vertShaderIndex);
    glDeleteObjectARB(fragShaderIndex);
#endif // OPENGL15

#if defined(OPENGL2)
    // link shader program object
    int linked;
    glLinkProgram(glslProgIndex);
    glGetProgramiv(glslProgIndex, GL_LINK_STATUS, &linked);
#endif // OPENGL2
#if defined(OPENGL15)
    // link shader program object
    int linked;
    glLinkProgramARB(glslProgIndex);
    glGetObjectParameterivARB(glslProgIndex, GL_OBJECT_LINK_STATUS_ARB, &linked);
#endif // OPENGL15
    printProgramInfoLog(glslProgIndex);
    if (linked == GL_FALSE) {
        fprintf(stderr, "error : GLSL link error.\n");
        return 0;
    }

#if defined(OPENGL2)
    glUseProgram(0);
#endif // OPENGL2
#if defined(OPENGL15)
    glUseProgramObjectARB(0);
#endif // OPENGL15

    return glslProgIndex;
}

/**
 * Outpur shader compile log.
 *
 * @param shader Shader object index.
 */
static void printShaderInfoLog(GLuint shader)
{
    GLsizei bufSize;

#if defined(OPENGL2)
    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &bufSize);
#endif // OPENGL2
#if defined(OPENGL15)
    glGetObjectParameterivARB(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &bufSize);
#endif // OPENGL15

    if (bufSize > 1)
    {
        GLchar* pInfoLog;

        pInfoLog = new GLchar[bufSize];
        GLsizei length;

#if defined(OPENGL2)
        glGetShaderInfoLog(shader, bufSize, &length, pInfoLog);
#endif // OPENGL2
#if defined(OPENGL15)
        glGetInfoLogARB(shader, bufSize, &length, pInfoLog);
#endif // OPENGL15
        fprintf(stderr, "InfoLog:\n%s\n\n", pInfoLog);
        delete [] pInfoLog;
    }
}

/**
 * Outpur shader program compile log.
 *
 * @param program Shader program object index.
 */
void printProgramInfoLog(GLuint program)
{
    GLsizei bufSize;

#if defined(OPENGL2)
    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &bufSize);
#endif // OPENGL2
#if defined(OPENGL15)
    glGetObjectParameterivARB(program, GL_OBJECT_INFO_LOG_LENGTH_ARB, &bufSize);
#endif // OPENGL15

    if (bufSize > 1)
    {
        GLchar *pInfoLog = new GLchar[bufSize];
        GLsizei length;

#if defined(OPENGL2)
        glGetProgramInfoLog(program, bufSize, &length, pInfoLog);
#endif // OPENGL2
#if defined(OPENGL15)
        glGetInfoLogARB(program, bufSize, &length, pInfoLog);
#endif // OPENGL15
        fprintf(stderr, "InfoLog:\n%s\n\n", pInfoLog);
        delete [] pInfoLog;
    }
}
