#ifndef _GLRC_COMPOSITOR_H_
#define _GLRC_COMPOSITOR_H_

#include "glrc.h"
#include "Matrix.h"

#include <string>

namespace glrc
{

class RTTBuffer;
class StatsRenderer;

/**
 * Rendering Compositor object class.
 *
 * glrc::Compositor object has its own off-screen buffer (RTTBuffer).
 * When you render your scene into the off-screen context, you can render
 * the resut to on-screen (or other Compositors) with blending, rotation,
 * scaling, distortion, color conversion, hdr processing and so on.
 * You can also use this Compositor to take a snapshot larger than your screen.
 */
class GLRC_API Compositor
{
public:

    /**
     * Constructor. You can create an instance without valid OpenGL contexts.
     */
    Compositor();

    /**
     * Destructor.
     */
    virtual ~Compositor();

	/**
	 * Initialize the off-screen buffer. You must initialize the buffer in a valid
	 * OpenGL context before rendering.
	 * Floating point buffer is not available when rttType is READPIXEL nor COPYSUBTEX.
     * At this time, width and height must be power of two!!
     * You can call setup() more than once to re-configure the off-screen buffer.
	 *
	 * @param rttType RenderToTexture type. GLRC_RTT_AUTO to select the best type automatically.
	 * @param width Width of the off-screen buffer.
	 * @param height Height of the off-screen buffer.
	 * @param bHasAlpha True if you need alpha channel.
	 * @param bFPBuffer True if you need a floating point buffer.
	 */
    void setup(GLRC_RTT type, int width, int height, bool bHasAlpha=true, bool bFPBuffer=false);

	/**
	 * Retrieve RenderToTexture type of the off-screen buffer.
	 *
	 * @return RenderToTexture type of the off-screen buffer.
	 */
    GLRC_RTT getBufferType() const;

	/**
	 * Retrieve width of the off-screen buffer.
	 *
	 * @return Width of the off-screen buffer.
	 */
    int getBufferWidth() const;

	/**
	 * Retrieve height of the off-screen buffer.
	 *
	 * @return Height of the off-screen buffer.
	 */
    int getBufferHeight() const;

	/**
	 * Check whether the off-screen has alpha channel or not.
	 *
	 * @return True if the off-screen buffer has alpha channel.
	 */
    bool hasAlpha() const;

	/**
	 * Check whether the off-screen is floating buffer type or not.
	 *
	 * @return True if the off-screen buffer is a floating point buffer.
	 */
    bool isFPBuffer() const;

    /**
     * Enable mesh distortion or not (use shader for distiortion).
     *
     * @param bEnable True to use mesh distortion, false to use shader for distortion.
     */
    void setMeshDistortion(bool bEnable);

    /**
     * Check whether mesh distortion is enable or not.
     *
     * @return True to use mesh distortion, false to use shader for distortion.
     */
    bool isMeshDistortion() const;

    /**
     * Set resolution of the distortion mesh.
     *
     * @param horResolution Horizontal resolution of te distortion mesh.
     * @param vertResolution Vertical resolution of te distortion mesh.
     */
    void setDistortionMeshResolution(int horResolution, int vertResolution);

    /**
     * Retrieve horizontal resolution of the distortion mesh.
     *
     * @return Horizontal resolution of te distortion mesh.
     */
    int  getHorizontalResolutionOfDisotrtionMesh() const;

    /**
     * Retrieve vertical resolution of the distortion mesh.
     *
     * @return Vertical resolution of te distortion mesh.
     */
    int  getVerticalResolutionOfDisotrtionMesh() const;

    /**
     * Set rendering area of the compositor in current OpenGL viewport.
     * Position and size is defined in normalized (0.0-1.0) coordinate
     * in the current OpenGL viewport.
     *
     * @param x X (left) position of the compositor in normalized (0.0-1.0) coordinate.
     * @param y Y (top) position of the compositor in normalized (0.0-1.0) coordinate.
     * @param width Width of the compositor in normalized (0.0-1.0) coordinate.
     * @param height Height of the compositor in normalized (0.0-1.0) coordinate.
     */
	void setViewport(float x, float y, float width, float height);

    /**
     * Retrieve x (left) position of the compositor.
     *
     * @param X (left) position of the compositor in normalized (0.0-1.0) coordinate.
     */
    float getX() const;

    /**
     * Retrieve y (left) position of the compositor.
     *
     * @param Y (left) position of the compositor in normalized (0.0-1.0) coordinate.
     */
    float getY() const;

    /**
     * Retrieve width of the compositor.
     *
     * @param Width of the compositor in normalized (0.0-1.0) coordinate.
     */
    float getWidth() const;

    /**
     * Retrieve height of the compositor.
     *
     * @param Height of the compositor in normalized (0.0-1.0) coordinate.
     */
    float getHeight() const;

    /**
     * Enable or disable off-screen rendering (and composite processing). 
	 * When disabled, both beginRender() / endRender() do nothing, so you can get usual rendering results.
     *
     * @param bEnabled True to enable off-screen rendering (and composite processing).
     */
	void setEnabled(bool bEnabled);

    /**
     * Check whether off-screen rendering (and composite processing) is enabled or not.
     *
     * @return True if off-screen rendering (and composite processing) is enabled.
     */
    bool getEnabled() const;

    /**
     * Declare to begin to render your scene. Activate off-screen rendering to capture the rendering result,
     * and modify current ModelView / Projection matrices.
     */
    void beginRender();

    /**
     * Declare to end the rendering. Deactivate off-screen rendering,
     * then render the compositor object to on-screen (or other Compositor)
     * with composite processings.
     */
    void endRender();

    /**
     * Enable or disable compositor rendering with OpenGL blending.
     * When blending is enabled and cleared with proper alpha value ( < 1.0),
     * you can overlay the rendering results with transparent background.
     *
     * @param bEnabled True to composite with blending.
     */
    void setBlendEnabled(bool bEnabled);

    /**
     * Check whether compositiong with blending is enabled or not.
     *
     * @return True if composition with blending is enabled.
     */
    bool getBlendEnabled() const;

    /**
     * Set rotation angle of the compositor.
     *
     * @param angle Rotation angle in degree (0.0 is default).
     */
    void setRotation(float angle);

    /**
     * Retrieve rotation angle of the compositor.
     *
     * @return Rotation angle in degree (0.0 is default).
     */
    float getRotation() const;

    /**
     * Set scale of the compositor.
     *
     * @param x Horizontal scale of the compositor (1.0 is default).
     * @param y Vertical scale of the compositor (1.0 is default).
     */
    void setScale(float x, float y);

    /**
     * Retrieve horizontal scale of the compositor.
     *
     * @return Horizontal scale of the compositor (1.0 is default).
     */
    float getScaleX() const;

    /**
     * Retrieve vertical scale of the compositor.
     *
     * @return Vertical scale of the compositor (1.0 is default).
     */
    float getScaleY() const;

    /**
     * Set flipping mode of the compositor.
     *
     * @param flip Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
     */
    void setFlip(unsigned int flip);

    /**
     * Retrieve flipping mode of the compositor.
     *
     * @return Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
     */
    unsigned int getFlip() const;

    /**
     * Enable or disable model-view matrix modification.
     *
     * @param bEnabled True to enable model-view matrix modification.
     */
    void setModelViewMatrixEnabled(bool bEnabled);

    /**
     * Check whether model-view matrix modification is enabled or not.
     *
     * @return True if model-view matrix modification is enabled.
     */
    bool getModelViewMatrixEnabled() const;

    /**
     * Set model-view modification matrix.
     *
     * @param matrix Model-view modification matrix.
     */
    void setModelViewMatrix(const Matrix& matrix);

    /**
     * Retrieve model-view modification matrix.
     *
     * @return Model-view modification matrix.
     */
    Matrix getModelViewMatrix() const;

    /**
     * Retrieve translation of the model-view modification matrix.
     *
     * @param x Distance in x-axis direction.
     * @param y Distance in y-axis direction.
     * @param z Distance in z-axis direction.
     */
    void getModelViewTranslation(float& x, float& y, float& z) const;

    /**
     * Retrieve rotation of the model-view modification matrix.
     *
     * @param angle Angle of rotation in degree.
     * @param x X element value of the axis vector.
     * @param y Y element value of the axis vector.
     * @param z Z element value of the axis vector.
     */
    void getModelViewRotation(float& angle, float& x, float& y, float& z) const;

    /**
     * Retrieve quatenion rotation of the model-view modification matrix.
     *
     * @param x X element of quatenion.
     * @param y Y element of quatenion.
     * @param z Z element of quatenion.
     * @param w W element of quatenion.
     */
    void getModelViewQuatenion(float& x, float&y, float& z, float& w) const;

    /**
     * Retrieve scaling of the model-view modification matrix.
     *
     * @param x Scaling factor in x-axis.
     * @param y Scaling factor in y-axis.
     * @param z Scaling factor in z-axis.
     */
    void getModelViewScaling(float& x, float& y, float& z) const;

    /**
     * Enable or disable projection matrix modification.
     *
     * @param bEnabled True to enable projection matrix modification.
     */
    void setProjectionMatrixEnabled(bool bEnabled);

    /**
     * Check whether projection matrix modification is enabled or not.
     *
     * @return True if projection matrix modification is enabled.
     */
    bool getProjectionMatrixEnabled() const;

    /**
     * Set projection modification matrix.
     *
     * @param matrix Projection modification matrix.
     */
    void setProjectionMatrix(const Matrix& matrix);

    /**
     * Retrieve projection modification matrix.
     *
     * @param matrix Projection modification matrix.
     */
	Matrix getProjectionMatrix() const;

    /**
     * Set projection modification matrix in glFrustum format.
     *
     * @param frustum Projection modification matrix in glFrustum format.
     */
    void setProjectionFrustum(float frustum[6]);

    /**
     * Retrieve projection modification matrix in glFrustum format.
     *
     * @param frustum Projection modification matrix in glFrustum format.
     */
    void getProjectionFrustum(float frustum[6]) const;

    /**
     * Load matrix configuration file.
     *
     * @param fileName Matrix configuration file name.
     * @return True if loaded successfully.
     */
    bool loadMatrixFile(const char* fileName);

    /**
     * Set effect processing mode to composite the rendering results.
     *
     * @param effectMode Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
     */
    void setEffectMode(unsigned int effectMode);

    /**
     * Retrieve effect processing mode to composite the rendering results.
     *
     * @return Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
     */
    unsigned int getEffectMode() const;

    /**
     * Load a distortion map file to distort the rendering results.
     *
     * @param fileName Distortion map file name.
     * @return True if the distortion map is loaded successfully.
     */
	bool setDistortionMap(const char* fileName);

    /**
     * Retrieve the distortion map file name currently loaded.
     *
     * @return Distortion map file name currently loaded.
     */
	const char* getDistortionMapFileName() const;

    /**
     * Load a blend map file to blend the rendering results.
     *
     * @param fileName Blend map file name.
     * @return True if the blend map is loaded successfully.
     */
	bool setBlendMap(const char* fileName);

    /**
     * Retrieve the blend map file name currently loaded.
     *
     * @return Blend map file name currently loaded.
     */
	const char* getBlendMapFileName() const;

    /**
     * Set gamma value for tone mapping of high-dynamic range rendering.
     * LDR = (HDR * exposure) ^ gamma.
     *
     * @param Gamma value (1.0 is default).
     */
    void setGamma(float gamma);

    /**
     * Retrieve gamma value for high-dynamic range rendering.
     *
     * @return Gamma value (1.0 is default).
     */
    float getGamma() const;

    /**
     * Set exposure value for tone mapping of high-dynamic range rendering.
     * LDR = (HDR * exposure) ^ gamma.
     *
     * @param Exposure value (1.0 is default).
     */
    void setExposure(float exposure);

    /**
     * Retrieve exposure value for high-dynamic range rendering.
     *
     * @return Exposure value (1.0 is default).
     */
    float getExposure() const;

    /**
     * Set color conversion matrix.
     *
     * @param matrix Color conversion matrix.
     */
	void setColorMatrix(const Matrix& matrix);

    /**
     * Retrieve color conversion matrix.
     *
     * @result Color conversion matrix.
     */
    Matrix getColorMatrix() const;

    /**
     * Save the current off-screen buffer to an image file.
     * Image file format is determined by its file extention.
     * Supported format is depending on DevIL library (and .hdr format).
     *
     * @param Image file name to save the off-screen buffer.
     * @return True if image is saved successfully.
     */
    bool saveBuffer(const char* fileName);

protected:

    /**
     * (Re-)build OpenGL shader depends on current effect mode. Called by setEffectMode().
     */
    void buildShader();

    /**
     * Create a shader program with specified code.
     *
     * @param vertShaderCode Vertex shader source code.
     * @param fragShaderCode Fragment shader source code.
     * @return Compiled/linked OpenGL shader program object index.
     */
    unsigned int createShader(const char* vertShaderCode, const char* fragShaderCode);

protected:

    RTTBuffer* m_pRTTBuffer;    //<! Off-screen buffer object.
    bool m_bEnabled;            //<! True if off-screen rendering (and composite processing) is enabled.

	float m_x;          //<! X (left) position of the compositor in normalized (0.0-1.0) coordinate.
	float m_y;          //<! Y (top) position of the compositor in normalized (0.0-1.0) coordinate.
	float m_width;      //<! Width of the compositor in normalized (0.0-1.0) coordinate.
	float m_height;     //<! Height of the compositor in normalized (0.0-1.0) coordinate.

    bool m_bBlendEnabled;               //<! True if compositiong with blending is enabled.
    float m_rotationAngle;              //<! Rotation angle in degree (0.0 is default).
    float m_scaleX;                     //<! Horizontal scale of the compositor (1.0 is default).
    float m_scaleY;                     //<! Vertical scale of the compositor (1.0 is default).
    unsigned int m_flip;                //<! Flipping mode of the compositor (GLRC_FLIP_*).
    bool m_bModelViewMatrixEnabled;     //<! True to enable model-view matrix modification.
	Matrix m_modelViewMatrix;           //<! Model-view modification matrix.
    bool m_bProjectionMatrixEnabled;    //<! True to enable projection matrix modification.
	Matrix m_projectionMatrix;          //<! Projection modification matrix.
    float m_frustum[6];                 //<! Projection modification matrix in glFrustum format.

    unsigned int m_effectMode;          //<! Effect mode (GLRC_EFFECT_*).
    float m_gamma[4];                   //<! Gamma value (1.0 is default) for HDR rendering.
    float m_exposure[4];                //<! Exposure value (1.0 is default) for HDR rendering.
	Matrix m_colorMatrix;               //<! Color conversion matrix.

    unsigned int m_shaderProgIndex;     //<! OpenGL Shader program object index.

    bool m_bMeshDistortion;             //<! True to use mesh distortion, false to use shader for distortion.
    int m_distMeshHorResolution;        //<! Horizontal resolution of the distortion mesh.
    int m_distMeshVertResolution;       //<! vertical resolution of the distortion mesh.
    unsigned int m_distTexIndex;        //<! Texture object index of the distortion map.
    std::string m_distortionFileName;   //<! File name of the distortion map currently loaded.
    std::string m_blendFileName;        //<! File name of the blend map currently loaded.
    unsigned int m_distMeshIndex;       //<! Display list index of the distortion mesh.
    unsigned int m_blendTexIndex;       //<! Texture object index of the blend map.
};

}; // namespace glrc

#endif // _GLRC_COMPOSITOR_H_
