#include <math.h>

#include "Matrix.h"

using namespace glrc;


#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795f
#endif // M_PI
#define RAD(x) (((x)*2.0f*M_PI/360.0f)-2*M_PI*((int)((x)/360.0f))) // Deg -> Rad

// some part of the code from OpenProducer by Don Burns.
#define INNER_PRODUCT(a,b,r,c) \
     ((a).m_value[r][0] * (b).m_value[0][c]) \
    +((a).m_value[r][1] * (b).m_value[1][c]) \
    +((a).m_value[r][2] * (b).m_value[2][c]) \
    +((a).m_value[r][3] * (b).m_value[3][c])

/**
 * Constructor. Default value is the identity matrix.
 */
Matrix::Matrix()
{
    makeIdentity();
}

/**
 * Constructor with initial value.
 *
 * @param matrix Each matrix element values.
 */
Matrix::Matrix(float matrix[4][4])
{
    set(matrix);
}

/**
 * Set matrix value.
 *
 * @param matrix Each matrix element values.
 */
void Matrix::set(float matrix[4][4])
{
    for (unsigned int i=0; i<4; ++i)
        for (unsigned int j=0; j<4; ++j)
            m_value[i][j] = matrix[i][j];
}

/**
 * Retrieve matrix value.
 *
 * @param matrix Each matrix element values.
 */
void Matrix::get(float matrix[4][4]) const
{
    for (unsigned int i=0; i<4; ++i)
        for (unsigned int j=0; j<4; ++j)
            matrix[i][j] = m_value[i][j];
}

/**
 * Make an identity matrix.
 */
void Matrix::makeIdentity()
{
    m_value[0][0] = 1.0f; m_value[1][0] = 0.0f; m_value[2][0] = 0.0f; m_value[3][0] = 0.0f;
    m_value[0][1] = 0.0f; m_value[1][1] = 1.0f; m_value[2][1] = 0.0f; m_value[3][1] = 0.0f;
    m_value[0][2] = 0.0f; m_value[1][2] = 0.0f; m_value[2][2] = 1.0f; m_value[3][2] = 0.0f;
    m_value[0][3] = 0.0f; m_value[1][3] = 0.0f; m_value[2][3] = 0.0f; m_value[3][3] = 1.0f;
}

/**
 * Make a frustum projection matrix
 *
 * @param l The coordinates for the left-vertical clipping planes.
 * @param r The coordinates for the right-vertical clipping planes.
 * @param b The coordinates for the bottom-horizontal clipping planes.
 * @param t The coordinates for the top-horizontal clipping planes.
 * @param n The distances to the near-depth clipping planes. Both distances must be positive.
 * @param f The distances to the far-depth clipping planes. Both distances must be positive.
 */
void Matrix::makeFrustum(float l, float r, float b, float t, float n, float f)
{
    if (l == r || b == t || n == f)
        return;

    m_value[0][0] = 2.0f * n / (r - l);
    m_value[0][1] = 0.0f;
    m_value[0][2] = 0.0f;
    m_value[0][3] = 0.0f;
    m_value[1][0] = 0.0f;
    m_value[1][1] = 2.0f * n / (t - b);
    m_value[1][2] = 0.0f;
    m_value[1][3] = 0.0f;
    m_value[2][0] = (r + l) / (r - l);
    m_value[2][1] = (t + b) / (t - b);
    m_value[2][2] = -(f + n) / (f - n);
    m_value[2][3] = -1.0f;
    m_value[3][0] = 0.0f;
    m_value[3][1] = 0.0f;
    m_value[3][2] = -2.0f * f * n / (f - n);
    m_value[3][3] = 0.0f;
}

/**
 * Translate the current matrix.
 *
 * @param x Distance in x-axis direction.
 * @param y Distance in y-axis direction.
 * @param z Distance in z-axis direction.
 */
void Matrix::translate(float x, float y, float z)
{
    float m[4][4];
    m[0][0] = 1.0f; m[1][0] = 0.0f; m[2][0] = 0.0f; m[3][0] =    x;
    m[0][1] = 0.0f; m[1][1] = 1.0f; m[2][1] = 0.0f; m[3][1] =    y;
    m[0][2] = 0.0f; m[1][2] = 0.0f; m[2][2] = 1.0f; m[3][2] =    z;
    m[0][3] = 0.0f; m[1][3] = 0.0f; m[2][3] = 0.0f; m[3][3] = 1.0f;

    Matrix mat(m);
    preMult(mat);
}

/**
 * Rotate the current matrix.
 *
 * @param angle Angle of rotation in degree.
 * @param x X element value of the axis vector.
 * @param y Y element value of the axis vector.
 * @param z Z element value of the axis vector.
 */
void Matrix::rotate(float angle, float x, float y, float z)
{
    float m[4][4];

    float inversenorm  = 1.0f / (float)sqrt(x*x+y*y+z*z);
    float coshalfangle = (float)cos(-0.5f*RAD(angle));
    float sinhalfangle = (float)sin(-0.5f*RAD(angle));

    float fv[4];
    fv[0] = x * sinhalfangle * inversenorm;
    fv[1] = y * sinhalfangle * inversenorm;
    fv[2] = z * sinhalfangle * inversenorm;
    fv[3] = coshalfangle;

    // Source: Gamasutra, Rotating Objects Using Quaternions
    //
    //http://www.gamasutra.com/features/programming/19980703/quaternions_01.htm
#define QX  fv[0]
#define QY  fv[1]
#define QZ  fv[2]
#define QW  fv[3]

    float wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;

    // calculate coefficients
    x2 = QX + QX;
    y2 = QY + QY;
    z2 = QZ + QZ;

    xx = QX * x2;
    xy = QX * y2;
    xz = QX * z2;

    yy = QY * y2;
    yz = QY * z2;
    zz = QZ * z2;

    wx = QW * x2;
    wy = QW * y2;
    wz = QW * z2;

    m[0][0] = 1.0f - (yy + zz);
    m[0][1] = xy - wz;
    m[0][2] = xz + wy;
    m[0][3] = 0.0f;

    m[1][0] = xy + wz;
    m[1][1] = 1.0f - (xx + zz);
    m[1][2] = yz - wx;
    m[1][3] = 0.0f;

    m[2][0] = xz - wy;
    m[2][1] = yz + wx;
    m[2][2] = 1.0f - (xx + yy);
    m[2][3] = 0.0f;

    m[3][0] = 0.0f;
    m[3][1] = 0.0f;
    m[3][2] = 0.0f;
    m[3][3] = 1.0f;

#undef QX 
#undef QY 
#undef QZ 
#undef QW 

    Matrix mat(m);
    preMult(mat);
}

/**
 * Scale the current matrix.
 *
 * @param x Scaling factor in x-axis.
 * @param y Scaling factor in y-axis.
 * @param z Scaling factor in z-axis.
 */
void Matrix::scale(float x, float y, float z)
{
    float m[4][4];
    m[0][0] =    x; m[1][0] = 0.0f; m[2][0] = 0.0f; m[3][0] = 0.0f;
    m[0][1] = 0.0f; m[1][1] =    y; m[2][1] = 0.0f; m[3][1] = 0.0f;
    m[0][2] = 0.0f; m[1][2] = 0.0f; m[2][2] =    z; m[3][2] = 0.0f;
    m[0][3] = 0.0f; m[1][3] = 0.0f; m[2][3] = 0.0f; m[3][3] = 1.0f;

    Matrix mat(m);
    preMult(mat);
}

/**
 * Pre multiply a matrix.
 *
 * @param matrix Matrix to multiply.
 */
void Matrix::preMult(const Matrix& matrix)
{
    float t[4];
    for (int col=0; col<4; ++col) {
        t[0] = INNER_PRODUCT(matrix, *this, 0, col);
        t[1] = INNER_PRODUCT(matrix, *this, 1, col);
        t[2] = INNER_PRODUCT(matrix, *this, 2, col);
        t[3] = INNER_PRODUCT(matrix, *this, 3, col);
        m_value[0][col] = t[0];
        m_value[1][col] = t[1];
        m_value[2][col] = t[2];
        m_value[3][col] = t[3];
    }
}   

/**
 * Post multiply a matrix.
 *
 * @param matrix Matrix to multiply.
 */
void Matrix::postMult(const Matrix& matrix)
{
    for (int row=0; row<4; ++row)
    {
        m_value[row][0] = INNER_PRODUCT( *this, matrix, row, 0);
        m_value[row][1] = INNER_PRODUCT( *this, matrix, row, 1);
        m_value[row][2] = INNER_PRODUCT( *this, matrix, row, 2);
        m_value[row][3] = INNER_PRODUCT( *this, matrix, row, 3);
    }
}

/**
 * Retrieve translation of the current matrix.
 * @param x Distance in x-axis direction.
 * @param y Distance in y-axis direction.
 * @param z Distance in z-axis direction.
 */
void Matrix::getTranslation(float& x, float&y, float& z) const
{
    x = m_value[3][0];
    y = m_value[3][1];
    z = m_value[3][2];
}

/**
 * Retrieve rotation of the current matrix.
 *
 * @param angle Angle of rotation in degree.
 * @param x X element value of the axis vector.
 * @param y Y element value of the axis vector.
 * @param z Z element value of the axis vector.
 */
void Matrix::getRotation(float& angle, float& x, float& y, float& z) const
{
    float qx, qy, qz, qw;
    getQuatenion(qx, qy, qz, qw);

    float sinhalfangle = (float)sqrt(qx*qx + qy*qy + qz*qz);
    angle = 2.0f * (float)atan2(sinhalfangle, qw);
    if (sinhalfangle)
    {
        x = qx / sinhalfangle;
        y = qy / sinhalfangle;
        z = qz / sinhalfangle;
    }
    else
    {
        x = 0.0f;
        y = 0.0f;
        z = 1.0f;
    }
}

/**
 * Retrieve translation of the matrix.
 *
 * @param x X element of quatenion.
 * @param y Y element of quatenion.
 * @param z Z element of quatenion.
 * @param w W element of quatenion.
 */
void Matrix::getQuatenion(float& x, float&y, float& z, float& w) const
{
    // Source: Gamasutra, Rotating Objects Using Quaternions
    //
    //http://www.gamasutra.com/features/programming/19980703/quaternions_01.htm

    float tr, s;
    float tq[4];
    int i, j, k;

    int nxt[3] = {1, 2, 0};

    tr = m_value[0][0] + m_value[1][1] + m_value[2][2] + 1.0f;

    // check the diagonal
    if (tr > 0.0f)
    {
        s = (float)sqrt(tr);
        w = s / 2.0f;
        s = 0.5f / s;
        x = (m_value[1][2] - m_value[2][1]) * s;
        y = (m_value[2][0] - m_value[0][2]) * s;
        z = (m_value[0][1] - m_value[1][0]) * s;
    }
    else
    {
        // diagonal is negative
        i = 0;
        if (m_value[1][1] > m_value[0][0])
            i = 1;
        if (m_value[2][2] > m_value[i][i])
            i = 2;
        j = nxt[i];
        k = nxt[j];

        s = (float)sqrt ((m_value[i][i] - (m_value[j][j] + m_value[k][k])) + 1.0f);

        tq[i] = s * 0.5f;

        if (s != 0.0f)
            s = 0.5f / s;

        tq[3] = (m_value[j][k] - m_value[k][j]) * s;
        tq[j] = (m_value[i][j] + m_value[j][i]) * s;
        tq[k] = (m_value[i][k] + m_value[k][i]) * s;

        x = tq[0];
        y = tq[1];
        x = tq[2];
        w = tq[3];
    }
}

/**
 * Retrieve scaling of the current matrix.
 *
 * @param x Scaling factor in x-axis.
 * @param y Scaling factor in y-axis.
 * @param z Scaling factor in z-axis.
 */
void Matrix::getScaling(float& x, float& y, float& z) const
{
    x = m_value[0][0];
    y = m_value[1][1];
    z = m_value[2][2];
}

/**
 * Assignment operator.
 *
 * @param matrix Assignment source transform matrix.
 */
Matrix& Matrix::operator=(const Matrix& matrix)
{
    for (unsigned int i=0; i<4; ++i)
        for (unsigned int j=0; j<4; ++j)
            m_value[i][j] = matrix.m_value[i][j];
    return *this;
}

 /**
 * Retrieve matrix value as float pointer (for glMultMatrixfv()).
 *
 * @return Matrix element values.
 */
float* Matrix::ptr() const
{
    return (float *)m_value;
}

