#ifndef _GLRC_MATRIX_H_
#define _GLRC_MATRIX_H_

#include "glrc.h"

namespace glrc
{

/**
 * Simple matrix class.
 */
class GLRC_API Matrix
{
public:

    /**
     * Constructor. Default value is the identity matrix.
     */
    Matrix();

    /**
     * Constructor with initial value.
     *
     * @param matrix Each matrix element values.
     */
    Matrix(float matrix[4][4]);

    /**
     * Set matrix value.
     *
     * @param matrix Each matrix element values.
     */
    void set(float matrix[4][4]);

    /**
     * Retrieve matrix value.
     *
     * @param matrix Each matrix element values.
     */
    void get(float matrix[4][4]) const;

    /**
     * Make an identity matrix.
     */
    void makeIdentity();

    /**
     * Make a frustum projection matrix
     *
     * @param l The coordinates for the left-vertical clipping planes.
     * @param r The coordinates for the right-vertical clipping planes.
     * @param b The coordinates for the bottom-horizontal clipping planes.
     * @param t The coordinates for the top-horizontal clipping planes.
     * @param n The distances to the near-depth clipping planes. Both distances must be positive.
     * @param f The distances to the far-depth clipping planes. Both distances must be positive.
     */
    void makeFrustum(float l, float r, float b, float t, float n, float f);

    /**
     * Translate the current matrix.
     *
     * @param x Distance in x-axis direction.
     * @param y Distance in y-axis direction.
     * @param z Distance in z-axis direction.
     */
    void translate(float x, float y, float z);


    /**
     * Rotate the current matrix.
     *
     * @param angle Angle of rotation in degree.
     * @param x X element value of the axis vector.
     * @param y Y element value of the axis vector.
     * @param z Z element value of the axis vector.
     */
    void rotate(float angle, float x, float y, float z);

    /**
     * Scale the current matrix.
     *
     * @param x Scaling factor in x-axis.
     * @param y Scaling factor in y-axis.
     * @param z Scaling factor in z-axis.
     */
    void scale(float x, float y, float z);

    /**
     * Pre multiply a matrix.
     *
     * @param matrix Matrix to multiply.
     */
    void preMult(const Matrix& matrix);

    /**
     * Post multiply a matrix.
     *
     * @param matrix Matrix to multiply.
     */
    void postMult(const Matrix& matrix);

    /**
     * Retrieve translation of the current matrix.
     * @param x Distance in x-axis direction.
     * @param y Distance in y-axis direction.
     * @param z Distance in z-axis direction.
     */
    void getTranslation(float& x, float&y, float& z) const;

    /**
     * Retrieve rotation of the current matrix.
     *
     * @param angle Angle of rotation in degree.
     * @param x X element value of the axis vector.
     * @param y Y element value of the axis vector.
     * @param z Z element value of the axis vector.
     */
    void getRotation(float& angle, float& x, float& y, float& z) const;

    /**
     * Retrieve translation of the matrix.
     *
     * @param x X element of quatenion.
     * @param y Y element of quatenion.
     * @param z Z element of quatenion.
     * @param w W element of quatenion.
     */
    void getQuatenion(float& x, float&y, float& z, float& w) const;

    /**
     * Retrieve scaling of the current matrix.
     *
     * @param x Scaling factor in x-axis.
     * @param y Scaling factor in y-axis.
     * @param z Scaling factor in z-axis.
     */
    void getScaling(float& x, float& y, float& z) const;

    /**
     * Assignment operator.
     *
     * @param matrix Assignment source transform matrix.
     */
    Matrix& operator=(const Matrix& matrix);

    /**
     * Retrieve matrix value as float pointer (for glMultMatrixfv()).
     *
     * @return Matrix element values.
     */
    float* ptr() const;

	// serialize
	// deserialize

private:

    float m_value[4][4];  //!< Matrix element values.
};

}; // namespace glrc

#endif // _GLRC_MATRIX_H_
