#ifndef _GLRC_PBUFFER_H_
#define _GLRC_PBUFFER_H_

#include "glrc.h"

#if defined(WIN32)
#  include <windows.h>
#  include "GLee.h"
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#  include <GL/glxext.h>
#elif defined(MACOS)
#  include <AGL/agl.h>
#endif

#include <string>
#include <vector>

namespace glrc
{

// The pixel format for the pbuffer is controlled by the mode string passed
// into the PBuffer constructor. This string can have the following attributes:
//
// r            - r pixel format (for float buffer).
// rg           - rg pixel format (for float buffer).
// rgb          - rgb pixel format. 8 bit or 16/32 bit in float buffer mode
// rgba         - same as "rgb alpha" string
// alpha        - must have alpha channel
// depth        - must have a depth buffer
// depth=n      - must have n-bit depth buffer
// stencil      - must have a stencil buffer
// double       - must support double buffered rendering
// samples=n    - must support n-sample antialiasing (n can be 2 or 4)
// float=n      - must support n-bit per channel floating point
//
// texture2D
// textureRECT
// textureCUBE  - must support binding pbuffer as texture to specified target
//              - binding the depth buffer is also supporting by specifying
//                '=depth' like so: texture2D=depth or textureRECT=depth
//              - the internal format of the texture will be rgba by default or
//                float if pbuffer is floating point
//

class PBuffer
{
public:

    PBuffer(const char* strMode, bool bManaged = false);
    virtual ~PBuffer();

    bool init(int width, int height, bool bShareContexts=false, bool bShareObjects=false);
    void destroy();

    void activate(PBuffer* pCurrent=NULL);
    void deactivate();

#if defined(WIN32)

    int  bind(int whichBuffer);
    int  release(int whichBuffer);
    int  swapBuffers();

    void handleModeSwitch();

#endif // WIN32

    unsigned int getSizeInBytes() { return m_width * m_height * (m_numComponents/8); }

    unsigned int copyToBuffer(void* pPtr, int width=-1, int height=-1);

    inline int  getNumComponents() { return m_numComponents; }
    inline int  getBitsPerComponent() { return m_bitsPerComponent; }
    inline int  getWidth() { return m_width; }
    inline int  getHeight() { return m_height; }
    inline bool isSharedContext() { return m_bSharedContext; }
#if defined(WIN32)
    inline bool isTexture() { return m_bIsTexture; }
#endif // WIN32

private:

    std::string getStringValue(std::string token);
    int getIntegerValue(std::string token);
#if defined(UNIX) || defined(WIN32)
    void parseModeString(const char* strMode, std::vector<int>* pfAttribList, std::vector<int>* pbAttribList);
#endif

protected:

#if defined(WIN32)

    HDC         m_hDC;
    HGLRC       m_hGLRC;
    HPBUFFERARB m_hPBuffer;

    HGLRC       m_hOldGLRC;
    HDC         m_hOldDC;

    bool m_bIsTexture;
    bool m_bIsBound;
    bool m_bIsActive;
	bool m_bIsDoubleBuffer;
    bool m_bManaged;

    std::vector<int> m_pfAttribList;
    std::vector<int> m_pbAttribList;

#elif defined(UNIX)

    Display    *m_pDisplay;
    GLXPbuffer  m_glxPbuffer;
    GLXContext  m_glxContext;

    Display    *m_pOldDisplay;
    GLXPbuffer  m_glxOldDrawable;
    GLXContext  m_glxOldContext;

    std::vector<int> m_pfAttribList;
    std::vector<int> m_pbAttribList;

    bool m_bIsBound;
    bool m_bIsActive;
	bool m_bIsDoubleBuffer;
    bool m_bManaged;

#elif defined(MACOS)

    AGLContext  m_context;
    WindowPtr   m_window;
    std::vector<int> m_pfAttribList;

#endif

    int m_width;
    int m_height;
    int m_numComponents;
    int m_bitsPerComponent;

    const char *m_strMode;
    bool m_bSharedContext;
    bool m_bShareObjects;
};

}; // namespace glrc

#endif // _GLRC_PBUFFER_H_
