#include "PBuffer.h"

#if defined(WIN32)
#  include <windows.h>
#  include "GLee.h"
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#  include <GL/glxext.h>
#elif defined(MACOS)
#  include <AGL/agl.h>
#endif

#ifdef _DEBUG
#define PB_FPF_NOTICE  printf
#define PB_FPF_WARNING printf
#define PB_FPF_ERROR   printf
#else
#define PB_FPF_NOTICE
#define PB_FPF_WARNING
#define PB_FPF_ERROR
#endif

using namespace glrc;

PBuffer::PBuffer(const char* strMode, bool bManaged)
{
    m_hDC = NULL;
    m_hGLRC = NULL;
    m_hPBuffer = NULL;
    m_hOldGLRC = NULL;
    m_hOldDC = NULL;
    m_bIsTexture = false;
    m_width = 0;
    m_height = 0;
    m_bSharedContext = false;
    m_bShareObjects = false;
    m_bIsBound = false;
    m_bIsActive = false;
	m_bIsDoubleBuffer = true;
    m_bManaged = bManaged;

    m_pfAttribList.push_back(WGL_DRAW_TO_PBUFFER_ARB);
    m_pfAttribList.push_back(true);
    m_pfAttribList.push_back(WGL_SUPPORT_OPENGL_ARB);
    m_pfAttribList.push_back(true);

    m_pbAttribList.push_back(WGL_PBUFFER_LARGEST_ARB);
    m_pbAttribList.push_back(true);

    PB_FPF_NOTICE("notice: declare a p-buffer with \"%s\" parameters.\n", strMode);
    m_strMode = strMode;
    parseModeString(m_strMode, &m_pfAttribList, &m_pbAttribList);

    m_pfAttribList.push_back(0);
    m_pbAttribList.push_back(0);
}

PBuffer::~PBuffer()
{
    if (m_bManaged)
        destroy();
}

bool PBuffer::init(int width, int height, bool bShareContexts, bool bShareObjects)
{
    HDC hdc = wglGetCurrentDC();
    HGLRC hglrc = wglGetCurrentContext();
    int format = 0;
    int nfattribs = 0;
    int niattribs = 0;

    m_width = width;
    m_height = height;

    m_bSharedContext = bShareContexts;
    m_bShareObjects = bShareObjects;

    if (m_bSharedContext)
    {
        // Get the pixel format for the on-screen window.
        format = GetPixelFormat(hdc);
        if (format == 0)
        {
            PB_FPF_ERROR("error : pbuffer creation error: GetPixelFormat() failed.\n");
            return false;
        }
    }
    else
    {
        unsigned int numFormats;
        wglChoosePixelFormatARB(hdc, &m_pfAttribList[0], NULL, 1, &format, &numFormats);
        if (numFormats == 0)
        {
            PB_FPF_ERROR("error : pbuffer creation error: Couldn't find a suitable pixel format.\n");
            return false;
        }
    }

    m_hPBuffer = wglCreatePbufferARB(hdc, format, m_width, m_height, &m_pbAttribList[0]);
    if (!m_hPBuffer)
    {
        DWORD err = GetLastError();
        PB_FPF_ERROR("error : pbuffer creation error: wglCreatePbufferARB() failed.\n");
        if (err == ERROR_INVALID_PIXEL_FORMAT)
        {
            PB_FPF_ERROR("error : ERROR_INVALID_PIXEL_FORMAT\n");
        }
        else if (err == ERROR_NO_SYSTEM_RESOURCES)
        {
            PB_FPF_ERROR("error : ERROR_NO_SYSTEM_RESOURCES\n");
        }
        else if (err == ERROR_INVALID_DATA)
        {
            PB_FPF_ERROR("error : ERROR_INVALID_DATA\n");
        }
        
        return false;
    }

    // Get the device context.
    m_hDC = wglGetPbufferDCARB(m_hPBuffer);
    if (!m_hDC)
    {
        PB_FPF_ERROR("error : pbuffer creation error: wglGetPbufferDCARB() failed.\n");
        return false;
    }

    if (m_bSharedContext)
    {
        // Let's use the same gl context..
        // Since the device contexts are compatible (i.e. same pixelformat),
        // we should be able to use the same gl rendering context.
        m_hGLRC = hglrc;
    }
    else
    {
        // Create a new gl context for the p-buffer.
        m_hGLRC = wglCreateContext(m_hDC);
        if (!m_hGLRC)
        {
            PB_FPF_ERROR("error : pbuffer creation error: wglCreateContext() failed.\n");
            return false;
        }

        if(m_bShareObjects)
        {
            if(!wglShareLists(hglrc, m_hGLRC))
            {
                PB_FPF_ERROR("error : pbuffer: wglShareLists() failed.\n");
                return false;
            }
        }
    }

    GLint texFormat = WGL_NO_TEXTURE_ARB;
    wglQueryPbufferARB(m_hPBuffer, WGL_TEXTURE_FORMAT_ARB, &texFormat);

//ATI??    if (texFormat != WGL_NO_TEXTURE_ARB)
        m_bIsTexture = true;

    // Determine the actual width and height we were able to create.
    wglQueryPbufferARB(m_hPBuffer, WGL_PBUFFER_WIDTH_ARB, &m_width);
    wglQueryPbufferARB(m_hPBuffer, WGL_PBUFFER_HEIGHT_ARB, &m_height);

    PB_FPF_NOTICE("notice : Created a %d x %d pbuffer.\n", m_width, m_height);

#ifdef _DEBUG
    // query pixel format
    int attributes[] = {
        WGL_RED_BITS_ARB,
        WGL_GREEN_BITS_ARB,
        WGL_BLUE_BITS_ARB,
        WGL_ALPHA_BITS_ARB,
        WGL_FLOAT_COMPONENTS_NV,
        WGL_DEPTH_BITS_ARB,
        WGL_SAMPLES_EXT,
        WGL_AUX_BUFFERS_ARB
    };
    int values[sizeof(attributes) / sizeof(int)];

    if (wglGetPixelFormatAttribivARB(m_hDC, format, 0, sizeof(attributes) / sizeof(int), attributes, values)) {
        PB_FPF_NOTICE("r:%d g:%d b:%d a:%d float:%d depth:%d samples:%d aux:%d.\n",
            values[0], values[1], values[2], values[3], values[4], values[5], values[6], values[7]);
    }
#endif

    return true;
}

void PBuffer::destroy()
{
    if (m_hPBuffer)
    {
        if (!m_bSharedContext)
            wglDeleteContext(m_hGLRC);
        wglReleasePbufferDCARB(m_hPBuffer, m_hDC);
        wglDestroyPbufferARB(m_hPBuffer);
        m_hGLRC = NULL;
        m_hDC = NULL;
        m_hPBuffer = NULL;
    }
}

void PBuffer::activate(PBuffer* pCurrent)
{
    if (pCurrent == this)
    {
        return;
    }

    if (NULL == pCurrent || !pCurrent->m_bIsActive)
    {
        if (m_bIsActive)
            return;

        m_hOldGLRC = wglGetCurrentContext();
        m_hOldDC = wglGetCurrentDC();
    }
    else
    {
        m_hOldGLRC = pCurrent->m_hOldGLRC;
        m_hOldDC = pCurrent->m_hOldDC;
        pCurrent->m_hOldGLRC = 0;
        pCurrent->m_hOldDC = 0;
        pCurrent->m_bIsActive = false;
    }

    if (!wglMakeCurrent(m_hDC, m_hGLRC))
        PB_FPF_ERROR("error : PBuffer::activate() failed.\n");

    m_bIsActive = true;
}

void PBuffer::deactivate()
{
    if (!m_bIsActive)
        return;

    if (!wglMakeCurrent(m_hOldDC, m_hOldGLRC))
        PB_FPF_ERROR("error : PBuffer::deactivate() failed.\n");

    m_hOldGLRC = 0;
    m_hOldDC = 0;
    m_bIsActive = false;
}

int PBuffer::bind(int whichBuffer)
{
    if (!m_bIsTexture)
    {
        PB_FPF_ERROR("error : PBuffer::bind() failed - pbuffer format does not support render to texture.\n");
        return 0;
    }

#if 0
    // SGG - with MRT it is legal to bind different buffers of a pbuffer simultaneously
    if (m_bIsBound)
    {
        PB_FPF_ERROR("error : PBuffer::bind() failed - pbuffer is already bound.\n");
        return 0;
    }
#endif

    if (m_bIsBound)
		return 1;

    int result = wglBindTexImageARB(m_hPBuffer, whichBuffer);
    if (!result)
        PB_FPF_ERROR("error : PBuffer::bind() failed.\n");

    m_bIsBound = true;

    return result;
}

int PBuffer::release(int whichBuffer)
{
    if (!m_bIsTexture)
    {
        PB_FPF_ERROR("error : PBuffer::release() failed - pbuffer format does not support render to texture.\n");
        return 0;
    }

#if 0
    // SGG - with MRT it is legal to bind different buffers of a pbuffer simultaneously
    if (!m_bIsBound)
    {
        PB_FBF_ERROR("error : PBuffer::release() failed - pbuffer is not bound.\n");
        return 0;
    }
#endif

    if (!m_bIsBound)
		return 1;

    int result = wglReleaseTexImageARB(m_hPBuffer, whichBuffer);
    if (!result)
        PB_FPF_ERROR("error : PBuffer::release() failed.\n");

    m_bIsBound = false;

    return result;
}

int PBuffer::swapBuffers()
{
	if (m_hDC && m_bIsDoubleBuffer)
	    return ::SwapBuffers(m_hDC);
	return 0;
}

void PBuffer::handleModeSwitch()
{
    int lost = 0;

    wglQueryPbufferARB(m_hPBuffer, WGL_PBUFFER_LOST_ARB, &lost);

    if (lost)
    {
        this->~PBuffer();
        init(m_width, m_height, m_bSharedContext, m_bShareObjects);
    }
}

unsigned int PBuffer::copyToBuffer(void* pPtr, int width, int height)
{
    GLenum format;
    GLenum type;
    switch (m_numComponents)
    {
    case 1: //
        format = GL_LUMINANCE; // is it right to ask for Red only component ?
        break;
    case 2:
        format = GL_LUMINANCE_ALPHA; //How to ask for GL_RG ??
        break;
    case 3:
        format = GL_RGB;
        break;
    case 4:
        format = GL_RGBA;
        break;
    }
    switch(m_bitsPerComponent)
    {
    case 8:
        type = GL_UNSIGNED_BYTE;
        break;
    case 32:
        type = GL_FLOAT;
        break;
    #ifdef GL_NV_half_float
    case 16:
        type = GL_HALF_FLOAT_NV;
        break;
    #endif
    default:
        PB_FPF_ERROR("error : unknown m_bitsPerComponent.\n");
        #   if defined(WIN32)
        _asm { int 3 }
        #   endif
    }

    activate();

    if ((width < 0) || (width > m_width))
        width = m_width;
    if ((height < 0) || (height > m_height))
        height = m_height;
    glReadPixels(0, 0, width, height, format, type, pPtr);

    deactivate();

    return width * height * (m_numComponents/8);
}

void PBuffer::parseModeString(const char* strMode, std::vector<int>* pfAttribList, std::vector<int>* pbAttribList)
{
    if (!strMode || strcmp(strMode, "") == 0)
        return;

    m_bitsPerComponent = 8;
    m_numComponents = 0;
    bool bIsFloatBuffer = false;
    bool bIsATIFloatBuffer = false;
    bool bIsTexture = false;
    bool bNeedAlpha = false;

    char* mode = strdup(strMode);

	std::vector<std::string> tokens;
    char* buf = strtok(mode, " ");
    while (buf != NULL)
    {
        if (strstr(buf, "ati_float") != NULL)
            bIsATIFloatBuffer = true;
        else if (strstr(buf, "float") != NULL)
            bIsFloatBuffer = true;

        if (strstr(buf, "texture") != NULL)
            bIsTexture = true;

        if (strstr(buf, "alpha") != NULL)
            bNeedAlpha = true;

        tokens.push_back(buf);
        buf = strtok(NULL, " ");
    }

    pfAttribList->push_back(WGL_PIXEL_TYPE_ARB);
#ifdef WGL_ATI_pixel_format_float
    if (bIsATIFloatBuffer) {
        pfAttribList->push_back(WGL_TYPE_RGBA_FLOAT_ATI);
    } else
#endif
    {
        pfAttribList->push_back(WGL_TYPE_RGBA_ARB);
    }

    for (unsigned int i=0; i<tokens.size(); i++)
    {
        std::string token = tokens[i];

        if (token == "rgb" && (m_numComponents <= 1))
        {
            m_numComponents += 3;
            continue;
        }
        else if (token == "rgb")
            PB_FPF_WARNING("warning : mistake in components definition (rgb + %d).\n", m_numComponents);

        if (token == "rgba" && (m_numComponents == 0))
        {
            m_numComponents = 4;
            continue;
        }
        else if (token == "rgba")
            PB_FPF_WARNING("warning : mistake in components definition (rgba + %d).\n", m_numComponents);

        if (token == "alpha" && (m_numComponents <= 3))
        {
            m_numComponents++;
            continue;
        }
        else if (token == "alpha")
            PB_FPF_WARNING("warning : mistake in components definition (alpha + %d).\n", m_numComponents);

        if (token == "r" && (m_numComponents <= 1))// && bIsFloatBuffer)
        {
            m_numComponents++;
            continue;
        }
        else if (token == "r")
            PB_FPF_WARNING("warning : mistake in components definition (r + %d)\n", m_numComponents);

        if (token == "rg" && (m_numComponents <= 1))// && bIsFloatBuffer)
        {
            m_numComponents += 2;
            continue;
        }
        else if (token == "r")
            PB_FPF_WARNING("warning : mistake in components definition (rg + %d)\n", m_numComponents);

        if (token.find("depth") == 0)
        {
            pfAttribList->push_back(WGL_DEPTH_BITS_ARB);
            pfAttribList->push_back(getIntegerValue(token));
            continue;
        }

        if (token.find("stencil") == 0)
        {
            pfAttribList->push_back(WGL_STENCIL_BITS_ARB);
            pfAttribList->push_back(8);
            continue;
        }

        if (token.find("samples") == 0)
        {
            pfAttribList->push_back(WGL_SAMPLE_BUFFERS_ARB);
            pfAttribList->push_back(1);
            pfAttribList->push_back(WGL_SAMPLES_ARB);
            pfAttribList->push_back(getIntegerValue(token));
            continue;
        }

        if (token.find("aux") == 0)
        {
            pfAttribList->push_back(WGL_AUX_BUFFERS_ARB);
            pfAttribList->push_back(getIntegerValue(token));
            continue;
        }

        if (token == "double")
        {
            pfAttribList->push_back(WGL_DOUBLE_BUFFER_ARB);
            pfAttribList->push_back(true);
			m_bIsDoubleBuffer = true;
            continue;
        }

        if (token.find("ati_float") == 0)
        {
            m_bitsPerComponent = getIntegerValue(token);
            // type already set above
            continue;

        }
        else if (token.find("float") == 0)
        {
            m_bitsPerComponent = getIntegerValue(token);
            // bIsFloatBuffer = true; done previously
            pfAttribList->push_back(WGL_FLOAT_COMPONENTS_NV);
            pfAttribList->push_back(true);
            continue;
        }

        if (token.find("texture") == 0)
        {
            if (token.find("textureRECT") == 0 || bIsFloatBuffer)
            {
                pbAttribList->push_back(WGL_TEXTURE_TARGET_ARB);
                pbAttribList->push_back(WGL_TEXTURE_RECTANGLE_NV);
            }
            else if (token.find("textureCUBE") == 0)
            {
                pbAttribList->push_back(WGL_TEXTURE_TARGET_ARB);
                pbAttribList->push_back(WGL_TEXTURE_CUBE_MAP_ARB);
            }
            else
            {
                pbAttribList->push_back(WGL_TEXTURE_TARGET_ARB);
                pbAttribList->push_back(WGL_TEXTURE_2D_ARB);
            }

            if (bIsFloatBuffer || bIsATIFloatBuffer)
            {
                if(m_numComponents == 0)
                {
                    PB_FPF_WARNING("warning : components not specified. assuming rgba...\n");
                    pfAttribList->push_back(WGL_RED_BITS_ARB);
                    pfAttribList->push_back(m_bitsPerComponent);
                    pfAttribList->push_back(WGL_GREEN_BITS_ARB);
                    pfAttribList->push_back(m_bitsPerComponent);
                    pfAttribList->push_back(WGL_BLUE_BITS_ARB);
                    pfAttribList->push_back(m_bitsPerComponent);
                    pfAttribList->push_back(WGL_ALPHA_BITS_ARB);
                    pfAttribList->push_back(m_bitsPerComponent);
                    m_numComponents = 4;
                }
            }

            if (bIsFloatBuffer)
            {
                switch(m_numComponents)
                {
                case 1:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_R_NV);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_FLOAT_R_NV);
                    break;
                case 2:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RG_NV);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_FLOAT_RG_NV);
                    break;
                case 3:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGB_NV);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_FLOAT_RGB_NV);
                    break;
                case 4:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RECTANGLE_FLOAT_RGBA_NV);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_FLOAT_RGBA_NV);
                    break;
                default:
                    PB_FPF_WARNING("warning : Bad number of components (r=1,rg=2,rgb=3,rgba=4): %d.\n", m_numComponents);
                    break;
                }
            }
            else
            {
                switch(m_numComponents)
                {
                case 3:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RGB_ARB);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_RGB_ARB);
                    break;
                case 4:
                    pfAttribList->push_back(WGL_BIND_TO_TEXTURE_RGBA_ARB);
                    pfAttribList->push_back(true);

                    pbAttribList->push_back(WGL_TEXTURE_FORMAT_ARB);
                    pbAttribList->push_back(WGL_TEXTURE_RGBA_ARB);
                    break;
                default:
                    PB_FPF_WARNING("warning : Bad number of components (r=1,rg=2,rgb=3,rgba=4): %d.\n", m_numComponents);
                    break;
                }
            }

            std::string option = getStringValue(token);
            if (option == "depth")
            {
                pfAttribList->push_back(WGL_BIND_TO_TEXTURE_DEPTH_NV);
                pfAttribList->push_back(true);

                pbAttribList->push_back(WGL_DEPTH_TEXTURE_FORMAT_NV);
                pbAttribList->push_back(WGL_TEXTURE_DEPTH_COMPONENT_NV);
            }
            continue;
        }

        if (token.find("mipmap") == 0 && bIsTexture)
        {
            pbAttribList->push_back(WGL_MIPMAP_TEXTURE_ARB);
            pbAttribList->push_back(true);

            continue;
        }

        PB_FPF_WARNING("warning : unknown pbuffer attribute: %s.\n", token.c_str());
    }

    if (m_numComponents > 0)
    {
        pfAttribList->push_back(WGL_RED_BITS_ARB);
        pfAttribList->push_back(m_bitsPerComponent);
    }
    if (m_numComponents > 1)
    {
        pfAttribList->push_back(WGL_GREEN_BITS_ARB);
        pfAttribList->push_back(m_bitsPerComponent);
    }
    if (m_numComponents > 2)
    {
        pfAttribList->push_back(WGL_BLUE_BITS_ARB);
        pfAttribList->push_back(m_bitsPerComponent);
    }
    if (m_numComponents > 3)
    {
        pfAttribList->push_back(WGL_ALPHA_BITS_ARB);
        pfAttribList->push_back(m_bitsPerComponent);
    }
}

std::string PBuffer::getStringValue(std::string token)
{
	size_t pos;
	if ((pos = token.find("=")) != token.npos)
    {
		std::string value = token.substr(pos+1, token.length()-pos+1);
		return value;
    }
    else
        return "";
}

int PBuffer::getIntegerValue(std::string token)
{
    size_t pos;
    if ((pos = token.find("=")) != token.npos)
    {
        std::string value = token.substr(pos+1, token.length()-pos+1);
        if (value.empty())
            return 1;
        return atoi(value.c_str());
    }
    else
        return 1;
}
