#if defined(WIN32)
#  include <windows.h>
#  include "GLee.h"
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#  include <GL/glxext.h>
#elif defined(MACOS)
#  include <AGL/agl.h>
#endif

#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "RTTBuffer.h"
#include "PBuffer.h"

using namespace glrc;

/**
 * Constructor. You can create an instance without valid OpenGL contexts.
 */
RTTBuffer::RTTBuffer()
{
    m_rttType = GLRC_RTT_AUTO;
	m_bFPBuffer = false;
    m_width = 256;
    m_height = 256;
    m_textureIndex = 0;
    m_pPBuffer = NULL;
    m_pTextureBuffer = NULL;
    m_pFPTextureBuffer = NULL;
	m_fboIndex = 0;
	m_depthIndex = 0;

	m_bRenderingToTexture = false;
	m_bUsingTexture = false;
}

/**
 * Destructor.
 */
RTTBuffer::~RTTBuffer()
{
	clear();
}

/**
 * Initialize the off-screen buffer. You must initialize the buffer in a valid
 * OpenGL context before rendering.
 * Floating point buffer is not available when rttType is READPIXEL nor COPYSUBTEX.
 * At this time, width and height must be power of two!!
 * You can call setup() more than once to re-configure the off-screen buffer.
 *
 * @param rttType RenderToTexture type. GLRC_RTT_AUTO to select the best type automatically.
 * @param width Width of the off-screen buffer.
 * @param height Height of the off-screen buffer.
 * @param bHasAlpha True if you need a alpha channel.
 * @param bFPBuffer True if you need a floating point buffer.
 */
void RTTBuffer::init(GLRC_RTT rttType, int width, int height, bool bHasAlpha, bool bFPBuffer)
{
//	clear();

	// RTT type auto detection
    m_rttType = rttType;
    if (m_rttType == GLRC_RTT_AUTO)
        m_rttType = GLRC_RTT_FBO;
    if (m_rttType == GLRC_RTT_FBO &&
        !GL_EXT_framebuffer_object)
        m_rttType = GLRC_RTT_PBUFFER_RTT;
    if (m_rttType == GLRC_RTT_PBUFFER_RTT &&
        (!WGL_ARB_pbuffer || !WGL_ARB_pixel_format || !WGL_ARB_render_texture))
        m_rttType = GLRC_RTT_PBUFFER;
    if (m_rttType == GLRC_RTT_PBUFFER &&
        (!WGL_ARB_pbuffer || !WGL_ARB_pixel_format))
        m_rttType = GLRC_RTT_COPYSUBTEX;
	m_bHasAlpha = bHasAlpha;
	m_bFPBuffer = bFPBuffer;

    m_width = width;
    m_height = height;
    m_textureIndex = 0;
    m_pTextureBuffer = NULL;
    m_pFPTextureBuffer = NULL;
    m_fboIndex = 0;
	GLuint format = m_bHasAlpha?GL_RGBA:GL_RGB;
	GLuint type = m_bFPBuffer?GL_FLOAT/*GL_HALF_FLOAT_NV*/:GL_UNSIGNED_BYTE;

	// create a texture object to be rendered.
    glGenTextures(1, &m_textureIndex);
    glBindTexture(GL_TEXTURE_2D, m_textureIndex);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// Initialize the texture depends on its RTT type.
    switch (m_rttType)
    {
        case GLRC_RTT_READPIXEL:
        case GLRC_RTT_COPYSUBTEX:
			if (!m_bFPBuffer) {
				m_pTextureBuffer = new unsigned char[m_width * m_height * (m_bHasAlpha?4:3)];
	            glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, type, m_pTextureBuffer);
			} else {
	            m_pFPTextureBuffer = new float[m_width * m_height * (m_bHasAlpha?4:3)];
	            glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, type, m_pFPTextureBuffer);
			}
            break;
        case GLRC_RTT_PBUFFER:
			if (!m_bFPBuffer)
	            m_pPBuffer = new PBuffer("rgba depth", true);
			else
	            m_pPBuffer = new PBuffer("rgba ati_float=16 depth", true);
            m_pPBuffer->init(width, height, false, true);
			if (!m_bFPBuffer) {
	            m_pTextureBuffer = new unsigned char[m_width * m_height * (m_bHasAlpha?4:3)];
	            glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, GL_RGBA, type, m_pTextureBuffer);
			} else {
	            m_pFPTextureBuffer = new float[m_width * m_height * (m_bHasAlpha?4:3)];
				glTexImage2D(GL_TEXTURE_2D, 0, m_bHasAlpha?GL_RGBA_FLOAT16_ATI:GL_RGB_FLOAT16_ATI, width, height, 0, format, type, m_pFPTextureBuffer);
			}
            break;
        case GLRC_RTT_PBUFFER_RTT:
			if (!m_bFPBuffer)
				m_pPBuffer = new PBuffer("rgba depth texture2D", true);
			else
				m_pPBuffer = new PBuffer("rgb ati_float=16 depth texture2D", true);
            m_pPBuffer->init(width, height, false, true);
            break;
        case GLRC_RTT_FBO:
            glGenFramebuffersEXT(1, &m_fboIndex);
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_fboIndex);
		    glGenRenderbuffersEXT(1, &m_depthIndex);

			if (!m_bFPBuffer)
				glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, type, NULL);
			else
				glTexImage2D(GL_TEXTURE_2D, 0, m_bHasAlpha?GL_RGBA_FLOAT16_ATI:GL_RGB_FLOAT16_ATI, width, height, 0, format, type, NULL);
            glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_textureIndex, 0);
			glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, m_depthIndex);
			glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height);
			glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, m_depthIndex);
            break;
    }
}

/**
 * Retrieve RenderToTexture type of the off-screen buffer.
 *
 * @return RenderToTexture type of the off-screen buffer.
 */
GLRC_RTT RTTBuffer::getType() const
{
    return m_rttType;
}

/**
 * Retrieve width of the off-screen buffer.
 *
 * @return Width of the off-screen buffer.
 */
int RTTBuffer::getWidth() const
{
    return m_width;
}

/**
 * Retrieve height of the off-screen buffer.
 *
 * @return Height of the off-screen buffer.
 */
int RTTBuffer::getHeight() const
{
    return m_height;
}

/**
 * Check whether the off-screen has alpha channel or not.
 *
 * @return True if the off-screen buffer has alpha channel or not.
 */
bool RTTBuffer::hasAlpha() const
{
	return m_bHasAlpha;
}

/**
 * Check whether the off-screen is floating buffer type or not.
 *
 * @return True if the off-screen buffer is a floating point buffer.
 */
bool RTTBuffer::isFPBuffer() const
{
	return m_bFPBuffer;
}

/**
 * Retrieve internal texture object index.
 *
 * @return Internal texture object index.
 */
unsigned int RTTBuffer::getTextureIndex() const
{
    return m_textureIndex;
}

/**
 * Retrieve internal texture buffer (invalid when rtt type is PBUFFER_RTT or FBO).
 *
 * @return Internal texture buffer.
 */
unsigned char* RTTBuffer::getTextureBuffer() const
{
	return m_pTextureBuffer;
}

/**
 * Retrieve internal floating point texture buffer (invalid when rtt type is PBUFFER_RTT or FBO).
 *
 * @return Internal floating point texture buffer.
 */
float* RTTBuffer::getFPTextureBuffer() const
{
	return m_pFPTextureBuffer;
}

/**
 * Retrieve internal P-Buffer object (valid when rtt type is PBUFFER or PBUFFER_RTT).
 *
 * @return Internal P-Buffer object.
 */
PBuffer* RTTBuffer::getPBuffer() const
{
    return m_pPBuffer;
}

/**
 * Retrieve internal frame buffer object (FBO) index (valid when rtt type is FBO).
 *
 * @return Internal frame buffer object (FBO) index.
 */
unsigned int RTTBuffer::getFBOIndex() const
{
    return m_fboIndex;
}

/**
 * Begin to render to the texture. Activate the RTT rendering context, then change the viewport to render.
 */
void RTTBuffer::beginRenderToTexture()
{
	if (m_bRenderingToTexture)
		return;
	m_bRenderingToTexture = true;

    activate();

	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, m_width, m_height); 
}

/**
 * Finish to render to the texture. Copy the rendered results to texture buffer (if needed), then restore the previous rendering context.
 */
void RTTBuffer::endRenderToTexture()
{
//	if (!m_bRenderingToTexture)
//		return;
	m_bRenderingToTexture = false;

	glPopAttrib();

    if (m_rttType == GLRC_RTT_FBO)
    {
        deactivate();
    }
    else
    {
        activate();

        if (m_rttType == GLRC_RTT_PBUFFER ||
			m_rttType == GLRC_RTT_PBUFFER_RTT)
			m_pPBuffer->swapBuffers();
        if (m_rttType == GLRC_RTT_PBUFFER_RTT)
            deactivate();

		// copy to texture 'before' switching the p-buffer context.
        copyToTexture();

        if (m_rttType == GLRC_RTT_PBUFFER)
            deactivate();
    }
}

/**
 * Check whether currently rendering to the texture or not.
 *
 * @return True if rendering to the texture.
 */
bool RTTBuffer::isRenderingToTexture() const
{
	return m_bRenderingToTexture;
}

/**
 * Bind the rendered texture to use.
 */
void RTTBuffer::beginUseTexture()
{
	if (m_bUsingTexture)
		return;
	m_bUsingTexture = true;

    glBindTexture(GL_TEXTURE_2D, m_textureIndex);
    switch (m_rttType)
    {
        case GLRC_RTT_PBUFFER_RTT:
            m_pPBuffer->bind(WGL_FRONT_LEFT_ARB);
            break;
    }
}

/**
 * Unbind the rendered texture.
 */
void RTTBuffer::endUseTexture()
{
	if (!m_bUsingTexture)
		return;
	m_bUsingTexture = false;

    switch (m_rttType)
    {
        case GLRC_RTT_PBUFFER_RTT:
            // release p-buffer texture for further rendering
            m_pPBuffer->release(WGL_FRONT_LEFT_ARB);
            break;
    }
}

/**
 * Check whether currently using the rendered texture or not.
 *
 * @return True if the rendered texture is currently used.
 */
bool RTTBuffer::isUsingTexture() const
{
	return m_bUsingTexture;
}

/**
 * Save the current off-screen buffer to an image file.
 * Image file format is determined by its file extention.
 * Supported format is depending on DevIL library (and .hdr format).
 *
 * @param Image file name to save the off-screen buffer.
 * @return True if image is saved successfully.
 */
bool RTTBuffer::saveBuffer(const char* fileName)
{
	if (m_rttType == GLRC_RTT_READPIXEL ||
		m_rttType == GLRC_RTT_COPYSUBTEX) return false;

    // store current RenderToTexture / UsingTexture state
    bool bIsUsingTexture = m_bUsingTexture;
    if (bIsUsingTexture)
        endUseTexture();

    activate();

    bool result = false;

    unsigned int imgIndex;
    ilGenImages(1, &imgIndex);
    ilBindImage(imgIndex);
    ilEnable(IL_FILE_OVERWRITE);

    if (!m_bFPBuffer)
    {
	    unsigned char* pBuffer = getBuffer();
        if (pBuffer) {
            if (ilTexImage(m_width, m_height, 1, m_bHasAlpha?4:3, m_bHasAlpha?IL_RGBA:IL_RGB, IL_UNSIGNED_BYTE, pBuffer))
                result = (ilSaveImage((ILstring)fileName) == IL_NO_ERROR);
	        delete [] pBuffer;
        }
    }
    else
    {
        float* pFPBuffer = getFPBuffer();
        if (pFPBuffer) {
            if (ilTexImage(m_width, m_height, 1, m_bHasAlpha?4:3, m_bHasAlpha?IL_RGBA:IL_RGB, IL_UNSIGNED_SHORT, pFPBuffer))
                result = (ilSaveImage((ILstring)fileName) == IL_NO_ERROR);
	        delete [] pFPBuffer;
        }
    }

    ilDeleteImages(1, &imgIndex);

    if (bIsUsingTexture)
        beginUseTexture();

	return result;
}

/*
 * Retrieve the current off-screen buffer as a bitmap buffer.
 *
 * @return Buffer pointer. NULL if fails.
 */
unsigned char* RTTBuffer::getBuffer()
{
	unsigned char* pBuffer = new unsigned char[m_width * m_height * (m_bHasAlpha?4:3)];
	if (getBuffer(pBuffer))
		return pBuffer;
	delete [] pBuffer;
	return NULL;
}

/*
 * Retrieve the current off-screen buffer as a bitmap buffer.
 *
 * @param pBuffer Buffer to retrieve the off-screen image.
 * @return True if buffer is retrieved successfully.
 */
bool RTTBuffer::getBuffer(unsigned char* pBuffer)
{
	if (m_rttType == GLRC_RTT_READPIXEL ||
		m_rttType == GLRC_RTT_COPYSUBTEX) return false;

    if (m_bFPBuffer)
        return false;

    // store current RenderToTexture / UsingTexture state
    bool bIsUsingTexture = m_bUsingTexture;
    if (bIsUsingTexture)
        endUseTexture();
    bool bIsRenderingToTexture = m_bRenderingToTexture;
    if (!bIsRenderingToTexture)
        beginRenderToTexture();

	glReadPixels(0, 0, m_width, m_height, m_bHasAlpha?GL_RGBA:GL_RGB, GL_UNSIGNED_BYTE, pBuffer);

    // restore current RenderToTexture / UsingTexture state
    if (!bIsRenderingToTexture)
        endRenderToTexture();
    if (bIsUsingTexture)
        beginUseTexture();

    return true;
}

/*
 * Retrieve the current off-screen buffer as a bitmap buffer.
 *
 * @return Buffer pointer. NULL if fails.
 */
float* RTTBuffer::getFPBuffer()
{
    float* pFPBuffer = new float[m_width * m_height * m_bHasAlpha?4:3];
	if (getFPBuffer(pFPBuffer))
		return pFPBuffer;
	delete [] pFPBuffer;
	return NULL;
}

/*
 * Retrieve the current off-screen buffer as a bitmap buffer.
 *
 * @param pFPBuffer Buffer to retrieve the off-screen image.
 * @return True if buffer is retrieved successfully.
 */
bool RTTBuffer::getFPBuffer(float* pFPBuffer)
{
	if (m_rttType == GLRC_RTT_READPIXEL ||
		m_rttType == GLRC_RTT_COPYSUBTEX) return false;

    if (!m_bFPBuffer)
        return false;

    // store current RenderToTexture / UsingTexture state
    bool bIsUsingTexture = m_bUsingTexture;
    if (bIsUsingTexture)
        endUseTexture();
    bool bIsRenderingToTexture = m_bRenderingToTexture;
    if (!bIsRenderingToTexture)
        beginRenderToTexture();

    glReadPixels(0, 0, m_width, m_height, m_bHasAlpha?GL_RGBA_FLOAT16_ATI:GL_RGB_FLOAT16_ATI, GL_FLOAT, pFPBuffer);

    // restore current RenderToTexture / UsingTexture state
    if (!bIsRenderingToTexture)
        endRenderToTexture();
    if (bIsUsingTexture)
        beginUseTexture();

    return true;
}

/**
 * Clear the internal data.
 */
void RTTBuffer::clear()
{
	if (glIsTexture(m_textureIndex))
		glDeleteTextures(1, &m_textureIndex);
	if (m_pTextureBuffer)
		delete [] m_pTextureBuffer;
	if (m_pFPTextureBuffer)
		delete [] m_pFPTextureBuffer;
	if (m_pPBuffer)
		delete m_pPBuffer;
	if (glIsFramebufferEXT(m_fboIndex))
		glDeleteFramebuffersEXT(1, &m_fboIndex);
	if (glIsRenderbufferEXT(m_fboIndex))
		glDeleteRenderbuffersEXT(1, &m_fboIndex);
}

/**
 * Ativate the P-Buffer context or bind the FBO.
 */
void RTTBuffer::activate()
{
    switch (m_rttType)
    {
        case GLRC_RTT_PBUFFER:
        case GLRC_RTT_PBUFFER_RTT:
            m_pPBuffer->activate();
            break;
        case GLRC_RTT_FBO:
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_fboIndex);
            glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_textureIndex, 0);
            break;
    }
}

/**
 * Deactivate P-Buffer context or unbind the FBO.
 */
void RTTBuffer::deactivate()
{
    switch (m_rttType)
    {
        case GLRC_RTT_PBUFFER:
        case GLRC_RTT_PBUFFER_RTT:
            m_pPBuffer->deactivate();
            break;
        case GLRC_RTT_FBO:
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
            break;
    }
}

/**
 * Copy the current buffer to the texture buffer (if needed).
 */
void RTTBuffer::copyToTexture()
{
    switch (m_rttType)
    {
        case GLRC_RTT_READPIXEL:
            glBindTexture(GL_TEXTURE_2D, m_textureIndex);
			if (!m_bFPBuffer) {
	            glReadPixels(0, 0, m_width, m_height, m_bHasAlpha?GL_RGBA:GL_RGB, GL_UNSIGNED_BYTE, m_pTextureBuffer);
				glTexImage2D(GL_TEXTURE_2D, 0, m_bHasAlpha?GL_RGBA:GL_RGB, m_width, m_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, m_pTextureBuffer);
			} else {
	            glReadPixels(0, 0, m_width, m_height, m_bHasAlpha?GL_RGBA:GL_RGB, GL_FLOAT, m_pFPTextureBuffer);
				glTexImage2D(GL_TEXTURE_2D, 0, m_bHasAlpha?GL_RGBA:GL_RGB, m_width, m_height, 0, GL_RGBA, GL_FLOAT, m_pFPTextureBuffer);
			}
            break;
        case GLRC_RTT_COPYSUBTEX:
            glBindTexture(GL_TEXTURE_2D, m_textureIndex);
            glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);
            break;
        case GLRC_RTT_PBUFFER:
            glBindTexture(GL_TEXTURE_2D, m_textureIndex);
            glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_width, m_height);
            break;
    }
}
