#ifndef _GLRC_RTTBUFFER_H_
#define _GLRC_RTTBUFFER_H_

#include "glrc.h"

namespace glrc
{
class PBuffer;

/**
 * Render To Texture (off-screen) buffer object class.
 */
class GLRC_API RTTBuffer
{
public:

	/**
	 * Constructor. You can create an instance without valid OpenGL contexts.
	 */
    RTTBuffer();

	/**
	 * Destructor.
	 */
    virtual ~RTTBuffer();

	/**
	 * Initialize the off-screen buffer. You must initialize the buffer in a valid
	 * OpenGL context before rendering.
	 * Floating point buffer is not available when rttType is READPIXEL nor COPYSUBTEX.
     * At this time, width and height must be power of two!!
     * You can call setup() more than once to re-configure the off-screen buffer.
	 *
	 * @param rttType RenderToTexture type. GLRC_RTT_AUTO to select the best type automatically.
	 * @param width Width of the off-screen buffer.
	 * @param height Height of the off-screen buffer.
	 * @param bHasAlpha True if you need a alpha channel.
	 * @param bFPBuffer True if you need a floating point buffer.
	 */
    void init(GLRC_RTT rttType, int width, int height, bool bHasAlpha=false, bool bFPBuffer=false);

	/**
	 * Retrieve RenderToTexture type of the off-screen buffer.
	 *
	 * @return RenderToTexture type of the off-screen buffer.
	 */
    GLRC_RTT getType() const;

	/**
	 * Retrieve width of the off-screen buffer.
	 *
	 * @return Width of the off-screen buffer.
	 */
    int getWidth() const;

	/**
	 * Retrieve height of the off-screen buffer.
	 *
	 * @return Height of the off-screen buffer.
	 */
    int getHeight() const;

	/**
	 * Check whether the off-screen has alpha channel or not.
	 *
	 * @return True if the off-screen buffer has alpha channel or not.
	 */
	bool hasAlpha() const;

	/**
	 * Check whether the off-screen is floating buffer type or not.
	 *
	 * @return True if the off-screen buffer is a floating point buffer.
	 */
	bool isFPBuffer() const;

	/**
	 * Retrieve internal texture object index.
	 *
	 * @return Internal texture object index.
	 */
    unsigned int getTextureIndex() const;

	/**
	 * Retrieve internal texture buffer (invalid when rtt type is PBUFFER_RTT or FBO).
	 *
	 * @return Internal texture buffer.
	 */
	unsigned char* getTextureBuffer() const;

	/**
	 * Retrieve internal floating point texture buffer (invalid when rtt type is PBUFFER_RTT or FBO).
	 *
	 * @return Internal floating point texture buffer.
	 */
	float* getFPTextureBuffer() const;

	/**
	 * Retrieve internal P-Buffer object (valid when rtt type is PBUFFER or PBUFFER_RTT).
	 *
	 * @return Internal P-Buffer object.
	 */
    PBuffer* getPBuffer() const;

	/**
	 * Retrieve internal frame buffer object (FBO) index (valid when rtt type is FBO).
	 *
	 * @return Internal frame buffer object (FBO) index.
	 */
    unsigned int getFBOIndex() const;

	/**
	 * Begin to render to the texture. Activate the RTT rendering context, then change the viewport to render.
	 */
    void beginRenderToTexture();

	/**
	 * Finish to render to the texture. Copy the rendered results to texture buffer (if needed), then restore the previous rendering context.
	 */
    void endRenderToTexture();

	/**
	 * Check whether currently rendering to the texture or not.
	 *
	 * @return True if rendering to the texture.
	 */
	bool isRenderingToTexture() const;

	/**
	 * Bind the rendered texture to use.
	 */
    void beginUseTexture();

	/**
	 * Unbind the rendered texture.
	 */
    void endUseTexture();

	/**
	 * Check whether currently using the rendered texture or not.
	 *
	 * @return True if the rendered texture is currently used.
	 */
	bool isUsingTexture() const;

    /**
     * Save the current off-screen buffer to an image file.
     * Image file format is determined by its file extention.
     * Supported format is depending on DevIL library (and .hdr format).
     *
     * @param Image file name to save the off-screen buffer.
     * @return True if image is saved successfully.
     */
    bool saveBuffer(const char* fileName);

    /*
     * Retrieve the current off-screen buffer as a bitmap buffer.
     *
     * @return Buffer pointer. NULL if fails.
     */
    unsigned char* getBuffer();

	/*
	 * Retrieve the current off-screen buffer as a bitmap buffer.
	 *
	 * @param pBuffer Buffer to retrieve the off-screen image.
	 * @return True if buffer is retrieved successfully.
	 */
	bool getBuffer(unsigned char* pFPBuffer);

    /*
     * Retrieve the current off-screen buffer as a bitmap buffer.
     *
     * @return Buffer pointer. NULL if fails.
     */
    float* getFPBuffer();

	/*
	 * Retrieve the current off-screen buffer as a bitmap buffer.
	 *
	 * @param pFPBuffer Buffer to retrieve the off-screen image.
	 * @return True if buffer is retrieved successfully.
	 */
	bool getFPBuffer(float* pFPBuffer);

protected:

	/**
	 * Clear the internal data.
	 */
	void clear();

	/**
	 * Ativate the P-Buffer context or bind the FBO.
	 */
    void activate();

	/**
	 * Deactivate P-Buffer context or unbind the FBO.
	 */
    void deactivate();

	/**
	 * Copy the current buffer to the texture buffer (if needed).
	 */
    void copyToTexture();

protected:

    GLRC_RTT m_rttType;	//!< RenderToTexture type.
	bool m_bHasAlpha;	//!< True if the off-screen buffer has alpha channel.
	bool m_bFPBuffer;	//!< True if the off-screen buffer is floating point buffer.
    int m_width;		//!< Width of the off-screen buffer.
    int m_height;		//!< Height of the off-screen buffer.

    unsigned int m_textureIndex;		//!< Rendered texture object index.
    unsigned char* m_pTextureBuffer;	//!< Rendered texture buffer.
    float* m_pFPTextureBuffer;			//!< Rendered texture buffer (for floating point buffer).
    PBuffer* m_pPBuffer;				//!< P-Buffer object (used in PBUFFER or PBUFFER_RTT).
    unsigned int m_fboIndex;			//!< Frame Buffer Object (FBO) index (used in FBO).
    unsigned int m_depthIndex;			//!< Depth render buffer object index (used in FBO).

	bool m_bRenderingToTexture;	//!< True if currently rendering to the texture.
	bool m_bUsingTexture;		//!< True if currently using the rendered texture.
};

}; // namespace glrc

#endif // _GLRC_RTTBUFFER_H_
