#ifdef WIN32
#include <windows.h>
#include <mmsystem.h>
#endif // WIN32

#include <stdio.h>
#include <stdlib.h>

#include "Statistics.h"

#include "glee.h"
#include "GL/gl.h"
#include "GL/glu.h"

#ifndef WIN32
#include "glut.h"
#endif

using namespace glrc;

/**
 * Constructor. You can create an instance without valid OpenGL contexts.
 */
Statistics::Statistics()
{
    m_bShow = true;
    m_updateInterval = 1000;

	m_waitVSyncMode = GLRC_VSYNC_UNKNOWN;

    m_frame = 0;
    m_lastTime = 0;
    m_fps = 0.0f;

	m_fontBase = 0;

#ifdef WIN32
    // Initialize high resolution timer.
	m_timePeriodIndex = 0;
	TIMECAPS timeCaps;
	if (::timeGetDevCaps(&timeCaps, sizeof(TIMECAPS)) == TIMERR_NOERROR) {
		m_timePeriodIndex = timeCaps.wPeriodMin;
		::timeBeginPeriod(m_timePeriodIndex);
	}
#endif // WIN32
}

/**
 * Destructor.
 */
Statistics::~Statistics()
{
#ifdef WIN32
	if (m_timePeriodIndex > 0)
		::timeEndPeriod(m_timePeriodIndex);
#endif // WIN32

#ifdef WIN32
	if (glIsList(m_fontBase))
		glDeleteLists(m_fontBase, 96);
#else // WIN32
	if (glIsList(m_fontBase))
		glDeleteLists(m_fontBase, 128);
#endif // WIN32
}

/**
 * Initilaize OpenGL objects. Create bitmap font display lists.
 */
void Statistics::init()
{
#ifdef WIN32
    // Use wgl function to create a bit map font.
	m_fontBase = glGenLists(96);						// Storage For 96 Characters

	HFONT	font;										// Windows Font ID
	HFONT	oldfont;									// Used For Good House Keeping

	font = CreateFont(-20,								// Height Of Font
					  0,								// Width Of Font
					  0,								// Angle Of Escapement
					  0,								// Orientation Angle
					  FW_NORMAL,						// Font Weight
					  FALSE,							// Italic
					  FALSE,							// Underline
					  FALSE,							// Strikeout
					  ANSI_CHARSET,						// Character Set Identifier
					  OUT_TT_PRECIS,					// Output Precision
					  CLIP_DEFAULT_PRECIS,				// Clipping Precision
					  ANTIALIASED_QUALITY,				// Output Quality
					  FF_DONTCARE|DEFAULT_PITCH,		// Family And Pitch
					  "Arial");							// Font Name

	HDC hDC = wglGetCurrentDC();
	oldfont = (HFONT)SelectObject(hDC, font);           // Selects The Font We Want
	wglUseFontBitmaps(hDC, 32, 96, m_fontBase);			// Builds 96 Characters Starting At Character 32
	SelectObject(hDC, oldfont);							// Selects The Font We Want
	DeleteObject(font);									// Delete The Font
#else // WIN32
    // Use glut function to create a bit map font.
	m_fontBase = glGenLists(128);
	for (unsigned char c=0; c<128; ++c) {
		glNewList(m_fontBase+c, GL_COMPILE);
	    glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, c);
		glEndList();
	}
#endif // WIN32
}

/**
 * Show statistics.
 */
void Statistics::show()
{
    m_bShow = true;
}

/**
 * Hide statistics.
 */
void Statistics::hide()
{
    m_bShow = false;
}

/**
 * Check whether the statistics is shown or not.
 */
bool Statistics::isShown() const
{
    return m_bShow;
}

/**
 * Set update interval.
 *
 * @param interval Interval count (1000 is default).
 */
void Statistics::setUpdateInterval(unsigned int interval)
{
    m_updateInterval = interval;
}

/**
 * Retrieve update interval.
 *
 * @return interval Interval count (1000 is default).
 */
unsigned int Statistics::getUpdateInterval() const
{
    return m_updateInterval;
}

/**
 * Set additional description text.
 *
 * @param text Additional description text.
 */
void Statistics::setDescription(const char* text)
{
    m_description = text;
}

 /**
 * Retrieve the additional description text.
 *
 * @return Additional description text.
 */
const char* Statistics::getDescription() const
{
    return m_description.c_str();
}

/**
 * Update FPS counting.
 */
void Statistics::update()
{
    m_frame++;
#ifdef WIN32
    int curTime = ::timeGetTime();
#else
    // Use glut to get time value.
    int curTime = glutGet(GLUT_ELAPSED_TIME);
#endif //WIN32
    if (curTime - m_lastTime > 1000) {
        m_fps = m_frame * 1000.0f / (curTime - m_lastTime);
        m_lastTime = curTime;
        m_frame = 0;
    }   
}

/**
 * Draw the statistics.
 */
void Statistics::draw()
{
    if (m_bShow)
    {
        GLint viewport[4];
        glGetIntegerv(GL_VIEWPORT, viewport);

        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        gluOrtho2D(viewport[0], viewport[2], viewport[1], viewport[3]);
        glScalef(1.0f, -1.0f, 1.0f);
        glTranslatef(0.0f, (float)-viewport[3], 0.0f);

        glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT); 
        glDisable(GL_LIGHTING);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_TEXTURE_2D);

        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

        char text[256];
		int y = 32;

        // FPS
        sprintf(text,"FPS: %4.2f %s", m_fps, m_waitVSyncMode==GLRC_VSYNC_UNKNOWN?"":(m_waitVSyncMode==GLRC_VSYNC_WAIT?"(v-sync locked)":"(v-sync free)"));
		drawString(16, y, text); y+=20;

        // GLRC related information
		if (glrcIsEnabled(GLRC_COMPOSITOR_2D))
		{
			switch (glrcGetBufferType())
			{
			case GLRC_RTT_READPIXEL:
				strcpy(text, "RTT: READPIXEL"); break;
			case GLRC_RTT_COPYSUBTEX:
				strcpy(text, "RTT: COPYSUBTEX"); break;
			case GLRC_RTT_PBUFFER:
				strcpy(text, "RTT: PBUFFER"); break;
			case GLRC_RTT_PBUFFER_RTT:
				strcpy(text, "RTT: PBUFFER_RTT"); break;
			case GLRC_RTT_FBO:
				strcpy(text, "RTT: FBO"); break;
			}
			sprintf(text, "%s%s %ix%i", text, glrcIsFPBuffer()?" (float)":"", glrcGetBufferWidth(), glrcGetBufferHeight());
			drawString(16, y, text); y+=20;
			sprintf(text, "Effect:");
			if (glrcGetEffectMode() & GLRC_EFFECT_DISTORTION)
				strcat(text, " DISTORTION");
			if (glrcGetEffectMode() & GLRC_EFFECT_HDR)
				strcat(text, " HDR");
			if (glrcGetEffectMode() & GLRC_EFFECT_COLORMATRIX)
				strcat(text, " COLORMATRIX");
			drawString(16, y, text); y+=20;
			sprintf(text, "exposure: %f", glrcGetExposure());
			drawString(16, y, text); y+=20;
			sprintf(text, "gamma: %f", glrcGetGamma());
			drawString(16, y, text); y+=20;
        }

        y+=20;

        // additional description
        std::string::size_type first = 0;
        std::string::size_type last = 0;
        while ((first = m_description.find_first_not_of("\n", last)) != std::string::npos )
        {
            last = m_description.find_first_of("\n", first);
            drawString(16, y, m_description.substr(first, last == std::string::npos ? last : last-first).c_str()); y+=20;
        }

        glPopAttrib();

        glPopMatrix();

        glMatrixMode(GL_MODELVIEW);
    }
}

/**
 * Set V-Sync waiting mode.
 *
 * @param V-Sync waiting mode.
 */
void Statistics::setWaitVSync(GLRC_VSYNC waitVSyncMode)
{
	if (waitVSyncMode != GLRC_VSYNC_UNKNOWN && WGL_EXT_swap_control) {
		wglSwapIntervalEXT(waitVSyncMode==GLRC_VSYNC_WAIT?1:0);
	    m_waitVSyncMode = waitVSyncMode;
	}
	else
	    m_waitVSyncMode = GLRC_VSYNC_UNKNOWN;
}

/**
 * Retrieve V-Sync waiting mode.
 *
 * @return V-Sync waiting mode.
 */
GLRC_VSYNC Statistics::getWaitVSync() const
{
    return m_waitVSyncMode;
}

/**
 * Draw bitmap font string.
 *
 * @param x X (left) position.
 * @param y Y (top) position.
 * @param string String to draw.
 */
void Statistics::drawString(int x, int y, const char* string)
{
    glRasterPos2i(x, y);

	glPushAttrib(GL_LIST_BIT);
#ifdef WIN32
	glListBase(m_fontBase-32);
#else // WIN32
	glListBase(m_fontBase);
#endif
	glCallLists(strlen(string), GL_UNSIGNED_BYTE, string);
	glPopAttrib();
}
