#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glut.h>

#include "glrc/glrc.h"
#include "gltb.h"

unsigned int compositorIndex;
int wwidth;
int wheight;

/* Initialize */
static void init()
{
    /* initialize glrc lirbary */
	glrcInit();

    /* create a compositor */
	glrcGenCompositors(1, &compositorIndex);
	glrcBindCompositor(GLRC_COMPOSITOR_2D, compositorIndex);
    glrcSetupBuffer(GLRC_RTT_AUTO, 1024, 1024, FALSE, FALSE);

    /* setup distortion and blendings */
	glrcSetDistortionMap("distort.png");
	glrcSetBlendMap("blend.png");
    glrcLoadMatrixFile("view.cfg");
	glrcSetEffectMode(GLRC_EFFECT_DISTORTION);
    glrcDisable(GLRC_MESHDISTORTION);

	/* initialize trackball manipulator */
	gltbInit(GLUT_LEFT_BUTTON);
	gltbAnimate(GL_TRUE);

	printf("GLRC distortion sample program\n");
	printf(" [d] key to toggle distortion.\n");
	printf(" [m] key to toggle mesh distortion / shader distortion.\n");
}

/* Draw a teapot scene */
static void scene_teapot()
{
	/* lighting parameters */
	static const GLfloat lightpos[] = { 0.0, 0.0, 5.0, 1.0 };
	static const GLfloat lightcol[] = { 1.0, 1.0, 1.0, 1.0 };
	static const GLfloat lightamb[] = { 0.1, 0.1, 0.1, 1.0 };

	/* material */
	static const GLfloat diffuse[] = { 0.6, 0.6, 0.6, 1.0 };
	static const GLfloat specular[] = { 0.3, 0.3, 0.3, 1.0 };

	if (!glrcIsEnabled(GLRC_PROJECTION))
	{
		/* initialize projection matrix (without view matrix file) */
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(60.0, (double)wwidth / (double)wheight, 0.1, 100.0);
	}

    /* clear background */
    glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_TEXTURE_2D);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

	if (!glrcIsEnabled(GLRC_MODELVIEW))
	{
		/* initialize model view matrix (without view matrix file) */
		glLoadIdentity();
	}

	/* setup lighting */
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightcol);
	glLightfv(GL_LIGHT0, GL_SPECULAR, lightcol);
	glLightfv(GL_LIGHT0, GL_AMBIENT, lightamb);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos);

	/* place a teapot */
	glTranslated(0.0, 0.0, -0.75);

	/* apply trackball manipulator matrix */
	gltbMatrix();

	/* setup material */
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 100.0f);

    /* draw a teapot */
	glutWireTeapot(0.25);

	glPopMatrix();
}

/* Display callback function */
static void display(void)
{
    /* activate compositor */
	glrcBindCompositor(GLRC_COMPOSITOR_2D, compositorIndex);

    /* clear background */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* begin to render a scene into off-screen buffer */
	glrcBeginRender();

      /* draw the scene */
      scene_teapot();

    /* now use the off-screen rendered scene image to display */
	glrcEndRender();

	glutSwapBuffers();
}

/* Resize callback function */
static void resize(int w, int h)
{
	wwidth = w;
	wheight = h;

	/* resize trackball input area */
	gltbReshape(w, h);

	/* setup viewport */
	glViewport(0, 0, w, h);
}

/* Mouse event callback function */
static void mouse(int button, int state, int x, int y)
{
	/* update trackball manipulator */
	gltbMouse(button, state, x, y);
}

/* Mouse drag callback function */
static void motion(int x, int y)
{
	/* update trackball manipulator */
	gltbMotion(x, y);
}

/* Keyboard input callback function */
static void keyboard(unsigned char key, int x, int y)
{
	switch (key)
	{
		case 'q':
		case '\033':
			/* ESC or 'q' key to exit */
			exit(0);
		case 'd':
			/* 'd' key to toggle distortion */
			if (glrcGetEffectMode() == GLRC_EFFECT_NONE)
			{
				glrcSetEffectMode(GLRC_EFFECT_DISTORTION);
				printf("Distortion enabled.\n");
			}
			else
			{
				glrcSetEffectMode(GLRC_EFFECT_NONE);
				printf("Distortion disabled.\n");
			}
		    glutPostRedisplay();
			break;
		case 'm':
			/* 'm' key to toggle mesh distortion / shader distortion */
			if (!glrcIsEnabled(GLRC_MESHDISTORTION))
			{
				glrcEnable(GLRC_MESHDISTORTION);
				printf("Mesh distortion enabled.\n");
			}
			else
			{
				glrcDisable(GLRC_MESHDISTORTION);
				printf("Shader distortion enabled.\n");
			}
		    glutPostRedisplay();
			break;
		default:
			break;
	}
}

/* Main routine */
int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(640, 480); 
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutReshapeFunc(resize);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutKeyboardFunc(keyboard);

	init();
	
	glutMainLoop();

	glrcDestroy();

	return 0;
}
