#if defined(WIN32)
#  include <windows.h>
#  include "GLee.h"
#  pragma warning (disable : 4786)
#elif defined(UNIX)
#  include <GL/glx.h>
#  include <GL/glxext.h>
#elif defined(MACOS)
#  include <AGL/agl.h>
#endif

// using DevIL library to load/save images for texture.
#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

// for supporting .hdr format.
extern "C" {
#include "rgbe.h"
}

#include "glrc.h"

#include "Matrix.h"
#include "Compositor.h"

#include <vector>

using namespace glrc;

static GLRC_ERROR g_lastError = GLRC_ERROR_NOERROR;	//!< Latest GLRC error code.

static std::vector<Compositor*> g_pCompositors;	//!< Global compositor array to be binded.
static Compositor* g_pComp = NULL;	//!< Current binded compositor.

// for supporting .hdr format.
static ILenum ILAPIENTRY LoadFunction_hdr(const ILstring fileName);
static ILenum ILAPIENTRY SaveFunction_hdr(const ILstring fileName);

/**
 * Initialize GLRC library. Call it before any GLRC functions.
 */
void glrcInit()
{
	if (g_pComp) {
		glrcSetError(GLRC_ERROR_ALREADY_INITIALIZED);
		return;
	}

	// initialize Glee library.
	GLeeInit();

	// initialize DevIL library.
    ilInit();
    ilRegisterLoad("hdr", LoadFunction_hdr);
    ilRegisterSave("hdr", SaveFunction_hdr);
    iluInit();
    ilutInit();
    ilutRenderer(ILUT_OPENGL);

	for (unsigned int i=0; i<g_pCompositors.size(); ++i)
		if (g_pCompositors[i] != NULL)
			delete g_pCompositors[i];
	g_pCompositors.clear();
	g_pComp = NULL;
}

/**
 * Terminate GLRC library.
 */
void glrcDestroy()
{
	// clear created compositors.
	for (unsigned int i=0; i<g_pCompositors.size(); ++i)
		if (g_pCompositors[i] != NULL)
			delete g_pCompositors[i];
	g_pCompositors.clear();
	g_pComp = NULL;

	// terminate DevIL library.
    ilShutDown();
}

/**
 * Create a new compositor(s).
 *
 * @param n Number of compositors to create.
 * @param compositors Indecies for new compositors.
 */
void glrcGenCompositors(unsigned int n, unsigned int* compositors)
{
	for (unsigned int i=0; i<n; ++i) {
		compositors[i] = g_pCompositors.size();
		g_pCompositors.push_back(new Compositor);
	}
}

/**
 * Delete specified compositor(s).
 *
 * @param n Number of compositors to delete.
 * @param compositors Indecies of compositors to delete.
 */
void glrcDeleteCompositors(unsigned int n, unsigned int* compositors)
{
	if (n < g_pCompositors.size() && g_pCompositors[n] != NULL)
	{
		delete g_pCompositors[n];
		g_pCompositors[n] = NULL;
	}
}

/**
 * Check whether the specified index is an index of an exisiting compositor.
 *
 * @param compositorIndex Index of compositor to check.
 * @return True if the specified index is a valid compositor's.
 */
int glrcIsCompositor(unsigned int compositorIndex)
{
	if (compositorIndex < g_pCompositors.size() && g_pCompositors[compositorIndex] != NULL)
		return TRUE;
	return FALSE;
}

/**
 * Bind the specified compositor as the current compositor. THe following GLRC functions
 * will be applied on this compositor. Take care about the current compositor!
 *
 * @param target Must be GLRC_COMPOSITOR_2D.
 * @param compositorIndex Index of compositor to bind.
 */
void glrcBindCompositor(unsigned int target, unsigned int compositorIndex)
{
	if (target == GLRC_COMPOSITOR_2D)
	{
		if (compositorIndex < g_pCompositors.size() && g_pCompositors[compositorIndex] != NULL)
			g_pComp = g_pCompositors[compositorIndex];
	}
}

/**
 * Initialize the off-screen buffer of the current compositor.
 * You must initialize the buffer in a valid OpenGL context before rendering.
 * Floating point buffer is not available when rttType is READPIXEL nor COPYSUBTEX.
 * At this time, width and height must be power of two!!
 * You can call setup() more than once to re-configure the off-screen buffer.
 *
 * @param rttType RenderToTexture type. GLRC_RTT_AUTO to select the best type automatically.
 * @param width Width of the off-screen buffer.
 * @param height Height of the off-screen buffer.
 * @param bHasAlpha True if you need alpha channel.
 * @param bFPBuffer True if you need a floating point buffer.
 */
void glrcSetupBuffer(GLRC_RTT rttType, int width, int height, int bHasAlpha, int bFPBuffer)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setup(rttType, width, height, bHasAlpha==TRUE, bFPBuffer==TRUE);
}

/**
 * Retrieve RenderToTexture type of the off-screen buffer.
 *
 * @return RenderToTexture type of the off-screen buffer.
 */
GLRC_RTT glrcGetBufferType()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return GLRC_RTT_AUTO;
	}
	return g_pComp->getBufferType();
}

/**
 * Retrieve width of the off-screen buffer.
 *
 * @return Width of the off-screen buffer.
 */
int glrcGetBufferWidth()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return g_pComp->getBufferWidth();
}

/**
 * Retrieve height of the off-screen buffer.
 *
 * @return Height of the off-screen buffer.
 */
int glrcGetBufferHeight()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return g_pComp->getBufferHeight();
}

/**
 * Check whether the off-screen has alpha channel type or not.
 *
 * @return True if the off-screen buffer has alpha channel point buffer.
 */
int glrcHasAlpha()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return (g_pComp->hasAlpha()==TRUE);
}

/**
 * Check whether the off-screen is floating buffer type or not.
 *
 * @return True if the off-screen buffer is a floating point buffer.
 */
int glrcIsFPBuffer()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return (g_pComp->isFPBuffer()==TRUE);
}

/**
 * Set rendering area of the current compositor in current OpenGL viewport.
 * Position and size is defined in normalized (0.0-1.0) coordinate
 * in the current OpenGL viewport.
 *
 * @param x X (left) position of the compositor in normalized (0.0-1.0) coordinate.
 * @param y Y (top) position of the compositor in normalized (0.0-1.0) coordinate.
 * @param width Width of the compositor in normalized (0.0-1.0) coordinate.
 * @param height Height of the compositor in normalized (0.0-1.0) coordinate.
 */
void glrcViewport(float x, float y, float width, float height)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setViewport(x, y, width, height);
}

/**
 * Retrieve x (left) position of the current compositor.
 *
 * @param X (left) position of the current compositor in normalized (0.0-1.0) coordinate.
 */
float glrcGetRenderX()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0.0f;
	}
	return g_pComp->getX();
}

/**
 * Retrieve y (top) position of the current compositor.
 *
 * @param Y (top) position of the current compositor in normalized (0.0-1.0) coordinate.
 */
float glrcGetRenderY()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0.0f;
	}
	return g_pComp->getY();
}

/**
 * Retrieve width of the current compositor.
 *
 * @param Width of the current compositor in normalized (0.0-1.0) coordinate.
 */
float glrcGetRenderWidth()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getWidth();
}

/**
 * Retrieve height of the current compositor.
 *
 * @param Height of the current compositor in normalized (0.0-1.0) coordinate.
 */
float glrcGetRenderHeight()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getHeight();
}

/**
 * Enable off-screen rendering (and composite processing) of the current compositor.
 * When disabled, both beginRender() / endRender() do nothing, so you can get usual rendering results.
 *
 * @param target Must be GLRC_COMPOSITOR_2D.
 */
void glrcEnable(unsigned int target)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}

    switch (target)
    {
    case GLRC_COMPOSITOR_2D:
		g_pComp->setEnabled(TRUE);
        break;
    case GLRC_BLEND:
		g_pComp->setBlendEnabled(TRUE);
        break;
    case GLRC_MESHDISTORTION:
		g_pComp->setMeshDistortion(TRUE);
        break;
    case GLRC_MODELVIEW:
		g_pComp->setModelViewMatrixEnabled(TRUE);
        break;
    case GLRC_PROJECTION:
		g_pComp->setProjectionMatrixEnabled(TRUE);
        break;
    }
}

/**
 * Disable off-screen rendering (and composite processing) of the current compositor.
 * When disabled, both beginRender() / endRender() do nothing, so you can get usual rendering results.
 *
 * @param target Must be GLRC_COMPOSITOR_2D.
 */
void glrcDisable(unsigned int target)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}

    switch (target)
    {
    case GLRC_COMPOSITOR_2D:
		g_pComp->setEnabled(FALSE);
        break;
    case GLRC_BLEND:
		g_pComp->setBlendEnabled(FALSE);
        break;
    case GLRC_MESHDISTORTION:
		g_pComp->setMeshDistortion(FALSE);
        break;
    case GLRC_MODELVIEW:
		g_pComp->setModelViewMatrixEnabled(FALSE);
        break;
    case GLRC_PROJECTION:
		g_pComp->setProjectionMatrixEnabled(FALSE);
        break;
    }
}

/**
 * Check whether off-screen rendering (and composite processing) of the current compositor is enabled or not.
 *
 * @param target Must be GLRC_COMPOSITOR_2D.
 * @return True if off-screen rendering (and composite processing) of the current compositor is enabled.
 */
int glrcIsEnabled(unsigned int target)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return FALSE;
	}

    switch (target)
    {
    case GLRC_COMPOSITOR_2D:
		return (g_pComp->getEnabled()==TRUE);
    case GLRC_BLEND:
		return (g_pComp->getBlendEnabled()==TRUE);
    case GLRC_MESHDISTORTION:
		return g_pComp->isMeshDistortion();
    case GLRC_MODELVIEW:
		return (g_pComp->getModelViewMatrixEnabled()==TRUE);
    case GLRC_PROJECTION:
		return (g_pComp->getProjectionMatrixEnabled()==TRUE);
    }
	return FALSE;
}

/**
 * Declare to begin to render your scene. Activate off-screen rendering to capture the rendering result,
 * and modify current ModelView / Projection matrices.
 */
void glrcBeginRender()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->beginRender();
}

/**
 * Declare to end the rendering. Deactivate off-screen rendering,
 * then render the current compositor object to on-screen (or other Compositor)
 * with composite processings.
 */
void glrcEndRender()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->endRender();
}

/**
 * Set rotation angle of the current compositor.
 *
 * @param angle Rotation angle in degree (0.0 is default).
 */
void glrcSetRotation(float angle)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setRotation(angle);
}

/**
 * Retrieve rotation angle of the current compositor.
 *
 * @return Rotation angle in degree (0.0 is default).
 */
float glrcGetRotation()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0.0f;
	}
	return g_pComp->getRotation();
}

/**
 * Set scale of the current compositor.
 *
 * @param x Horizontal scale of the current compositor (1.0 is default).
 * @param y Vertical scale of the current compositor (1.0 is default).
 */
void glrcSetScale(float x, float y)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setScale(x, y);
}

/**
 * Retrieve horizontal scale of the current compositor.
 *
 * @return Horizontal scale of the current compositor (1.0 is default).
 */
float glrcGetScaleX()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getScaleX();
}

/**
 * Retrieve vertical scale of the current compositor.
 *
 * @return Vertical scale of the current compositor (1.0 is default).
 */
float glrcGetScaleY()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getScaleY();
}

/**
 * Set flipping mode of the current compositor.
 *
 * @param flip Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
 */
void glrcSetFlip(unsigned int flip)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setFlip(flip);
}

/**
 * Retrieve flipping mode of the current compositor.
 *
 * @return Flip mode (combination of GLRC_FLIP_HORIZONTAl and GLRC_FLIP_VERTICAL, or GLRC_FLIP_NONE).
 */
unsigned int glrcGetFlip()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return GLRC_FLIP_NONE;
	}
	return g_pComp->getFlip();
}

/**
 * Set model-view modification matrix of the current compositor.
 *
 * @param matrix Model-view modification matrix.
 */
void glrcSetModelViewMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
    mat.set(matrix);
	g_pComp->setModelViewMatrix(mat);
}

/**
 * Retrieve model-view modification matrix of the current compositor.
 *
 * @return Model-view modification matrix.
 */
void glrcGetModelViewMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
	mat = g_pComp->getModelViewMatrix();
    mat.get(matrix);
}

/**
 * Retrieve translation of the model-view modification matrix of the current compositor.
 *
 * @param x Distance in x-axis direction.
 * @param y Distance in y-axis direction.
 * @param z Distance in z-axis direction.
 */
void glrcGetModelViewTranslation(float* x, float* y, float* z)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->getModelViewTranslation(*x, *y, *z);
}

/**
 * Retrieve rotation of the model-view modification matrix.
 *
 * @param angle Angle of rotation in degree.
 * @param x X element value of the axis vector.
 * @param y Y element value of the axis vector.
 * @param z Z element value of the axis vector.
 */
void glrcGetModelViewRotation(float* angle, float* x, float* y, float* z)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->getModelViewRotation(*angle, *x, *y, *z);
}

/**
 * Retrieve quatenion rotation of the model-view modification matrix.
 *
 * @param x X element of quatenion.
 * @param y Y element of quatenion.
 * @param z Z element of quatenion.
 * @param w W element of quatenion.
 */
void glrcGetModelViewQuatenion(float* x, float* y, float* z, float* w)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->getModelViewQuatenion(*x, *y, *z, *w);
}

/**
 * Retrieve scaling of the model-view modification matrix.
 *
 * @param x Scaling factor in x-axis.
 * @param y Scaling factor in y-axis.
 * @param z Scaling factor in z-axis.
 */
void glrcGetModelViewScaling(float* x, float* y, float* z)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->getModelViewScaling(*x, *y, *z);
}

/**
 * Set projection modification matrix of the current compositor.
 *
 * @param matrix Projection modification matrix.
 */
void glrcSetProjectionMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
    mat.set(matrix);
	g_pComp->setProjectionMatrix(mat);
}

/**
 * Retrieve projection modification matrix of the current compositor.
 *
 * @param matrix Projection modification matrix.
 */
void glrcGetProjectionMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
	mat = g_pComp->getProjectionMatrix();
    mat.get(matrix);
}

/**
 * Set projection modification matrix of the current compositor in glFrustum format.
 *
 * @param matrix Projection modification matrix in glFrustum format.
 */
void glrcSetProjectionFrustum(float frustum[6])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setProjectionFrustum(frustum);
}

/**
 * Retrieve projection modification matrix of the current compositor in glFrustum format.
 *
 * @param matrix Projection modification matrix in glFrustum format.
 */
void glrcGetProjectionFrustum(float frustum[6])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->getProjectionFrustum(frustum);
}

/**
 * Load matrix configuration file.
 *
 * @param fileName Matrix configuration file name.
 * @return 1 if loaded successfully.
 */
int glrcLoadMatrixFile(const char* fileName)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return false;
	}
	return g_pComp->loadMatrixFile(fileName)?1:0;
}

/**
 * Set effect processing mode to composite the rendering results.
 *
 * @param effectMode Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
 */
void glrcSetEffectMode(unsigned int effectMode)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setEffectMode(effectMode);
}

/**
 * Retrieve effect processing mode to composite the rendering results.
 *
 * @return Effect mode (combination of GLRC_EFFECT_* or GLRC_EFFECT_ALL/NONE).
 */
unsigned int glrcGetEffectMode()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return false;
	}
	return g_pComp->getEffectMode();
}

/**
 * Load a distortion map file to distort the rendering results.
 *
 * @param fileName Distortion map file name.
 * @return True if the distortion map is loaded successfully.
 */
void glrcSetDistortionMap(const char* fileName)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setDistortionMap(fileName);
}

/**
 * Retrieve the distortion map file name currently loaded.
 *
 * @return Distortion map file name currently loaded.
 */
const char* glrcGetDistortionMapFileName()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return "";
	}
	return g_pComp->getDistortionMapFileName();
}

/**
 * Set resolution of the distortion mesh.
 *
 * @param horResolution Horizontal resolution of the distortion mesh.
 * @param vertResolution Vertical resolution of the distortion mesh.
 */

void glrcSetDistortionMeshResolution(int horResolution, int vertResolution)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setDistortionMeshResolution(horResolution, vertResolution);
}

/**
 * Retrieve horizontal resolution of the distortion mesh.
 *
 * @return Horizontal resolution of the distortion mesh.
 */
int glrcGetHorizontalResolutionOfDisotrtionMesh()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return g_pComp->getHorizontalResolutionOfDisotrtionMesh();
}

/**
 * Retrieve vertical resolution of the distortion mesh.
 *
 * @return Vertical resolution of the distortion mesh.
 */
int glrcGetVerticalResolutionOfDisotrtionMesh()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 0;
	}
	return g_pComp->getVerticalResolutionOfDisotrtionMesh();
}

/**
 * Load a blend map file to distort the rendering results.
 *
 * @param fileName Blend map file name.
 * @return True if the blend map is loaded successfully.
 */
void glrcSetBlendMap(const char* fileName)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setBlendMap(fileName);
}

/**
 * Retrieve the blend map file name currently loaded.
 *
 * @return Blend map file name currently loaded.
 */
const char* glrcGetBlendMapFileName()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return "";
	}
	return g_pComp->getBlendMapFileName();
}

/**
 * Set gamma value for tone mapping of high-dynamic range rendering.
 * LDR = (HDR * exposure) ^ gamma.
 *
 * @param Gamma value (1.0 is default).
 */
void  glrcSetGamma(float gamma)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setGamma(gamma);
}

/**
 * Retrieve gamma value for high-dynamic range rendering.
 *
 * @return Gamma value (1.0 is default).
 */
float glrcGetGamma()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getGamma();
}

/**
 * Set exposure value for tone mapping of high-dynamic range rendering.
 * LDR = (HDR * exposure) ^ gamma.
 *
 * @param Exposure value (1.0 is default).
 */
void glrcSetExposure(float exposure)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
	g_pComp->setExposure(exposure);
}

/**
 * Retrieve exposure value for high-dynamic range rendering.
 *
 * @return Exposure value (1.0 is default).
 */
float glrcGetExposure()
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return 1.0f;
	}
	return g_pComp->getExposure();
}

/**
 * Set color conversion matrix.
 *
 * @param matrix Color conversion matrix.
 */
void glrcSetColorMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
    mat.set(matrix);
	g_pComp->setColorMatrix(mat);
}

/**
 * Retrieve color conversion matrix.
 *
 * @result Color conversion matrix.
 */
void glrcGetColorMatrix(float matrix[4][4])
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return;
	}
    glrc::Matrix mat;
    mat = g_pComp->getColorMatrix();
    mat.get(matrix);
}

/**
 * Save the current off-screen buffer to an image file.
 * Image file format is determined by its file extention.
 * Supported format is depending on DevIL library (and .hdr format).
 *
 * @param Image file name to save the off-screen buffer.
 * @return True if image is saved successfully.
 */
int glrcSaveBuffer(const char* fileName)
{
	if (!g_pComp) {
		glrcSetError(GLRC_ERROR_NOT_INITIALIZED);
		return false;
	}

    if (g_pComp->saveBuffer(fileName))
        return TRUE;
    return FALSE;
}

/**
 * Take a snapshot in specified size and save to an image file.
 * You have to provide 'complete' scene rendering function. 'Complete' means that
 * this function may create a new context for off-screen rendering (though sharing
 * display list and texture objects), so you should confirm about your contents.
 * At this time, width and height must be power of two!!
 *
 * @param fileName Image file name to save the taken snapshot.
 * @param width Width of the snapshot.
 * @param height Height of the snapshot.
 * @param bHasAlpha True if you need alpha channel.
 * @param bFPBuffer True if you need a floating point buffer.
 * @param pDisplayFunc 'Complete' rendering function like 'void display()'.
  */
int glrcSaveSnapshot(const char* fileName, int width, int height, int bHasAlpha, int bFPBuffer, void (*pDisplayFunc)(void))
{
    int result = FALSE;

    unsigned int ssIndex;
    glrcGenCompositors(1, &ssIndex);
    glrcBindCompositor(GLRC_COMPOSITOR_2D, ssIndex);
    glrcSetupBuffer(GLRC_RTT_AUTO, width, height, (bHasAlpha==TRUE), (bFPBuffer==TRUE));
    glrcBeginRender();

    if (pDisplayFunc)
        pDisplayFunc();

    glrcBindCompositor(GLRC_COMPOSITOR_2D, ssIndex);
    glrcEndRender();

    if (glrcSaveBuffer(fileName))
        result = TRUE;

    glrcDeleteCompositors(1, &ssIndex);

    return result;
}

/**
 * Set the latest error code. Used inside of GLRC library.
 *
 * @param error The latest error code (GLRC_ERROR_*).
 */
void glrcSetError(GLRC_ERROR error)
{
	g_lastError = error;
}

/**
 * Retrieve the latest error code.
 *
 * @return the latest error code (GLRC_ERROR_*).
 */
GLRC_ERROR glrcGetError()
{
	GLRC_ERROR error = g_lastError;
	g_lastError = GLRC_ERROR_NOERROR;
	return error;
}

/**
 * .hdr format image file loading function for DevIL library.
 *
 * @param fileName File name of the .hdr file to load.
 * @return Result code.
 */
static ILenum ILAPIENTRY LoadFunction_hdr(const ILstring fileName)
{
    int width, height;
    FILE* fp = fopen(fileName, "rb");

    if (fp == NULL)
        return IL_COULD_NOT_OPEN_FILE;

    RGBE_ReadHeader(fp, &width, &height, NULL);
    if (!ilTexImage(width, height, 1, 3, IL_RGB, IL_FLOAT, NULL)) {
        fclose(fp);
        return IL_INTERNAL_ERROR;
    }

    // Set the origin via the register functions.
    ilRegisterOrigin(IL_ORIGIN_UPPER_LEFT);

    RGBE_ReadPixels_RLE(fp, reinterpret_cast<float*>(ilGetData()), width, height);
    fclose(fp);

    return IL_NO_ERROR;
}

/**
 * .hdr format image file saving function for DevIL library.
 *
 * @param fileName File name of the .hdr file to save.
 * @return Result code.
 */
static ILenum ILAPIENTRY SaveFunction_hdr(const ILstring fileName)
{
    ILint width, height, type;
    ilGetIntegerv(IL_IMAGE_WIDTH, &width);
    ilGetIntegerv(IL_IMAGE_HEIGHT, &height);
    ilGetIntegerv(IL_IMAGE_TYPE, &type);
    if (type != IL_FLOAT)
        return IL_INTERNAL_ERROR;

    FILE* fp = fopen(fileName, "wb");
    if (fp == NULL)
        return IL_COULD_NOT_OPEN_FILE;
    RGBE_WriteHeader(fp, width, height, NULL);
    RGBE_WritePixels_RLE(fp, reinterpret_cast<float*>(ilGetData()), width ,height);
    fclose(fp);

    return IL_NO_ERROR;
}
