#ifndef _SCENE_H_
#define _SCENE_H_

#include <QGLViewer/frame.h>

#include "math/TransformMatrix.h"

namespace projection
{
class ProjectionModel;
class SceneContent;

/**
 * Scene class.
 */
class Scene : public QObject
{
    Q_OBJECT

public:

    /**
     * Constructor.
     *
     * @param pModel Projection model.
     */
    Scene(ProjectionModel* pModel);

    /**
     * Destructor.
     */
    virtual ~Scene();

    /**
     * Retrieve name of the scene content.
     *
     * @param index Index of the scene content.
     * @return Name of the scene content.
     */
    QString getContentName(int index) const;

    /**
     * Retrieve the number of the scene content.
     *
     * @return Number of the scene content.
     */
    unsigned int getNumContents() const;

    /**
     * Retrieve scene content object.
     *
     * @param shapeName Name of the scene content object.
     * @return Scene content object.
     */
    SceneContent* getContent(const QString& contentName) const;

    /**
     * Select scene content by name.
     *
     * @param contentName Name of the scene content to select.
     */
    void setCurrentContent(const QString& contentName);

    /**
     * Retrieve the name of the current scene content.
     *
     * @return Name of the current scene content.
     */
    QString getCurrentContentName() const;

    /**
     * Retrieve current scene content object.
     *
     * @return Current scene content object.
     */
    SceneContent* getCurrentContent() const;

    /**
     * Set viewpoint of the scene viewer.
     *
     * @param data Viewpoint as XML data.
     */
    void setCameraFrame(const QString& data);

    /**
     * Retrieve viewpoint of the scene viewer.
     *
     * @return Viewpoint as XML data.
     */
    const qglviewer::Frame& getCameraFrame() const { return m_cameraFrame; }

    /**
     * Reset viewpoint of the camera.
     */
    void resetCamera();

    /**
     * Set transform matrix of the scene.
     *
     * @param matrix Transform matrix of the scene.
     */
    void setMatrix(const TransformMatrix& matrix);

    /**
     * Retrieve transform matrix of the scene.
     *
     * @param Transform matrix of the scene.
     */
    const TransformMatrix& getMatrix() const { return m_transformMatrix; }

    /**
     * Initialize screen content. 
     */
    void init();

    /**
     * Draw the scene content.
     */
    void draw();

    /**
     * Show or hide the scene content in design view.
     *
     * @param bShowInDesignView True to show the scene content in design view.
     */
    void setShowInDesignView(bool bShowInDesignView);

    /**
     * Check whether the scene content is shown in design view.
     *
     * @return True if the scene content is shown in design view.
     */
    bool getShowInDesignView() { return m_bShowInDesignView; }

    /**
     * Set size of the scene for SceneViewer.
     *
     * @param size Size of the scene.
     */
    void setSize(float size);

    /**
     * Retrieve size of the scene.
     *
     * @return Size of the scene.
     */
    float getSize() const;

    /**
     * Enable or disable wire frame drawing of the scene content.
     *
     * @param bShowFrame True to draw as a wire frame model.
     */
    void setShowFrame(bool bShowFrame);

    /**
     * Check whether drawing the scene content as a wire frame model or not.
     *
     * @return True if the scene content is drawn as a wire frame model.
     */
    bool getShowFrame() { return m_bShowFrame; }

    /**
     * Show or hide a grid frame with the scene content.
     *
     * @param bShowGrid True to show a grid frame, false to hide.
     */
    void setShowGrid(bool bShowGrid);

    /**
     * Check whether the grid frame is shown or not.
     *
     * @return True if the grid frame is shown.
     */
    bool getShowGrid() { return m_bShowGrid; }

    /**
     * Set size of the grid frame.
     *
     * @param gridSize Size of the grid frame.
     */
    void setGridSize(float gridSize);

    /**
     * Retrieve the size of the grid frame.
     *
     * @return Size of the grid frame.
     */
    float getGridSize() { return m_gridSize; }

    /**
     * Notify to redraw the screen.
     */
    void notifyRedraw();

    /**
     * Restore the scene content from XML data.
     *
     * @param element Parent XML element of the scene content data.
     */
    void initFromDOMElement(const QDomElement& element);

    /**
     * Store the current scene content as XML data.
     *
     * @param name XML node name of the data.
     * @param doc XML document to store the data.
     * @return Current scene content data as XML data.
     */
    QDomElement domElement(const QString& name, QDomDocument& doc) const;

protected slots:

    /**
     * Update viewpoint data.
     *
     * @param frame Viewpoint data.
     */
    void updateCameraFrame(const qglviewer::Frame& frame);

protected:

    TransformMatrix m_transformMatrix;  //!< Transform matrix of the scene.
    qglviewer::Frame m_cameraFrame;     //!< View point.
    bool m_bShowInDesignView;           //!< True if the scene content is shown in design view.
    float m_size;                       //!< Size of the scene.
    bool m_bShowFrame;                  //!< True if the scene content is drawn as a wire frame model.
    bool m_bShowGrid;                   //!< True if the grid frame is shown.
    int m_gridSize;                     //!< Size of the grid frame.

	QList<SceneContent*> m_pContents;   //!< Scene content objects.
    SceneContent* m_pCurrentContent;    //!< Current scene content object.

    ProjectionModel* m_pModel;          //!< Projection model.
};

}; // projection

#endif // _SCENE_H_
