#ifndef _SCENE_MODEL_H_
#define _SCENE_MODEL_H_

#include "Scene/SceneContent.h"

#include "glm/glm.h"

namespace projection
{

/**
 * Model scene class.
 */
class SceneModel : public SceneContent
{
public:

    /**
     * Constructor.
     *
     * @param pScene Scene data.
     */
    SceneModel(Scene* pScene);

    /**
     * Destructor.
     */
    virtual ~SceneModel();

    /**
     * Retrieve screen content name.
     */
    QString getName() const { return "Model"; }

    /**
     * Initialize screen content.
     */
    virtual void init();

    /**
     * Draw the scene content.
     *
     * @param cameraFrame View point.
     */
    virtual void draw(const qglviewer::Frame& cameraFrame);

    /**
     * Load a model file.
     *
     * @param fileName File name of a model to load.
     */
    void setFileName(const QString& fileName);

    /**
     * Retrieve file name of the model.
     *
     * @return File name of the model.
     */
    QString getFileName() const;

    /**
     * Enable or disable back face culling.
     *
     * @param bEnable True to enable back face culling, false to disable.
     */
    void setBackFaceCulling(bool bEnable);

    /**
     * Checke whether back face culling is enabled or not.
     *
     * @param bEnable True to enable back face culling, false to disable.
     */
    bool getBackFaceCulling() const;

    /**
     * Enable or disable smoothing.
     *
     * @param bEnable True to enable smoothing, false to disable.
     */
    void setSmoothing(bool bEnable);

    /**
     * Checke whether smoothing is enabled or not.
     *
     * @param bEnable True to enable smoothing, false to disable.
     */
    bool getSmoothing() const;

    /**
     * Retrieve bounding box of the scene content.
     *
     * @param min One corner of the scene content.
     * @param max Another corner of the scene content.
     */
    void getBoundingBox(qglviewer::Vec& min, qglviewer::Vec& max);

    /**
     * Restore the scene content from XML data.
     *
     * @param element Parent XML element of the scene content data.
     */
    virtual void initFromDOMElement(const QDomElement& element);

    /**
     * Store the current scene content as XML data.
     *
     * @param name XML node name of the data.
     * @param doc XML document to store the data.
     * @return Current scene content data as XML data.
     */
    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;

protected:

    QString m_fileName;         //!< File name of the model.
    bool m_bBackFaceCulling;    //!< True to cull back face of the model.
    bool m_bSmoothing;          //!< True to smooth surface of the model.
    GLMmodel* m_pModel;         //!< glm model data structure.
    GLuint m_modelIndex;        //!< gml model display list index.
};

}; // projection

#endif // _SCENE_MODEL_H_
