#include "ProjectionModel.h"
#include "Scene.h"
#include "scene/SceneEmpty.h"
#include "scene/SceneTeapot.h"
#include "scene/SceneCubeMap.h"
#include "scene/SceneModel.h"
#include "RSync.h"

#include "GUIControler.h"

using namespace projection;

/**
 * Notify to redraw the scene content.
 */
void SceneContent::notifyRedraw()
{
    m_pScene->notifyRedraw();
}

/**
 * Constructor.
 *
 * @param pModel Projection model.
 */
Scene::Scene(ProjectionModel* pModel) : QObject()
{
    m_pModel = pModel;

    m_pContents.push_back(new SceneEmpty(this));
    m_pContents.push_back(new SceneTeapot(this));
    m_pContents.push_back(new SceneCubeMap(this));
    m_pContents.push_back(new SceneModel(this));
    m_pCurrentContent = m_pContents[1]; // Teapot

    m_bShowInDesignView = true;
    m_size = 1000.0f;
    m_bShowFrame = true;
    m_bShowGrid = true;
	m_gridSize = 5.0;

	qglviewer::Quaternion rot =
        qglviewer::Quaternion(0.431798, 0.149583, 0.349953, 0.817746).inverse();
	qglviewer::Vec pos = qglviewer::Vec(0.0f, 0.0f, 2.0f);
	pos = rot * pos;
	m_cameraFrame.setOrientation(rot);
	m_cameraFrame.setPosition(pos);
}

/**
 * Destructor.
 */
Scene::~Scene()
{
    while (m_pContents.count() > 0) {
        SceneContent* pContent = m_pContents.takeAt(0);
        if (pContent) delete pContent;
    }
}

/**
 * Retrieve name of the scene content.
 *
 * @param index Index of the scene content.
 * @return Name of the scene content.
 */
QString Scene::getContentName(int index) const
{
    if (index < m_pContents.size())
        return m_pContents[index]->getName();
    return "None";
}

/**
 * Retrieve the number of the scene content.
 *
 * @return Number of the scene content.
 */
unsigned int Scene::getNumContents() const
{
    return m_pContents.size();
}

/**
 * Retrieve scene content object.
 *
 * @param shapeName Name of the scene content object.
 * @return Scene content object.
 */
SceneContent* Scene::getContent(const QString& contentName) const
{
    for (int i=0; i<m_pContents.size(); ++i)
        if (m_pContents[i]->getName() == contentName)
            return m_pContents[i];
    return NULL;
}

/**
 * Select scene content by name.
 *
 * @param contentName Name of the scene content to select.
 */
void Scene::setCurrentContent(const QString& contentName)
{
    if (m_pCurrentContent && m_pCurrentContent->getName() == contentName)
        return;

    for (int i=0; i<m_pContents.size(); ++i) {
		if (m_pContents[i]->getName() == contentName) {
			m_pCurrentContent = m_pContents[i];
			notifyRedraw();
		}
	}
}

/**
 * Retrieve the name of the current scene content.
 *
 * @return Name of the current scene content.
 */
QString Scene::getCurrentContentName() const
{
    return m_pCurrentContent->getName();
}

/**
 * Retrieve current scene content object.
 *
 * @return Current scene content object.
 */
SceneContent* Scene::getCurrentContent() const
{
    return m_pCurrentContent;
}

/**
 * Set viewpoint of the scene viewer.
 *
 * @param data Viewpoint as XML data.
 */
void Scene::setCameraFrame(const QString& data)
{
    QDomDocument doc;
    doc.setContent(data);
    m_cameraFrame.initFromDOMElement(doc.firstChildElement("CameraFrame"));

    if (!m_pModel->getRSync()->isServer())
        m_pModel->updateViewAndOffscreens();
}

/**
 * Reset viewpoint of the camera.
 */
void Scene::resetCamera()
{
    m_cameraFrame.setPosition(qglviewer::Vec(0.0f, -0.00001f, 0.0f));
    m_pModel->getGUI()->reserSceneCamera();
}

/**
 * Set transform matrix of the scene.
 *
 * @param matrix Transform matrix of the scene.
 */
void Scene::setMatrix(const TransformMatrix& matrix)
{
    m_transformMatrix = matrix;
    m_pModel->updateViewAndOffscreens();
}

/**
 * Initialize screen content. 
 */
void Scene::init()
{
	// initialize contents when the content shown in the first time
    for (int i=0; i<m_pContents.size(); ++i)
		m_pContents[i]->init();
}

/**
 * Draw the scene content.
 */
void Scene::draw()
{
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    if (m_bShowFrame)
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    glPushMatrix();
    glMultMatrixd(m_transformMatrix.getMatrix());
    m_pCurrentContent->draw(m_cameraFrame);
    glPopMatrix();

    if (m_bShowFrame)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    glPopAttrib();
}

/**
 * Show or hide the scene content in design view.
 *
 * @param bShowInDesignView True to show the scene content in design view.
 */
void Scene::setShowInDesignView(bool bShowInDesignView)
{
    m_bShowInDesignView = bShowInDesignView;
    notifyRedraw();
}

/**
 * Enable or disable wire frame drawing of the scene content.
 *
 * @param bShowFrame True to draw as a wire frame model.
 */
void Scene::setShowFrame(bool bShowFrame)
{
    m_bShowFrame = bShowFrame;
    notifyRedraw();
}

/**
 * Set size of the scene for SceneViewer.
 *
 * @param size Size of the scene.
 */
void Scene::setSize(float size)
{
    m_size = size;
    m_pModel->getGUI()->setSceneSize(m_size);
}

/**
 * Retrieve size of the scene.
 *
 * @return Size of the scene.
 */
float Scene::getSize() const
{
/*
    qglviewer::Vec min, max;
    m_pCurrentContent->getBoundingBox(min, max);
	min = m_transformMatrix.getFrame().transformOf(min);
	max = m_transformMatrix.getFrame().transformOf(max);
    float minLength = sqrtf(min.x*min.x+min.y*min.y+min.z*min.z);
    float maxLength = sqrtf(max.x*max.x+max.y*max.y+max.z*max.z);
    return minLength>maxLength?minLength:maxLength;
*/
    return m_size;
}

/**
 * Show or hide a grid frame with the scene content.
 *
 * @param bShowGrid True to show a grid frame, false to hide.
 */
void Scene::setShowGrid(bool bShowGrid)
{
    m_bShowGrid = bShowGrid;
    notifyRedraw();
}

/**
 * Set size of the grid frame.
 *
 * @param gridSize Size of the grid frame.
 */
void Scene::setGridSize(float gridSize)
{
    m_gridSize = gridSize;
    notifyRedraw();
}

/**
 * Notify to redraw the screen.
 */
void Scene::notifyRedraw()
{
    if (m_pModel->getRSync()->isServer())
    {
        QDomDocument doc;
        doc.appendChild(domElement("Scene", doc));
        m_pModel->getRSync()->sendChanges(RSYNC_COMMAND_SCENE, doc.toString(0));
    }

    m_pModel->getGUI()->updateSceneViewer();
    m_pModel->updateViewAndOffscreens();
}

/**
 * Restore the scene content from XML data.
 *
 * @param element Parent XML element of the scene content data.
 */
void Scene::initFromDOMElement(const QDomElement& element)
{
    if (!element.isNull())
    {
        if (!element.firstChildElement("TransformMatrix").isNull())
		    m_transformMatrix.initFromDOMElement(element.firstChildElement("TransformMatrix"));
        setCurrentContent(element.attribute("contentName"));
        m_bShowInDesignView = element.attribute("showInDesignView").toLower()=="true";
        m_size = element.attribute("size").toFloat();
        m_bShowFrame = element.attribute("showFrame").toLower()=="true";
        m_cameraFrame.initFromDOMElement(element.firstChildElement("CameraFrame"));

        QDomElement contents = element.firstChildElement("Contents");
        QDomElement content = contents.firstChild().toElement();
        while (!content.isNull()) {
            SceneContent* pContent = getContent(content.attribute("name"));
            if (pContent)
                pContent->initFromDOMElement(content);
            content = content.nextSiblingElement();
        }
    }
}

/**
 * Store the current scene content as XML data.
 *
 * @param name XML node name of the data.
 * @param doc XML document to store the data.
 * @return Current scene content data as XML data.
 */
QDomElement Scene::domElement(const QString& name, QDomDocument& doc) const
{
    QDomElement de = doc.createElement(name);
	de.appendChild(m_transformMatrix.domElement("TransformMatrix", doc));
    de.setAttribute("contentName", getCurrentContentName());
    de.setAttribute("showInDesignView", (m_bShowInDesignView?"true":"false"));
    de.setAttribute("size", m_size);
    de.setAttribute("showFrame", (m_bShowFrame?"true":"false"));
    de.appendChild(m_cameraFrame.domElement("CameraFrame", doc));

    QDomElement contents = doc.createElement("Contents");
    for (int i=0; i<m_pContents.size(); ++i)
        contents.appendChild(m_pContents[i]->domElement("Content", doc));
    de.appendChild(contents);

    return de;
}

/**
 * Update viewpoint data.
 *
 * @param frame Viewpoint data.
 */
void Scene::updateCameraFrame(const qglviewer::Frame& frame)
{
    m_cameraFrame = frame;

    if (m_pModel->getRSync()->isServer())
    {
        QDomDocument doc;
        doc.appendChild(m_cameraFrame.domElement("CameraFrame", doc));
        m_pModel->getRSync()->sendChanges(RSYNC_COMMAND_VIEWPOINT, doc.toString(0));
    }

    m_pModel->updateViewAndOffscreens();
}
