#include "ProjectionModel.h"
#include <QMessageBox>

#include "Scene.h"
#include "Screen.h"
#include "Channel.h"
#include "gui/QDesignViewWidget.h"

#include "glrc/glrc.h"

using namespace projection;

QDesignViewWidget::QDesignViewWidget(QWidget* pParent, const QGLWidget* pSharedWidget, Qt::WFlags flags)
    : QGLViewer(pParent, pSharedWidget, flags)
{
	m_pModel = NULL;

//	setShortcut(DRAW_AXIS, 0);
	setShortcut(DRAW_GRID, 0);
//	setShortcut(DISPLAY_FPS, 0);
	setShortcut(ENABLE_TEXT, 0);
	setShortcut(EXIT_VIEWER, 0);
	setShortcut(SAVE_SCREENSHOT, 0);
	setShortcut(CAMERA_MODE, 0);
	setShortcut(FULL_SCREEN, 0);
	setShortcut(STEREO, 0);
	setShortcut(ANIMATION, 0);
	setShortcut(HELP, 0);
	setShortcut(EDIT_CAMERA, 0);
	setShortcut(MOVE_CAMERA_LEFT, 0);
	setShortcut(MOVE_CAMERA_RIGHT, 0);
	setShortcut(MOVE_CAMERA_UP, 0);
	setShortcut(MOVE_CAMERA_DOWN, 0);
	setShortcut(INCREASE_FLYSPEED, 0);
	setShortcut(DECREASE_FLYSPEED, 0);
    setStateFileName(QString::null);

    m_viewMode = DESIGN_VIEWMODE_PERSPECTIVE;
    m_opMode = DESIGN_OPMODE_PROJECTOR;
    m_transMode = DESIGN_TRANSMODE_LOCAL;

    m_bGridFrameIsDrawn = true;
	m_gridFrameSize = 1.0f;
	m_gridFrameSubdivisions = 10;
    m_bInfoIsDrawn = true;

    m_infoFont.setFamily("Helvetica");
    m_infoFont.setPointSize(14);
}

void QDesignViewWidget::init()
{
    glrcInit();

    setBackgroundColor(QColor(64, 64, 64));

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

QDesignViewWidget::~QDesignViewWidget()
{
}

void QDesignViewWidget::setModel(ProjectionModel* pModel)
{
    m_pModel = pModel;

    camera()->setSceneRadius(m_pModel->getScreen()->getSize());
    camera()->setZClippingCoefficient(10.0f);
    camera()->setOrientation(qglviewer::Quaternion(-0.272906, 0.353225, 0.141683, 0.883561));
    camera()->showEntireScene();
}

void QDesignViewWidget::draw()
{
	if (!isVisible()) return;

    glViewport(0, 0, width(), height());

    if (m_pModel->getScene()->getShowInDesignView())
    {
	    glPushMatrix();
	    glMultMatrixd(m_pModel->getScene()->getCameraFrame().inverse().matrix());
	    m_pModel->getScene()->draw();
	    glPopMatrix();
    }

    glPushMatrix();
    m_pModel->getScreen()->draw();
    m_pModel->getScreen()->drawFrame();
    for (unsigned int i=0; i<m_pModel->getNumChannels(); ++i)
    {
        if (m_pModel->getChannel(i)->getProjector()->getShowArea())
            m_pModel->getScreen()->draw(PROJECT_SOURCE_PROJECTORAREA, m_pModel->getChannel(i));
        if (m_pModel->getChannel(i)->getView()->getShowArea())
            m_pModel->getScreen()->draw(PROJECT_SOURCE_VIEWAREA, m_pModel->getChannel(i));
        if (m_pModel->getChannel(i)->getShowProjectedScene())
            m_pModel->getScreen()->draw(PROJECT_SOURCE_SCENE, m_pModel->getChannel(i));
    }
    glPopMatrix();

    for (unsigned int i=0; i<m_pModel->getNumChannels(); ++i)
    {
        glPushMatrix();
        m_pModel->getChannel(i)->drawFrustums(m_pModel->getChannel(i)==m_pModel->getSelectedChannel());
        glPopMatrix();
    }

    glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);
	for (unsigned int i=0; i<m_pModel->getNumChannels(); ++i)
    {
        if (m_pModel->getChannel(i)->getProjector()->getShow())
        {
            glPushMatrix();
            glMultMatrixd(m_pModel->getChannel(i)->getProjector()->getTransformMatrix().getMatrix());
	        glTranslatef(0.0f, 0.0f, 0.05f);
	        renderText(0, 0, 0, m_pModel->getChannel(i)->getName(), m_infoFont, 5000);
	        glPopMatrix();
        }
        if (m_pModel->getChannel(i)->getView()->getShow())
        {
            glPushMatrix();
            glMultMatrixd(m_pModel->getChannel(i)->getView()->getTransformMatrix().getMatrix());
	        glTranslatef(0.0f, 0.0f, 0.05f);
	        renderText(0, 0, 0, m_pModel->getChannel(i)->getName(), m_infoFont, 5000);
	        glPopMatrix();
        }
    }
	glPopAttrib();

	if (m_bGridFrameIsDrawn)
	{
        glPushMatrix();
        glRotatef(90.0f, 1.0f, 0.0f, 0.0f); // X-Y plane to X-Z plane
		glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT);
		glDisable(GL_LIGHTING);
		glColor3f(0.5f, 0.5f, 0.5f);
		drawGrid(m_gridFrameSize, m_gridFrameSubdivisions);
		glPopAttrib();
        glPopMatrix();
	}

    if (m_bInfoIsDrawn)
    {
        int x = m_infoFont.pointSize() * 0.5;
        int y = m_infoFont.pointSize() * 1.5;
        int yoffset = m_infoFont.pointSize() * 1.5;
        startScreenCoordinatesSystem();
        switch (m_viewMode)
        {
        case DESIGN_VIEWMODE_PERSPECTIVE:
            drawText(x, y, QString("View Mode: Perspective"), m_infoFont); y += yoffset; break;
        case DESIGN_VIEWMODE_CAMERA:
            drawText(x, y, QString("View Mode: Camera"), m_infoFont); y += yoffset; break;
        case DESIGN_VIEWMODE_TOP:
            drawText(x, y, QString("View Mode: Top"), m_infoFont); y += yoffset; break;
        case DESIGN_VIEWMODE_SIDE:
            drawText(x, y, QString("View Mode: Side"), m_infoFont); y += yoffset; break;
        case DESIGN_VIEWMODE_FRONT:
            drawText(x, y, QString("View Mode: Front"), m_infoFont); y += yoffset; break;
        }
        drawText(x, y, QString("Operation Mode: %1").arg(m_opMode==DESIGN_OPMODE_PROJECTOR?"Projector":"View"), m_infoFont); y += yoffset;
        drawText(x, y, QString("Translation Mode: %1").arg(m_transMode==DESIGN_TRANSMODE_LOCAL?"Local":"World"), m_infoFont); y += yoffset;
        if (m_pModel->getSelectedChannel())
            drawText(x, y, QString("Channel: %1").arg(m_pModel->getSelectedChannel()->getName()), m_infoFont); y += yoffset;
        stopScreenCoordinatesSystem();

#ifdef _DEBUG
        drawText(x, y, QString("UpdateViews: %1").arg(m_pModel->m_updateViewsCount), m_infoFont); y += yoffset;
        drawText(x, y, QString("UpdateOffscreen: %1").arg(m_pModel->m_updateOffscreenCount), m_infoFont); y += yoffset;
#endif // _DEBUG
    }
}

void QDesignViewWidget::setViewMode(DESIGN_VIEWMODE viewMode)
{
    if (m_viewMode == viewMode)
        return;

    if (viewMode == DESIGN_VIEWMODE_PERSPECTIVE)
	{
		camera()->playPath(0);
		camera()->setType(qglviewer::Camera::PERSPECTIVE);
	}
    else
    {
		camera()->deletePath(0);
		camera()->addKeyFrameToPath(0);

        if (viewMode == DESIGN_VIEWMODE_CAMERA)
        {
	        camera()->setType(qglviewer::Camera::PERSPECTIVE);
	        qglviewer::Quaternion quat(qglviewer::Vec(1.0, 0.0, 0.0), 0.0);
	        camera()->setOrientation(quat);
	        camera()->setPosition(qglviewer::Vec(0.0f, -0.00001f, 0.0f));
        }
        else
        {
	        camera()->setType(qglviewer::Camera::ORTHOGRAPHIC);
            switch (viewMode)
            {
            case DESIGN_VIEWMODE_TOP:
                {
	                qglviewer::Quaternion quat(qglviewer::Vec(1.0, 0.0, 0.0), -M_PI/2.0);
	                camera()->setOrientation(quat);
                }
                break;
            case DESIGN_VIEWMODE_SIDE:
                {
	                qglviewer::Quaternion quat(qglviewer::Vec(0.0, 1.0, 0.0), M_PI/2.0);
	                camera()->setOrientation(quat);
                }
                break;
            case DESIGN_VIEWMODE_FRONT:
                {
	                qglviewer::Quaternion quat(qglviewer::Vec(1.0, 0.0, 0.0), 0.0);
	                camera()->setOrientation(quat);
                }
                break;
            }
	        camera()->showEntireScene();
        }
    }

    m_viewMode = viewMode;

    updateGL();
}

DESIGN_VIEWMODE QDesignViewWidget::getViewMode() const
{
    return m_viewMode;
}

void QDesignViewWidget::setOpMode(DESIGN_OPMODE opMode)
{
    m_opMode = opMode;
	updateGL();
}

DESIGN_OPMODE QDesignViewWidget::getOpMode() const
{
    return m_opMode;
}

void QDesignViewWidget::setTransMode(DESIGN_TRANSMODE transMode)
{
	m_transMode = transMode;
	updateGL();
}

DESIGN_TRANSMODE QDesignViewWidget::getTransMode() const
{
	return m_transMode;
}

void QDesignViewWidget::setGridFrameIsDrawn(bool drawn)
{
	m_bGridFrameIsDrawn = drawn;
	updateGL();
    gridIsDrawnChanged(drawn);
}

bool QDesignViewWidget::gridFrameIsDrawn() const
{
	return m_bGridFrameIsDrawn;
}

void QDesignViewWidget::setGridFrameSize(float size)
{
	m_gridFrameSize = size;
	updateGL();
}

float QDesignViewWidget::getGridFrameSize() const
{
	return m_gridFrameSize;
}

void QDesignViewWidget::setGridFrameSubdivisions(int sub)
{
	m_gridFrameSubdivisions = sub;
	updateGL();
}

int QDesignViewWidget::getGridFrameSubdivisions() const
{
	return m_gridFrameSubdivisions;
}

void QDesignViewWidget::setInfoIsDrawn(bool drawn)
{
	m_bInfoIsDrawn = drawn;
	updateGL();
    infoIsDisplayedChanged(drawn);
}

bool QDesignViewWidget::infoIsDrawn() const
{
	return m_bInfoIsDrawn;
}

void QDesignViewWidget::keyReleaseEvent(QKeyEvent* pEvent)
{
    QGLViewer::keyReleaseEvent(pEvent);

    switch (pEvent->key())
    {
	case Qt::Key_G:
        setGridFrameIsDrawn(!gridFrameIsDrawn());
		break;
	case Qt::Key_I:
        setInfoIsDrawn(!infoIsDrawn());
		break;
	case Qt::Key_F1:
		setViewMode(DESIGN_VIEWMODE_PERSPECTIVE);
		break;
	case Qt::Key_F2:
		setViewMode(DESIGN_VIEWMODE_CAMERA);
		break;
	case Qt::Key_F3:
		setViewMode(DESIGN_VIEWMODE_TOP);
		break;
	case Qt::Key_F4:
		setViewMode(DESIGN_VIEWMODE_SIDE);
		break;
	case Qt::Key_F5:
		setViewMode(DESIGN_VIEWMODE_FRONT);
		break;
	case Qt::Key_Space:
		m_pModel->selectNextChannel();
		break;
    default:
        break;
	}

    if (!m_pModel->getSelectedChannel())
        return;

    processChannelKeyEvent(m_pModel->getSelectedChannel(), pEvent);
}

void QDesignViewWidget::processChannelKeyEvent(Channel* pChannel, QKeyEvent* pEvent)
{
    if (!pChannel)
        return;

    bool bShift = (pEvent->modifiers() & Qt::ShiftModifier);
#ifdef WIN32
    bShift = (GetAsyncKeyState(VK_SHIFT)<0);
#endif // WIN32
    bool bCtrl = (pEvent->modifiers() & Qt::ControlModifier);
//  bool bAlt = (pEvent->modifiers() & Qt::AltModifier);
    bool bKeypad = (pEvent->modifiers() & Qt::KeypadModifier);
    bool bCapsLock = false;
#ifdef WIN32
    bCapsLock = (GetAsyncKeyState(VK_CAPITAL) < 0);
#endif // WIN32

    float step = 1.0f;
	if (bShift && bCtrl)
		step = 0.01f;
	else if (bShift || bCtrl)
		step = 0.1f;

	Frustum* pFrustum = pChannel->getProjector();
	if (m_opMode == DESIGN_OPMODE_VIEW)
		pFrustum = pChannel->getView();

	ProjectionMatrix projMatrix = pFrustum->getProjectionMatrix();
	TransformMatrix xformMatrix = pFrustum->getTransformMatrix();
    qglviewer::Vec pos = xformMatrix.getPosition();
	qglviewer::Vec rot = xformMatrix.getRotation();
    qglviewer::Quaternion quatRot = xformMatrix.getFrame().rotation();

    if (bKeypad)
    {
        switch (pEvent->key())
        {
	    case Qt::Key_8:
	    case Qt::Key_Up:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
                xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1], pos[2]-0.1f*step));
            else
    		    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(0.0f, 0.0f, -0.1f*step));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_2:
	    case Qt::Key_Down:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
                xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1], pos[2]+0.1f*step));
            else
    		    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(0.0f, 0.0f, 0.1f*step));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_6:
	    case Qt::Key_Right:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
    		    xformMatrix.setPosition(qglviewer::Vec(pos[0]+0.1f*step, pos[1], pos[2]));
            else
	    	    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(0.1f*step, 0.0f, 0.0f));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_4:
	    case Qt::Key_Left:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
                xformMatrix.setPosition(qglviewer::Vec(pos[0]-0.1f*step, pos[1], pos[2]));
            else
    		    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(-0.1f*step, 0.0f, 0.0f));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_9:
	    case Qt::Key_PageUp:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
                xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1]+0.1f*step, pos[2]));
            else
    		    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(0.0f, 0.1f*step, 0.0f));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_3:
	    case Qt::Key_PageDown:
            if (m_transMode == DESIGN_TRANSMODE_WORLD)
                xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1]-0.1f*step, pos[2]));
            else
    		    xformMatrix.setPosition(pos + quatRot * qglviewer::Vec(0.0f, -0.1f*step, 0.0f));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Plus:
			projMatrix.setOffaxisY(projMatrix.getOffaxisY()-0.01f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
            break;
	    case Qt::Key_Minus:
			projMatrix.setOffaxisY(projMatrix.getOffaxisY()+0.01f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
            break;
	    case Qt::Key_Asterisk:
			projMatrix.setOffaxisX(projMatrix.getOffaxisX()+0.01f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
            break;
	    case Qt::Key_Slash:
			projMatrix.setOffaxisX(projMatrix.getOffaxisX()-0.01f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
            break;
        }
    }
    else
    {
        switch (pEvent->key())
	    {
	    case Qt::Key_PageUp:
		    xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1]+0.1f*step, pos[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_PageDown:
		    xformMatrix.setPosition(qglviewer::Vec(pos[0], pos[1]-0.1f*step, pos[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Left:
			xformMatrix.setRotation(qglviewer::Vec(rot[0]+1.0f*step, rot[1], rot[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Right:
			xformMatrix.setRotation(qglviewer::Vec(rot[0]-1.0f*step, rot[1], rot[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Up:
		    xformMatrix.setRotation(qglviewer::Vec(rot[0], rot[1]+1.0f*step, rot[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Down:
		    xformMatrix.setRotation(qglviewer::Vec(rot[0], rot[1]-1.0f*step, rot[2]));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Home:
		    xformMatrix.setRotation(qglviewer::Vec(rot[0], rot[1], rot[2]-1.0f*step));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_End:
		    xformMatrix.setRotation(qglviewer::Vec(rot[0], rot[1], rot[2]+1.0f*step));
		    pFrustum->setTransformMatrix(xformMatrix);
		    break;
	    case Qt::Key_Comma:
			projMatrix.setFOV(projMatrix.getFOV()-1.0f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
            break;
	    case Qt::Key_Period:
			projMatrix.setFOV(projMatrix.getFOV()+1.0f*step);
		    pFrustum->setProjectionMatrix(projMatrix);
		    break;
	    case Qt::Key_Q:
			if (m_opMode == DESIGN_OPMODE_PROJECTOR)
                setOpMode(DESIGN_OPMODE_VIEW);
            else
                setOpMode(DESIGN_OPMODE_PROJECTOR);
		    break;
	    case Qt::Key_T:
			if (m_transMode == DESIGN_TRANSMODE_LOCAL)
                setTransMode(DESIGN_TRANSMODE_WORLD);
            else
                setTransMode(DESIGN_TRANSMODE_LOCAL);
		    break;
        }
    }
}

void QDesignViewWidget::initFromDOMElement(const QDomElement& element)
{
	camera()->deletePath(0);

    QDomElement elementWithoutGeom = element;
    if (!elementWithoutGeom.firstChildElement("Geometry").isNull())
        elementWithoutGeom.removeChild(elementWithoutGeom.firstChildElement("Geometry"));

	QGLViewer::initFromDOMElement(elementWithoutGeom);

    QDomElement display = element.firstChildElement("Display");
	if (!display.isNull()) {
        QString viewMode = display.attribute("viewMode", "projector");
        if (viewMode == "perspective")
            m_viewMode = DESIGN_VIEWMODE_PERSPECTIVE;
        else if (viewMode == "camera")
            m_viewMode = DESIGN_VIEWMODE_CAMERA;
        else if (viewMode == "top")
            m_viewMode = DESIGN_VIEWMODE_TOP;
        else if (viewMode == "side")
            m_viewMode = DESIGN_VIEWMODE_SIDE;
        else if (viewMode == "front")
            m_viewMode = DESIGN_VIEWMODE_FRONT;
        if (display.attribute("opMode", "projector")=="projector")
    		m_opMode = DESIGN_OPMODE_PROJECTOR;
        else
    		m_opMode = DESIGN_OPMODE_VIEW;
        if (display.attribute("transMode", "local")=="local")
    		m_transMode = DESIGN_TRANSMODE_LOCAL;
        else
    		m_transMode = DESIGN_TRANSMODE_WORLD;
		m_bGridFrameIsDrawn = (display.attribute("gridFrameIsDrawn", "true")=="true");
		m_gridFrameSize = display.attribute("gridFrameSize", "1.0").toFloat();
		m_gridFrameSubdivisions = display.attribute("gridFrameSubdivisions", "10").toInt();
		m_bInfoIsDrawn = (display.attribute("infoIsDrawn", "true")=="true");
	}
}

QDomElement QDesignViewWidget::domElement(QDomDocument& doc) const
{
    QDomElement de = QGLViewer::domElement("DesignView", doc);
	QDomElement display = de.firstChildElement("Display");
	if (!display.isNull()) {
        switch (m_viewMode) {
        case DESIGN_VIEWMODE_PERSPECTIVE: display.setAttribute("viewMode", "perspective"); break;
        case DESIGN_VIEWMODE_CAMERA:      display.setAttribute("viewMode", "camera"); break;
        case DESIGN_VIEWMODE_TOP:         display.setAttribute("viewMode", "top"); break;
        case DESIGN_VIEWMODE_SIDE:        display.setAttribute("viewMode", "side"); break;
        case DESIGN_VIEWMODE_FRONT:       display.setAttribute("viewMode", "front"); break;
        }
		display.setAttribute("opMode", m_opMode==DESIGN_OPMODE_PROJECTOR?"projector":"view");
		display.setAttribute("transMode", m_transMode==DESIGN_TRANSMODE_LOCAL?"local":"world");
		display.setAttribute("gridFrameIsDrawn", m_bGridFrameIsDrawn?"true":"false");
		display.setAttribute("gridFrameSize", QString::number(m_gridFrameSize));
		display.setAttribute("gridFrameSubdivisions", QString::number(m_gridFrameSubdivisions));
		display.setAttribute("infoIsDrawn", m_bInfoIsDrawn?"true":"false");
	}
	return de;
}
