#include "math/TransformMatrix.h"

#include <QGLViewer/quaternion.h>
#include <QGLViewer/vec.h>

#define RAD(x) (((x)*2.0f*M_PI/360.0f)-2*M_PI*((int)((x)/360.0f))) // Deg -> Rad

using namespace projection;

/**
 * Constructor.
 *
 * Default transform matrix value is identity matrix.
 */
TransformMatrix::TransformMatrix()
{
    m_pos.setValue(0.0f, 0.0f, 0.0f);
    m_rot.setValue(0.0f, 0.0f, 0.0f);
    m_scale.setValue(1.0f, 1.0f, 1.0f);
    m_bUseScale = true;
}

/**
 * Destructor.
 */
TransformMatrix::~TransformMatrix()
{
}

/**
 * Retrieve the trasform matrix as a qglviewer::Frame. No scaling!
 *
 * @return Trasform matrix as a qglviewer::Frame.
 */
qglviewer::Frame TransformMatrix::getFrame() const
{
    qglviewer::Frame frame = qglviewer::Frame(qglviewer::Vec(m_pos[0], m_pos[1], m_pos[2]),
						                      qglviewer::Quaternion(qglviewer::Vec(0.0f, 1.0f, 0.0f), RAD(m_rot[0])) *
						                      qglviewer::Quaternion(qglviewer::Vec(1.0f, 0.0f, 0.0f), RAD(m_rot[1])) *
						                      qglviewer::Quaternion(qglviewer::Vec(0.0f, 0.0f, 1.0f), RAD(m_rot[2])));
	return frame;
}

double* TransformMatrix::getMatrix()
{
    qglviewer::Frame frame = getFrame();
    double mat[4][4];
    unsigned int i;
    frame.rotation().getMatrix(mat);
    for (i=0; i<16; ++i)
        m_mat[i] = mat[i/4][i%4];
    for (i=0; i<16; ++i)
        m_mat[i] = mat[i/4][i%4];
    m_mat[12] = frame.position()[0];
    m_mat[13] = frame.position()[1];
    m_mat[14] = frame.position()[2];
    double scale[] = { m_scale[0], m_scale[1], m_scale[2], 1.0 };
    for (i=0; i<16; ++i)
        m_mat[i] = m_mat[i] * scale[i%4];

    return m_mat;
}

/**
 * Retrieve transform parameters like 'Rotate 45 1 0 0;'.
 *
 * @param numSpaces Number of space characters for indent.
 * @return Transform params.
 */
QString TransformMatrix::getParams(int space) const
{
    QString spc;
    spc.fill(' ', space);

    QString result;
    if (m_scale != qglviewer::Vec(1.0, 1.0, 1.0))
        result += spc + QString("Scale %1 %2 %3;\n").arg(m_scale[0]).arg(m_scale[1]).arg(m_scale[2]);
    if (m_rot[2] != 0.0)
        result += spc + QString("Rotate %1 0 0 1;\n").arg(m_rot[2]);
    if (m_rot[1] != 0.0)
        result += spc + QString("Rotate %1 1 0 0;\n").arg(m_rot[1]);
    if (m_rot[0] != 0.0)
        result += spc + QString("Rotate %1 0 1 0;\n").arg(m_rot[0]);
    if (m_pos != qglviewer::Vec(0.0, 0.0, 0.0))
        result += spc + QString("Translate %1 %2 %3;\n").arg(m_pos[0]).arg(m_pos[1]).arg(m_pos[2]);

    return result;
}

/**
 * Restore the transform matrix from XML data.
 *
 * @param element Parent XML element of the transform matrix data.
 */
void TransformMatrix::initFromDOMElement(const QDomElement& element)
{
	m_pos.initFromDOMElement(element.firstChildElement("position"));
	QDomElement rot = element.firstChildElement("rotation");
	m_rot[0] = rot.attribute("h").toFloat();
	m_rot[1] = rot.attribute("p").toFloat();
	m_rot[2] = rot.attribute("r").toFloat();
    if (!element.firstChildElement("scaling").isNull())
	    m_scale.initFromDOMElement(element.firstChildElement("scaling"));
}

/**
 * Store the current transform matrix as XML data.
 *
 * @param name XML node name of the data.
 * @param doc XML document to store the data.
 * @return Current transform matrix data as XML data.
 */
QDomElement TransformMatrix::domElement(const QString& name, QDomDocument& doc) const
{
	QDomElement de = doc.createElement(name);

	de.appendChild(m_pos.domElement("position", doc));

	QDomElement rot = doc.createElement("rotation");
	rot.setAttribute("h", QString::number(m_rot[0]));
	rot.setAttribute("p", QString::number(m_rot[1]));
	rot.setAttribute("r", QString::number(m_rot[2]));
	de.appendChild(rot);

    de.appendChild(m_scale.domElement("scaling", doc));

	return de;
}
