#include <stdio.h>
#include <stdlib.h>

#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "Scene.h"
#include "scene/SceneCubeMap.h"

using namespace projection;

#define SIZE 1.0f
#define TEXCOORD_MIN 0.001f
#define TEXCOORD_MAX 0.999f

/**
 * Constructor.
 *
 * @param pScene Scene data.
 */
SceneCubeMap::SceneCubeMap(Scene* pScene) : SceneContent(pScene)
{
    for (unsigned int i=0; i<6; ++i)
        m_textureIndex[i] = 0;
}

/**
 * Destructor.
 */
SceneCubeMap::~SceneCubeMap()
{
    for (unsigned int i=0; i<6; ++i)
        if (glIsTexture(m_textureIndex[i]))
            glDeleteTextures(1, &m_textureIndex[i]);
}

/**
 * Initialize screen content.
 */
void SceneCubeMap::init()
{
}

/**
 * Draw the scene content.
 *
 * @param cameraFrame View point data.
 */
void SceneCubeMap::draw(const qglviewer::Frame&)
{
    glPushMatrix();

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    for (unsigned int i=0; i<6; ++i) {
        if (!glIsTexture(m_textureIndex[i]) && !m_textureFileNames[i].isEmpty())
        {
            m_textureIndex[i] = ilutGLLoadImage((ILstring)m_textureFileNames[i].toLocal8Bit().data());
        }
    }

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_FRONT]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f(-SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f( SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f(-SIZE,  SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f( SIZE,  SIZE, -SIZE);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_BACK]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f(-SIZE, -SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f(-SIZE,  SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f( SIZE, -SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f( SIZE,  SIZE,  SIZE);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_LEFT]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f(-SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f(-SIZE, -SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f(-SIZE,  SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f(-SIZE,  SIZE,  SIZE);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_RIGHT]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f( SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f( SIZE,  SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f( SIZE, -SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f( SIZE,  SIZE,  SIZE);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_TOP]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f(-SIZE,  SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f( SIZE,  SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f(-SIZE,  SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f( SIZE,  SIZE,  SIZE);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, m_textureIndex[(int)CUBEMAP_FACE_BOTTOM]);
    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MIN); glVertex3f(-SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MIN, TEXCOORD_MAX); glVertex3f(-SIZE, -SIZE,  SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MIN); glVertex3f( SIZE, -SIZE, -SIZE);
    glTexCoord2f(TEXCOORD_MAX, TEXCOORD_MAX); glVertex3f( SIZE, -SIZE,  SIZE);
    glEnd();

    glPopAttrib();

    glPopMatrix();
}

/**
 * Retrieve bounding box of the scene content.
 *
 * @param min One corner of the scene content.
 * @param max Another corner of the scene content.
 */
void SceneCubeMap::getBoundingBox(qglviewer::Vec& min, qglviewer::Vec& max)
{
    min.setValue(-SIZE, -SIZE, -SIZE);
    max.setValue( SIZE,  SIZE,  SIZE);
}

/**
 * Set texture to a face.
 *
 * @param face Face to set the texture.
 * @param fileName Texture file name.
 */
void SceneCubeMap::setTexture(CUBEMAP_FACE face, const QString& fileName)
{
    if ((int)face < 0 || (int)face > 5) return;
    if (m_textureFileNames[(int)face] == fileName) return;

    if (glIsTexture(m_textureIndex[(int)face]))
        glDeleteTextures(1, &m_textureIndex[(int)face]);
    m_textureIndex[(int)face] = 0;
    m_textureFileNames[(int)face] = fileName;

	notifyRedraw();
}

/**
 * Retrieve texture file name of the specified face.
 *
 * @param face Face to retrieve the texture file name
 * @return Texture file name.
 */
QString SceneCubeMap::getTexture(CUBEMAP_FACE face) const
{
    if ((int)face < 0 || (int)face > 5) return QString::null;

    return m_textureFileNames[(int)face];
}

/**
 * Restore the scene content from XML data.
 *
 * @param element Parent XML element of the scene content data.
 */
void SceneCubeMap::initFromDOMElement(const QDomElement& element)
{
    if (!element.isNull())
    {
        QDomElement texture = element.firstChildElement("texture");
        while (!texture.isNull())
        {
            if (texture.attribute("face") == "front")
                setTexture(CUBEMAP_FACE_FRONT, texture.attribute("fileName"));
            else if (texture.attribute("face") == "back")
                setTexture(CUBEMAP_FACE_BACK, texture.attribute("fileName"));
            else if (texture.attribute("face") == "left")
                setTexture(CUBEMAP_FACE_LEFT, texture.attribute("fileName"));
            else if (texture.attribute("face") == "right")
                setTexture(CUBEMAP_FACE_RIGHT, texture.attribute("fileName"));
            else if (texture.attribute("face") == "top")
                setTexture(CUBEMAP_FACE_TOP, texture.attribute("fileName"));
            else if (texture.attribute("face") == "bottom")
                setTexture(CUBEMAP_FACE_BOTTOM, texture.attribute("fileName"));
            texture = texture.nextSiblingElement("texture");
        }
    }
}

/**
 * Store the current scene content as XML data.
 *
 * @param name XML node name of the data.
 * @param doc XML document to store the data.
 * @return Current scene content data as XML data.
 */
QDomElement SceneCubeMap::domElement(const QString& name, QDomDocument& doc) const
{
    QDomElement de = doc.createElement(name);
    de.setAttribute("name", getName());

    for (unsigned int i=0; i<6; ++i)
    {
        QDomElement texture = doc.createElement("texture");
        switch (i) {
            case 0: texture.setAttribute("face", "front"); break;
            case 1: texture.setAttribute("face", "back"); break;
            case 2: texture.setAttribute("face", "left"); break;
            case 3: texture.setAttribute("face", "right"); break;
            case 4: texture.setAttribute("face", "top"); break;
            case 5: texture.setAttribute("face", "bottom"); break;
        }
        texture.setAttribute("fileName", m_textureFileNames[i]);
        de.appendChild(texture);
    }
    return de;
}
