#include <stdio.h>
#include <stdlib.h>
#include <glut.h>

#include "Scene.h"
#include "scene/SceneModel.h"

using namespace projection;

/**
 * Constructor.
 *
 * @param pScene Scene data.
 */
SceneModel::SceneModel(Scene* pScene) : SceneContent(pScene)
{
    m_bBackFaceCulling = true;
    m_bSmoothing = true;

    m_pModel = NULL;
    m_modelIndex = 0;
}

/**
 * Destructor.
 */
SceneModel::~SceneModel()
{
    if (m_pModel)
        glmDelete(m_pModel);
    if (glIsList(m_modelIndex))
        glDeleteLists(m_modelIndex, 1);
}

/**
 * Initialize screen content.
 */
void SceneModel::init()
{
}

/**
 * Draw the scene content.
 *
 * @param cameraFrame View point data.
 */
void SceneModel::draw(const qglviewer::Frame&)
{
    glPushAttrib(GL_ENABLE_BIT);

    if (m_modelIndex)
    {
        if (m_bBackFaceCulling)
            glEnable(GL_CULL_FACE);
        else
            glDisable(GL_CULL_FACE);
        glCallList(m_modelIndex);
    }

    glPopAttrib();
}

/**
 * Load a model file.
 *
 * @param fileName File name of a model to load.
 */
void SceneModel::setFileName(const QString& fileName)
{
    if (m_fileName != fileName)
    {
        m_fileName = fileName;

        if (m_pModel)
            glmDelete(m_pModel);
        if (glIsList(m_modelIndex))
            glDeleteLists(m_modelIndex, 1);

        m_pModel = glmReadOBJ(fileName.toLocal8Bit().data());
        glmFacetNormals(m_pModel);
        glmVertexNormals(m_pModel, 89.5f);
        m_modelIndex = glGenLists(1);
        glNewList(m_modelIndex, GL_COMPILE);
        glmDraw(m_pModel, (m_bSmoothing?GLM_SMOOTH:GLM_FLAT) | GLM_MATERIAL);
        glEndList();

        notifyRedraw();
    }
}

/**
 * Retrieve file name of the model.
 *
 * @return File name of the model.
 */
QString SceneModel::getFileName() const
{
    return m_fileName;
}

/**
 * Enable or disable back face culling.
 *
 * @param bEnable True to enable back face culling, false to disable.
 */
void SceneModel::setBackFaceCulling(bool bEnable)
{
    m_bBackFaceCulling = bEnable;
	notifyRedraw();
}

/**
 * Checke whether back face culling is enabled or not.
 *
 * @param bEnable True to enable back face culling, false to disable.
 */
bool SceneModel::getBackFaceCulling() const
{
    return m_bBackFaceCulling;
}

/**
 * Enable or disable smoothing.
 *
 * @param bEnable True to enable smoothing, false to disable.
 */
void SceneModel::setSmoothing(bool bEnable)
{
    if (m_bSmoothing != bEnable)
    {
        m_bSmoothing = bEnable;
        if (m_pModel) {
            if (glIsList(m_modelIndex))
                glDeleteLists(m_modelIndex, 1);
            m_modelIndex = glGenLists(1);
            glNewList(m_modelIndex, GL_COMPILE);
            glmDraw(m_pModel, (m_bSmoothing?GLM_SMOOTH:GLM_FLAT) | GLM_MATERIAL);
            glEndList();

            notifyRedraw();
        }
    }
}

/**
 * Checke whether smoothing is enabled or not.
 *
 * @param bEnable True to enable smoothing, false to disable.
 */
bool SceneModel::getSmoothing() const
{
    return m_bSmoothing;
}

/**
 * Retrieve bounding box of the scene content.
 *
 * @param min One corner of the scene content.
 * @param max Another corner of the scene content.
 */
void SceneModel::getBoundingBox(qglviewer::Vec& min, qglviewer::Vec& max)
{
    if (m_pModel)
    {
        GLfloat scale[3];
        glmDimensions(m_pModel, scale);
        min.setValue(-scale[0]/2.0f, -scale[1]/2.0f, -scale[2]/2.0f);
        max.setValue( scale[0]/2.0f,  scale[1]/2.0f,  scale[2]/2.0f);
    }
}

/**
 * Restore the scene content from XML data.
 *
 * @param element Parent XML element of the scene content data.
 */
void SceneModel::initFromDOMElement(const QDomElement& element)
{
    if (!element.isNull())
    {
        setFileName(element.attribute("fileName"));
        setBackFaceCulling(element.attribute("backFaceCulling")=="true"?true:false);
        setSmoothing(element.attribute("smoothing")=="true"?true:false);
    }
}

/**
 * Store the current scene content as XML data.
 *
 * @param name XML node name of the data.
 * @param doc XML document to store the data.
 * @return Current scene content data as XML data.
 */
QDomElement SceneModel::domElement(const QString& name, QDomDocument& doc) const
{
    QDomElement de = doc.createElement(name);
    de.setAttribute("name", getName());
    de.setAttribute("fileName", m_fileName);
    de.setAttribute("backFaceCulling", m_bBackFaceCulling?"true":"false");
    de.setAttribute("smoothing", m_bSmoothing?"true":"false");
    return de;
}
